/*+++++++++++
.LANGUAGE       C
.IDENTIFICATION astropos.c
.AUTHOR         Francois Ochsenbein [ESO-IPG]
.KEYWORDS       Test of Coordinate Transformations
.ENVIRONMENT    Any
.COMMENTS

.VERSION 1.0	27-Oct-1988: Creation
.VERSION 2.0	17-Feb-1992: Tested with TurboC
.VERSION 4.0	28-Apr-1995
.VERSION 4.1	28-Jun-1996: Take care of input precision
.VERSION 4.2	28-Jan-1997: Option -c -r
.VERSION 4.3	13-Jul-1998: Also error ellipses
.VERSION 4.4	09-Nov-1998: Don't force precision, except n.
.VERSION 4.5	18-Apr-2002; Accept Equinox/Epoch as Columns
.VERSION 4.6	19-Sep-2002: -mas -p options
.VERSION 4.7	25-Mar-2004: -n+ option
.VERSION 4.8	21-Mar-2007: Cosmetics
.VERSION 4.9	22-Jul-2008: -d option
.VERSION 4.91	06-Apr-2009: Pb in -r option...
.VERSION 4.92	12-Apr-2010: added readline
.VERSION 4.93	28-Aug-2010: 64-bit
.VERSION 4.94	25-Sep-2010: Accept -c1 (previously  -c1-1 required)
.VERSION 5.0 	17-Apr-2012: Use projections (+ for offsets / & for rho,theta)
.VERSION 5.01	06-Jul-2012: Bug for '&' (offsets in rho,theta)
.VERSION 5.1	15-Aug-2012: Accept CD matrix etc
.VERSION 5.11	29-Jun-2013: -mas was not working in non-tabular mode
	TODO: accept scale +  projection

-------------------------------------------------------------------------*/

#include <astropos.h>
#include <tranum.h>	/* Interpretation ascii text */
#include <ctype.h>
#include <stdio.h>
#include <unistd.h>	/* exit  */
#include <stdlib.h>	/* atoi  */
#include <strings.h>	/* strcy */
#include <string.h>	/* strcy */
#ifndef HAS_READLINE
#define HAS_READLINE 0
#endif
#if HAS_READLINE
#include <readline/readline.h>
#include <readline/history.h>
#define HISTORY_FILE ".queries"
char *history_file;
#endif

#define VERSION "5.11"

static char usage[] = "\
Usage: astropos [-v] -[2..9][dp.+m] [-d[delim]] [-cx-y[%off][,pm]] [-r] [-b]\n\
 [c=center] [x0=center] [y0=center] [CD=CD_matrix]\n\
 [-mas] [-pmas] input_sys:equinox/epoch output_sys:equinox/epoch...\n\
" ;

static char help[] = "\
    -v is a verbose option\n\
    -d defines a tab-separated input (-d, for a comma-separated input)\n\
   -c= specifies a center position; the input consists of offsets from there\n\
    -c option specifies the column positions (from 1) containing the RA/DE,\n\
       eventually the offset (arcsec/mas) preceded by a % sign or / (polar),\n\
       and eventually the proper motions (byte or char-delimited columns)\n\
    -r option asks to replace the input coordinates by the transformed ones\n\
       (default is to add the new coordinates preceded by a ;\n\
    -b asks to remove the useless blanks (output is shorter)\n\
  -mas when the input positions are in mas\n\
  +mas when the input offsets (rectangular or polar) are in mas\n\
 -pmas when the input proper motions are in mas/yr (default: arcsec/yr)\n\
    -2  to -9  indicate the precision for edition, from arcmin to 0.01mas\n\
       may be followed by the symbols\n\
       dot (.) to force the precision to the specified number\n\
       plus(+) to force the minimal precision to the specified number\n\
       'p' to edit the proper motions even if non-significant\n\
       'd' to edit in decimal degrees\n\
       'm' to edit in mas\n\
The known systems of coordinates (input_sys and output_sys) may be:\n\
     4=FK4; 5=FK5 ; I=ICRS ; G=Galactic; E=Ecliptic; S=Supergalactic\n\
Equinox and Epochs may be specified in:\n\
        Besselian years (prefix B)\n\
     or Julian years (prefix J)\n\
     or Julian date (prefix JD)\n\
     or as a column in the input stream when equinox replaced by -cx-y\n\
The input must consist of positions in the specified system, followed by:\n\
   '%' followed by an x and y offset (expressed in arcsec or mas with +mas)\n\
   '/' followed by a rho (expressed in arcsec/mas) and pos.angle (in deg)\n\
   ',' followed by the velocity components\n\
       mu_1(arcsec/yr)  mu_2(arcsec/yr)  RadialVel.(km/s)  Parallax(arcsec)\n\
       eventually followed by a colon (:) and the error ellipses\n\
" ;

static char *fl[] = {"?", "FK4", "Gal", "Sup", "Ecl", "FK5", "ICRS"};
static char verbop;

static char uprec ;		/* 'd' (dec. degrees) or 'm' (mas)    	*/
static int qprec ;
static int mprec ;		/* Minimal Precision, e.g. -4+		*/
static int odec ;		/* Offset of Declination Part for Decl 	*/
static int olen ;		/* Total length of edited coos         	*/
static double pm_factor = 1;	/* Changed by -pmas option		*/
static double of_factor = 3600;	/* Changed by +mas option		*/
static char *arg_center;	/* Center specified via -c= */
static double acenter[2];	/* Value of arg_center */
static int acenter_prec = -1;	/* Precision of arg_center      */
static double CD[2][2] ;	/* (x,y) = CD * (pixels-pix0)	*/
static double pix0[2] ;		/* (x,y) = CD * (pixels-pix0)	*/
static int _proj_ = _PROJ_ARC_;	/* Default Projection		*/

/* First/last column of input for Position Proper_Motion RVel Plx Eq Ep ox oy */
static int posdata[16] = {-1,-1,-1,-1,-1,-1,-1,-1, -1,-1, -1,-1,-1,-1,-1,-1 };
#define POS_coo     0
#define POS_pm      2
#define POS_rv      4
#define POS_plx     6
#define POS_eq      8
#define POS_ep     10
#define POS_of     12  /* Offset in x/y/theta */
/* Position in bytes of input */
static int posbyte[16] = {-1,-1,-1,-1,-1,-1,-1,-1, -1,-1, -1,-1,-1,-1,-1,-1 };
static char sepchar;		/* -d option */
static char separes = ';';	/* separates original data from result */

static struct spos s[120];
static int ns = 0;		/* Number of systems (0 = input) */
static char optr ;
static char optb ;
static char optp ;
static char opfp ;		/* Force Precision (e.g. 4.) */

static char *prompt;
static char prbuf[80];

#define ITEMS(x)	sizeof(x)/sizeof(x[0])
#define put(x)		printf("%s\n", x)
#define f0	s[0].frame

static int strloc(text, c)
/*++++++++++++++++
.PURPOSE  Locate specified character
.RETURNS  Index of located char
-----------------*/
  char *text;     /* IN: String to interpret */
  int  c;         /* IN: Char to look for */
{
  char *s;

    for ( s = text; *s; s++)      if (*s == c)    break;
    return(s-text);
}

static int set_posdata(str, posdata)
/*++++++++++
.PURPOSE Interpret the "-c" argument
.RETURNS Number of bytes interpreted.
-----------*/
  char	*str;	/* IN: String to interpret, e.g.  1-12,14-18, ... */
  int *posdata; /* OUT: The list of begin-end bytes */
{
  char *p; int k, n ;

    k=POS_coo, p=str; 	/* k=0: position, then pm ... */
    while (*p) {
	if (!isdigit(*p)) switch(*p) {
	    case '%': case '+': k = POS_of;    p++; continue;
	    case '/': case '&': k = POS_of+1;  p++; continue;
	    case ',': k = POS_pm;    p++; continue;
	    case ':': case '.': k++; p++; continue;
	    default:
		fprintf(stderr, "#***can't understand '%c' in -c%s\n", *p, str);
		return(0);
	}
    	for (n=0; isdigit(*p); p++) n = n*10 + (*p - '0') ;
    	if (n <= 0) {
	    fprintf(stderr, "#***invalid column (%d) in -c%s\n", n, p);
	    return(0);
	}
    	posdata[k] = n -1;
    	/*if (*p != '-') return(p-str) ; */
	if (*p == '-')
    	    for (++p, n=0; isdigit(*p); p++) n = n*10 + (*p - '0') ;
    	if (n < posdata[k]) {
	    fprintf(stderr, "#***impossible column %d-%d in -c%s\n",
		    posdata[k], n, str);
	    return(0) ;
    	}
    	posdata[++k] = n ;
	if (n >= BUFSIZ) {
	    fprintf(stderr, "#***impossible column %d(>%d) in -c%s\n",
		    n, BUFSIZ-1, str);
	    return(0) ;
	}
    }
    return(p-str) ;
}

static int set_bytes(byte, field, text)
/*++++++++++
.PURPOSE Transform a field position into a byte position
.RETURNS Number of fields involved (field[1] - field[0])
.REMARKS Set to -1 for field missing
-----------*/
  int *byte;	/* OUT: position in bytes (from 0) */
  int *field;	/* IN: position in field (from 0) */
  char	*text;	/* IN: String containing the fields */
{
  char *p; int k;

    byte[0] = byte[1] = -1;	/* Not here */
    if ((field[0]<0) || (field[1]<field[0]))
	return(-1);

    /* Locate position of field[0] */
    for (k=0, p=text; k<field[0]; k++) {
	p += strloc(p, sepchar);
	if (*p) p++;
	else return(-1);	/* Not here */
    }

    byte[0] = p-text;
    /* Locate now the other fields */
    while (1) {
	p += strloc(p, sepchar);
	if (!*p) break;		/* EOS found */
	if (++k >= field[1]) break;
	p++;			/* skip TAB  */
    }
    p += strloc(p, sepchar);
    byte[1] =  p-text;

    return( byte[1] - byte[0]);
}

static int copy_fields(dest, src, size, item)
/*++++++++++
.PURPOSE Copy src into dest, sizeof(dest)=size
.RETURNS Length of dest
-----------*/
  char *dest;	/* OUT: destination, max=COOLEN (32) */
  char *src;	/* IN: Source text */
  int size;	/* IN: length of source text */
  int item;	/* IN: index in posbyte / posdata  */
{
  int len;
    len = posbyte[item+1] -  posbyte[item];
    if (len>size) len=size;
    strncpy(dest, src+posbyte[item], len);
    if (len<size) dest[len] = 0;
    if ((sepchar) && ((posdata[item+1]-posdata[item])>1)) {
        int i;
	for (i=0; i<len; i++) {
	    if(dest[i] == sepchar) dest[i] = ' ';
	}
    }
    return(len);
}

static double get_date(str)
/*++++++++++
.PURPOSE Get a date in Julian Years.
.RETURNS the Julian date / -1.e12 when nothing found
-----------*/
  char	*str;
{
  double atof(), date;
  char	type;
  char	*p;

    type = 'J';			/* Default	*/
    p = str;
    while (isspace(*p)) p++ ;
    if (!*p) return(-1.e12) ;

    if (isalpha(*p))	type = toupper(*p), p++;
    while (isspace(*p)) p++ ;
    date = atof(p);
    switch(type) {
	case 'J':
	    if (!*p) date = 2000.;
	    break ;
	case 'D':	/* Julian Date */
	    if (!*p) date = -1.e12 ;
	    else date = JDtoJY(date);
	    break;
	case 'B':	/* Julian Date */
	    if (!*p) date = 1950.;
	    else date = BYtoJY(date);
	    break ;
	default:
	    fprintf(stderr, "++++Bad date ignored: %s\n", str) ;
	    return(-1.e12) ;
    }
    return(date);
}

int get_frame(str, s)
/*++++++++++
.PURPOSE Get all info for a Frame
.RETURNS The type of frame
-----------*/
  char *str;	/* IN: String */
  struct spos *s;	/* MOD: Info	*/
{
  int 	i, m;
  double atof();
  char	*q, x, get_Equinox;

    q = "J2000"; 	/* Default */
    m = strloc(str, ':');
    if (str[m]) x = toupper(str[m-1]), str += m+1 ;
    else if (isalpha(*str)) x = toupper(*str) ;
    else if (atoi(str) >= 2000) x = '5' ;
    else x = '4' ;
    switch(x) {
    	case 'I':   i = _COO_ICRS_;		get_Equinox=0; break;
    	case 'G':   i = _COO_GAL_;		get_Equinox=0; break;
    	case 'S':   i = _COO_SUPERGAL_;		get_Equinox=0; break;
    	case 'E':   i = _COO_ECL_;		get_Equinox=1; break;
	case 'B':
    	case '4':   i = _COO_FK4_; q="B1950";	get_Equinox=1; break;
	case 'J':
    	case '5':   i = _COO_FK5_;  		get_Equinox=1; break;
    	default :
	    fprintf(stderr, "****Coordinate Type must be 4 / 5 / G / E / S \n");
	    fprintf(stderr, "%s%s", usage, help);
	    exit(1);
    }
    s->frame = i;

    m = strloc(str, '/');
    x = str[m]; str[m] = 0;
    if (get_Equinox) {
	s->epoch = s->equinox = get_date(q) ;	/* Default Values */
	if ((str[0] == '-') && (str[1] == 'c')) {
	     set_posdata(str+2, posdata+POS_eq);
	     memcpy(posdata+POS_ep, posdata+POS_eq, 2*sizeof(int)) ;
	}
	else {
	    if (*str) s->epoch = s->equinox = get_date(str) ;
	}
    }
    else s->epoch = get_date(q);		/* Default Epoch  */

    str[m] = x;
    str += m;

    if (*str) {
	if (*str == '/') str++;
	if ((str[0] == '-') && (str[0] == 'c'))
	    set_posdata(str+2, posdata+POS_ep) ;
	else if (*str) s->epoch = get_date(str);
    }
    return(i);
}


static char *edit(s, has_pm)
/*++++++++++
.PURPOSE Edit the Results
.RETURNS The edited result
-----------*/
  struct spos *s;	/* IN: Info	*/
  int has_pm;		/* IN: non-zero if proper motions real */
{
  static char *masfmt1[] = { "%10.0f", "%12.1f", "%13.2f" };
  static char masfmt[20]; 	/* Final Format of mas Edition */
  static char *pmfmt[] = { "%+8.4f", "%+9.2f" };
  static double temp_coo[2] = {359., 89.} ;
  double q[2], v[3];
  static char line[BUFSIZ];
  int prec, i, show_equinox, width_coo;
  char *p, *fmt;
  static char sep = 0;

    width_coo = posbyte[1] - posbyte[0] ;
    if (width_coo < 8) width_coo = 31 ;

    /* Separator for position needed ? */
    if (sepchar && ((posdata[1] - posdata[0])>1)) 	/* Need 2 columns */
	sep = sepchar;


    /* Initialize the formats for edition (when precision specified) */
    if (qprec && (!odec)) {
	if (uprec == 'm') {
	    olen = width_coo;
	    p = masfmt1[qprec-7];	/* Format of 1 coo */
	    odec = atoi(p+1) ;		/* Width of RA */
	    i = olen-2*odec;		/* Number of blanks to insert */
	    if (i>0) sprintf(masfmt, "%s%%%ds%s", p, i, p) ;
	    else sprintf(masfmt, "%s%%s%s", p, p) ;
	    odec += i;			/* Where Dec starts */
	}
	else {
	    if (uprec == 'd') ed2_o(line, width_coo+1, temp_coo, qprec);
	    ed2_q(line, width_coo+1, temp_coo, qprec);
	    olen = strlen(line) ;
	    odec = strloc(line, '+') ;
	}
    }
    tr_sqv(s, q, v);
    if (s->precision < mprec) {		/* V4.7: Minimal Precision */
	if (s->precision > 0) s->precision  = mprec;
    }
    if (uprec == 'm') {
	i = qprec ? qprec : s->precision;
	if (i<7) i=7;
	if (i>9) i=9;
	if (!masfmt[0]) sprintf(masfmt, "%s%s", masfmt1[i-7], masfmt1[i-7]);
	sprintf(line, masfmt, q[0]*3.6e6, "", q[1]*3.6e6) ;
    }
    else if (((s->frame == _COO_FK4_) || (s->frame == _COO_FK5_))
      && (uprec == 0)) {
	if (qprec) {
	    /* Here we align the RA / Dec whatever the precision is */
	    if ((!opfp) && (qprec > s->precision) && (qprec >= 5)) {
		prec = s->precision ;
		if (prec < 4) prec = 4 ;
		ed_q1(line, q, prec, 0); 	/* Edit RA */
		i = strlen(line) ;
		if ((prec < 5) && (qprec > 4)) line[i++] = '.' ;
		while (i<odec) line[i++] = ' ' ;
		prec = s->precision ;
		if (prec == 4) prec++ ;
		ed_q1(line+i, q, prec, 1) ;	/* Edit Dec */
		i = strlen(line) ;
		if ((prec == 5) && (qprec > 5)) line[i++] = '.' ;
		while (i<olen) line[i++] = ' ' ;
		line[i] = 0 ;
	    }
	    else ed_q(line, q, qprec); 	/* With Full Precision */
	}
	else ed2_q(line, width_coo+1, q, s->precision);
    }
    else {
    	if (qprec) ed_o(line, q, qprec);
    	else ed2_o(line, width_coo+1, q, s->precision);
    }

    if (sep) {				/* Add separator */
	i = 0;	/* non-zero if sign found */
	for (p=line+strlen(line)-1; (i==0) && (p>line); p--) {
	    if ((*p == '+') || (*p == '-')) i = 1;
	}
	if (p == line) {	/* no sign ? */
	    while (*p == ' ') p++;
	    while (isgraph(*p)) p++;
	}
	if (isspace(*p)) *p = sep;
	else {
	    ++p; i = strlen(p);
	    while (i>=0) p[i+1] = p[i];
	    *p = sep;
	}
    }
    /* if (posdata[0] >= 0) return(line) ; */

    /* Add the proper motions, separated by " , " */
    p = line + strlen(line) ;
    if (has_pm) {
	sep = 	/* Separator between proper motion components ? */
	    posdata[POS_pm+1]-posdata[POS_pm]>1 ? sepchar : 0;
        strcpy(p, " , "); p += strlen(p);
        if (posdata[0] >= 0) ; 	/* Taken from Input File */
        else sprintf(p, "Vel%s", pm_factor==1 ? " (\"/yr)" : ":mas/yr ");
        p += strlen(p);
        fmt = pm_factor==1 ? pmfmt[0] : pmfmt[1];  /* Format arcsec/mas/yr */
        if (posdata[0] >= 0) ; 	/* Taken from Input File */
        else { *(p++) = ':'; sprintf(p, fmt, v[0]*pm_factor); p += strlen(p); }
        sprintf(p, fmt, v[1]*pm_factor);  p += strlen(p);
	if (sep) *(p++) = sep;
        sprintf(p, fmt, v[2]*pm_factor);  p += strlen(p);
    }
    if (posdata[0] >= 0) ;	/* Taken from File */
    else {
	sprintf(p, " \n\t\t%s Ep=J%10.5f", fl[s->frame&0xf], s->epoch);
	p += strlen(p);
	show_equinox  = (s->frame == _COO_FK4_) || (s->frame == _COO_FK5_)
		     || (s->frame == _COO_ECL_) ;
	if (show_equinox) p += strlen(p),
	    sprintf(p, " Eq=J%10.5f", s->equinox);
    }

    if (optb) {		/* Remove useless blanks */
	char *b;
	p=b=line;
	while (*p) {
	    if (*p == ' ') {
		while (*p == ' ') p++;
		if ((b>line) && (*p != sep)) *(b++) = ' ';
	    }
	    else if (*p == sep) {
		*(b++) = *(p++) ;
		while (*p == ' ') p++;
	    }
	    else *(b++) = *(p++) ;
	}
	if ((b>line) && (b[-1] == ' ')) b--;
	*b = 0;
    }

    return(line);
}

static void xy2pos(double xy[2], double o[2], double center[2])
/*++++++++++++++++
.PURPOSE  Compute the Position from the x,y
.RETURNS  ---
-----------------*/
{
  static double Rlocal[3][3], Center[2] = { 9.e9, -9.e9 };
  double  XY[2], u[3], pix[2] ;
    if ((center[0] != Center[0]) || (center[1] != Center[1])) {
	Center[0] = center[0];
	Center[1] = center[1];
        tr_oR(center, Rlocal) ;
    }
    pix[0] = xy[0] - pix0[0] ;
    pix[1] = xy[1] - pix0[1] ;
    XY[0] = CD[0][0] * pix[0] + CD[0][1] * pix[1] ;
    XY[1] = CD[1][0] * pix[0] + CD[1][1] * pix[1] ;
    tr_pu(_proj_, XY, u) ;
    tr_uu1(u, u, Rlocal) ;
    tr_uo(u, o) ;
}

/*============================================================================*/
int main (argc, argv)
/*++++++++++++++++
.PURPOSE  The program transforms astronomical coordinates
.RETURNS  ---
.REMARKS
-----------------*/
  int argc; 		/* IN: Number of arguments */
  char **argv;		/* IN: Vector of arguments */
{
  char *p, *c, *pm, *abuf, buf[BUFSIZ], apos[64];
  double eq0=0, ep0=0;	/* Default equinox & epoch */
  double q[2], dq[2], w[4], v[3];
  int 	i, n, nd, lineno, has_pm, has_of, arg_eq, has_CD=0;


	/* Look at Options */
    while (-- argc > 0) {
	 p = *++argv;
	 if (verbop) fprintf(stderr, " %s", p);
	 arg_eq = p[strloc(p, '=')];
	 if (arg_eq || (*p == '-')) {
	     if (*p == '-') p++;
	     switch(*p) {
	       case 'r':  optr=1; continue;
	       case 'h':  printf("%s%s", usage, help); exit(0);
	       case 'v':  if(strncmp(p, "ver", 3) == 0) {
			    printf("astropos (CDS) Version %s\n", VERSION);
			    exit(0);
			}
			verbop = isdigit(p[1]) ? atoi(p+1) : 1;
			fprintf(stderr, "# Arguments:");
			continue;
	       case 'b': optb = 1; continue;
	       case 'c':
		 while (isalpha(*p)) p++;
		 if (*p == '=') {	/* A center is specified */
		    arg_center = p+1;
    		    acenter_prec = tr_aq(arg_center, strlen(arg_center), 
			    acenter);
		    if (acenter_prec <= 2) {
			fprintf(stderr, "#***Bad center: %s\n", p);
			exit(1);
		    }
		    continue;
		 }
		 i = set_posdata(p, posdata);
		 if (!p[i]) continue ;
		 fprintf(stderr, "\n****Bad argument: '%s'\n", *argv);
		 exit(1) ;
	       case 'd':	/* V4.9 */
		 sepchar = p[1];
		 if (!sepchar) sepchar = '\t';
		 continue;
	       case '2': case '3': case '4': case '5':
	       case '6': case '7': case '8': case '9':
		 qprec = *p -'0' + 1;
		 for (++p; *p; p++) {
		    if (*p == '.') opfp = 1;
		    else if (*p == '+') { mprec = qprec; qprec = 0; }
		    else if (*p == 'm') uprec = *p;
		    else if (*p == 'd') uprec = *p;
		    else if (*p == 'p') optp  = 1;
		    else fprintf(stderr, "\n++++Unknown option '%c' in '%s'\n",
			*p, *argv);
		 }
		 continue;
	       case 'm':	/*milli-arc-sec */
	         if (!qprec) qprec = 7;
	         uprec = *p;
	         continue;
	       case 'p':	/* Proper motions + Parallaxes in mas */
	         if (p[1] == 'm') pm_factor = 1000.;
	         else goto bad_option;
	         continue;
	       case 'x': 	/* Center of CCD */
		 if (!arg_eq) goto bad_option;
		 pix0[0] = atof(strchr(p, '=')+1);
		 continue;
	       case 'y':
		 if (!arg_eq) goto bad_option;
		 pix0[1] = atof(strchr(p, '=')+1);
		 continue;
	       case 'C':
		 if ((p[1] != 'D') || (p[2] != '=')) goto bad_option;
		 if(tr_afn(p+3, -1, &CD[0][0], 4) != 4) {
		     fprintf(stderr, "#***Bad CD matrix: %s\n", p);
		     fprintf(stderr, "#   %s\n",
			 "must be value_11,value_12,value_21,value_22");
		     exit(1);
		 }
		 has_CD = 1;
		 continue;
	     bad_option:
	       default:
		 fprintf(stderr, "\n****Bad option: -%s\n", p);
		 fprintf(stderr, "%s", usage);
		 exit(1);
	     }
	 }
	 else if ((p[0]=='+') && (p[1] == 'm')) {
	     of_factor *= 1000.;
	     continue;
	 }
	 if (ns >= ITEMS(s)) {
	     fprintf(stderr, "\n****Limited to %d coordinate computations\n",
		(int)(ITEMS(s)-1));
	    exit(1);
	 }
	 get_frame(p, s+ns);
	 if (ns == 0) { ep0 = s[0].epoch; eq0 = s[0].equinox; }
	 ns++;
    }
    if (verbop) fprintf(stderr, "\n");
    if (ns==1) s[ns++] = s[0];
    if (ns < 2) {
	fprintf(stderr,
	    "****You didn't specify at least input and output frames\n");
	fprintf(stderr, "%s%s", usage, help);
	exit(1);
    }
    /* Replacement of position + proper motion: require position BEFORE */
    if (optr && (posdata[POS_pm]>=0) && (posdata[POS_pm] <= posdata[POS_coo])){
	fprintf(stderr, "****astropos: %s\n",
		"the proper motions must FOLLOW the position!");
	exit(1);
    }

    /* Decide for a prompt */
    if ((!prompt) && isatty(0)) {
	p =  arg_center ? "offset from center" : "Position %x.y/rho.pa" ;
	sprintf(prbuf, "\nInput %s %s , mu1 mu2 [RV [plx]]: ",
		fl[f0&0xf], p);
	prompt = prbuf;
    }

    /* Char separating results: use same separators */
    if(sepchar) separes = sepchar;
    abuf = (char *)0;

    for (lineno=0;;lineno++) {
	if (abuf && (abuf != buf)) free(abuf);
	abuf = (char *)0;
	if (prompt && isatty(0)) {
#if HAS_READLINE
	    if (!history_file) {
	        using_history();
	        stifle_history(1000);
	        history_file = getenv("HISTORY_FILE");
	        if (!history_file) {
	            p = getenv("HOME");
	            i = p ? strlen(p)+1 : 0;
	            history_file = malloc(i+1+sizeof(HISTORY_FILE));
	            if (p) {
		        strcpy(history_file, p);
		        strcat(history_file, "/");
    	            }
    	            else *history_file = 0;
    	            strcat(history_file, HISTORY_FILE);
    	        }
    	        if (read_history(history_file))
    	            write_history(history_file);
            }
            abuf = readline(prompt);
	    if (!abuf) { printf("\n"); break; }
            /* printf("%s", prompt); */
	    add_history(abuf);
#else
	    if (prompt) printf("%s", prompt);
    	    /* printf("\nInput Position , mu1 mu2 RV plx "); */
            if (!fgets (buf, sizeof(buf), stdin))   break;
	    abuf = buf;
#endif
	}
	else { 	/* Not a tty: read */
	    if (!fgets (buf, sizeof(buf), stdin))   break;
	    abuf = buf;
	}
	i = strlen(abuf)-1; if (abuf[i] == '\n') abuf[i] = 0;
	if (!abuf[0])         { put(abuf) ; continue ; }
	if (abuf[0] == '%')   { put(abuf) ; continue ; }
	if (abuf[0] == '#')   { put(abuf) ; continue ; }
	if (abuf[0] == '!')   { put(abuf) ; continue ; }

    	for (i=0; i<4; i++)	w[i] = 0;
	if (posdata[0] >= 0) {	/* Data in dedicated Columns 	*/
	    /* V4.9: char-delimited columns => convert posdata into posbyte */
	    if (sepchar) for (i=0; i<ITEMS(posdata); i += 2) {
		if (posdata[i]>=0) {
		    if (set_bytes(posbyte+i, posdata+i, abuf)<0) {
			/* If position not there ==> may be a comment! */
			if (i == 0) break;
		    }
		}
		else { posbyte[i] = posdata[i]; posbyte[i+1] = posdata[i+1]; }
	    }
	    else memcpy(posbyte, posdata, sizeof(posdata));
	    if ((p = arg_center))         /* Center specified via argument c= */
		i = strlen(p);
	    else if (posbyte[0]<0) { put(abuf) ; continue ; }  /* No position */
	    else { p = abuf+posbyte[0]; i = posbyte[1] - posbyte[0]; }
	    for (n=nd=0; n<i; n++) {		/* Look for text */
		if (p[n] == sepchar) continue;
		if (isalpha(p[n])) break ;
		if (isdigit(p[n])) nd++;
		if ((p[n] == '(') || (p[n] == '[')) break ;
	    }
	    if ((n<i)||(nd==0)) { put(abuf); continue ; }
	    /* V4.9: if separators exist, remove them first... */
	    i = copy_fields(apos, abuf, sizeof(apos), POS_coo);
	    if (uprec == 'm') {
		n = tr_afn(apos, i, q, 2) ;	/* 2-d Vector */
		q[0] /= 3.6e6; q[1] /= 3.6e6;
		/* Assumed precision: 10 bytes for an angle in mas,
		   and another byte for the decimal point.
		*/
		if (n>0) {
		    n = 7 + (i-22)/2 ;
		    if (n<7) n = 7;
		    if (n>9) n = 9;
		}
	    }
    	    else if ((f0 == _COO_FK4_) || (f0 == _COO_FK5_))
    		 n = tr_aq(apos, i, q);
    	    else n = tr_ao(apos, i, q);
	    s->precision = n ;
	    /* Get Equinox & Epoch */
	    if (posbyte[POS_eq] >= 0) {	/* Equinox */
		n = abuf[posbyte[POS_eq+1]]; abuf[posbyte[POS_eq+1]] = 0;
		s->equinox = s->epoch = get_date(abuf+posbyte[POS_eq]);
		abuf[posbyte[POS_eq+1]] = n;
	    }
	    if (posbyte[POS_ep] >= 0) {	/* Epoch   */
		n = abuf[posbyte[POS_ep+1]]; abuf[posbyte[POS_ep+1]] = 0;
		s->epoch = get_date(abuf+posbyte[POS_ep]);
		abuf[posbyte[POS_ep+1]] = n;
	    }
	    if (s->equinox <= -1.e12) s->equinox = eq0;
	    if (s->epoch   <= -1.e12) s->epoch   = ep0;
	    /* Get offset from input */
	    if ((has_of = posbyte[POS_of+1]  >= 0)) { int ofx;
		ofx = posbyte[POS_of]>=0 ? POS_of : POS_of+1 ;
		n = 2;	/* 2 fields */
		i = copy_fields(apos, abuf, sizeof(apos), ofx);
		tr_af1(apos, i, dq, 2) ;
		if (ofx != POS_of) { /* Polar: rho + theta */
		    double rho = dq[0];
		    dq[0] = rho * sind(dq[1]);
		    dq[1] = rho * cosd(dq[1]);
		}
		q[0] += dq[0]/of_factor/cosd(q[1]);
		q[1] += dq[1]/of_factor;
	    }
	    n = 0 ;
	    if ((has_pm = posbyte[POS_pm] >= 0)) {    /* Proper Motion */
		n=2;	/* 2 fields */
		/* V4.9: if separators exist, remove them first... */
		i = copy_fields(apos, abuf, sizeof(apos), POS_pm);
		tr_af1(apos, i, w, 2) ;
	    }
	    if (posbyte[POS_rv] >= 0) n++, /* Radial Veloci */
		tr_af1(abuf+posbyte[4], posbyte[5]-posbyte[4], w+2, 1) ;
	    if (posbyte[POS_plx] >= 0) n++, /* Parallax     */
		tr_af1(abuf+posbyte[6], posbyte[7]-posbyte[6], w+3, 1) ;
	}
	else { int inext=0; char cnext;	/* Free-text input positions */
	    if (isalpha(abuf[0])) { put(abuf); continue; }
	    if ((p = arg_center)) { 
		i = strlen(arg_center); 
		q[0] = acenter[0]; q[1] = acenter[1];
		cnext = '%';
		n = acenter_prec;
	    }
	    else {		/* position in text */
		p = abuf;
		for (i=0; isprint(abuf[i]); i++) {
		    if (strchr(",%/&", abuf[i])) break;
		}
		inext = i;
		cnext = abuf[i];
	        if (uprec == 'm') {
		    n = tr_afn(p, i, q, 2) ;	/* 2-d Vector */
		    q[0] /= 3.6e6; q[1] /= 3.6e6;
		    if (n>0) {
		        n = 7 + (i-22)/2 ;
		        if (n<7) n = 7;
		        if (n>9) n = 9;
		    }
		}
		else if ((f0 == _COO_FK4_) 
	              || (f0 == _COO_FK5_) || (f0 == _COO_ICRS_))
    		     n = tr_aq(p, i, q);
    	        else n = tr_ao(p, i, q);
	    }
    	    if (n <= 0)	{ put(abuf) ; continue ; }
	    s->precision = n ;
	    p = abuf+inext;	    /* Next part of position: offset , pm */
	    if ((cnext == '%') || (cnext == '/') || (cnext == '&')) {
		/* There are offsets */
		char which_offset = cnext;
		if (*p == cnext) p++;
		i = strloc(p, ','); cnext = p[i];    /* Then the pm... */
		tr_af1(p, i, dq, 2) ;
		if (which_offset != '%') { 	 /* Polar: rho + theta */
		    double rho = dq[0];
		    dq[0] = rho * sind(dq[1]);
		    dq[1] = rho * cosd(dq[1]);
		}
		if (has_CD) xy2pos(dq, q, q);
		else {
		    q[0] += dq[0]/of_factor/cosd(q[1]);
		    q[1] += dq[1]/of_factor;
		}
		p += i;
	    }
	    if ((has_pm = cnext == ',')) p++;
    	    n = tr_af1(p, strlen(p), w, 4);
	}
    	v[0] = kms_pc * w[2] * w[3]/pm_factor ;
    	v[1] = w[0]/pm_factor ;
    	v[2] = w[1]/pm_factor ;
    	tr_qvs(q, v, s);

    	if ( (n == 0) && (f0 == _COO_FK4_))	/* Supply Proper Motions */
    		fk4_s0(s);

    	/* Edit Input Data	*/

    	if (posdata[0]<0)
	    printf("\nInput: "), put(edit(s, has_pm|optp));

	/* Compute & Edit Output Data	*/
    	for (n=1; n<ns; n++) {
    	    tr_ss(s, s+n);
	    c = edit(s+n, has_pm|optp);
	    if (posdata[0]<0) { printf("\nOutput:"); puts(c); continue; }
	    pm = strstr(c, " , ");

	    if (sepchar && pm) {	/* Use delimiter for proper motions */
		*(pm++) = sepchar;
		p = pm+2;
		if (optb) { while(*p == ' ') p++; }
		for (i=0; p[i]; i++) pm[i] = p[i];
		pm[i] = 0;
		--pm;			/* pm -> beginning of proper motions */
	    }

	    if (n == 1) {		/* First line: issue input...  */
	        if (optr) {		/* Replace current coordinates */
		    fwrite(abuf, 1, posbyte[0], stdout);
		    i = pm ? pm-c : strlen(c);
		    fwrite( c , 1, i,  stdout);	/* Coordinates */
		    if (pm && ((i = posbyte[POS_pm]-posbyte[1])>=0)) {
			fwrite(abuf+posbyte[1], 1, i, stdout);
	    	        printf("%s%s", pm+1, abuf+posbyte[POS_pm+1]) ;
		    }
		    else  printf("%s", abuf+posbyte[1]);  /* Added V4.91 */
		}
		else printf("%s%c%s", abuf, separes, c) ;
	    }
	    else printf("%c%s", separes, c) ;
    	}
	if (posdata[0]>=0) putchar('\n') ;
    }
    if (abuf != buf) free(abuf);
    exit(0);
}
