/*++++++++++++++
.IDENTIFICATION ccd.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein, Isabelle Bourekeb [CDS]
.ENVIRONMENT    Unix, CDS Catalogues
.KEYWORDS
.VERSION   1.0  30-May-2001: Installation
.VERSION   1.1  05-Apr-2004: Edition Sexagesimal
.VERSION   1.2  03-May-2005: Addition ccd_locate
.VERSION   1.3  15-Oct-2005: Add ccd_fromfile
.VERSION   1.4  30-Nov-2005: Accept lines without '=' embedded in input
.VERSION   1.5  15-Dec-2005: Accept arguments sigma=  score=
.VERSION   1.51 16-Jan-2006: Verification scale changes
.VERSION   1.52 25-Mar-2006: Remove the NULL values
.VERSION   1.53 22-Jan-2007: Accept a header which specifies X Y V
.VERSION   1.54 15-Feb-2007: Edition of x0 y0 with adapted number of decimals
.VERSION   1.55 01-Apr-2007: Cosmetics
.VERSION   1.56 25-Aug-2009
.VERSION   1.57 04-Aug-2013: Check the number of outliers; faster getScore
.VERSION   1.58 23-Sep-2013: bug fix
.VERSION   1.59 07-Oct-2013: add error= in vizPosition

.COMMENTS
Computation of Transformations (x,y) to Positions
-------------------------------------------------
Let's assume we've to compute the (x0 y0) and CD matrix such that
    | X |     | CD00   CD01 |  | x-x0 |
    |   | =   |             |  |      |  (Z = A (z-z0) in complex)
    | Y |     | CD10   CD11 |  | y-y0 |

   var(x)   = Sum((x-x0)^2)
   var(y)   = Sum((y-y0)^2)
   cov(x,y) = Sum((x-x0)(y-y0))
   cov(X,x) = Sum(X(x-x0))
   cov(X,y) = Sum(X(y-y0))
   cov(Y,x) = Sum(Y(x-x0))
   cov(Y,y) = Sum(Y(y-y0))

   the terms of the CD matrix ,   if  D = var(x).var(y) - cov(x,y)^2
   CD00 = [var(y).cov(X,x) - cov(x,y).cov(X,y)] / D
   CD01 = [var(x).cov(X,y) - cov(x,y).cov(X,x)] / D
   CD10 = [var(y).cov(Y,x) - cov(x,y).cov(Y,y)] / D
   CD11 = [var(x).cov(Y,y) - cov(x,y).cov(Y,x)] / D

   while the (x0 y0) can be computed from residuals

   | x0 |    | E(x) |          | E(X) |
   |    | =  |      | - CD^-1^ |      |
   | y0 |    | E(y) |          | E(Y) |

   E(x) = Sum(x)/n = math. expectation
---------------*/

#define VERSION "1.59"
#include <astropos.h>
#include <tranum.h>
#include <string.h>	/* Standard String manipulations */
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <ctype.h>
#include <time.h>

#ifndef MAX_REFSTARS
#define MAX_REFSTARS 2000
#endif

static int _proj_ = _PROJ_TAN_ ;	/* Default Projection	*/
static int verbop = 0 ;			/* Display ccd_compute  */
static double Rlocal[3][3];
static double vartot ;
static double sigma = -1.;		/* Last computed sigma  */
static double ascale[2];		/* Scale in argument 	*/
static double ascore[2];		/* Score in argument 	*/
static double asigma[2];		/* sigma in argument 	*/

static int test_points = 36;		/* -t option */

typedef int (*SORT_FCT)();

#define MIN(a,b)	((a)<=(b)?a:b)
#define MAX(a,b)	((a)>=(b)?a:b)
#define ITEMS(a)	(sizeof(a)/sizeof(a[0]))
#define bzero(a,l)      memset(a,0,l)
#define index(x)	(int)((intp)(x))

/*==================================================================
		Basic Utilities
 *==================================================================*/

static void computeXY(double *XY, double *xy, int n,
       double *xy0, double CD[2][2])
/*++++++++++++++++
.PURPOSE  Compute the XY array from xy: XY = CD * (xy - xy0)
	(in complex form:  Z = A * (z - z0))
.RETURNS  ---
-----------------*/
{
    while (--n>=0) {
	XY[0] = CD[0][0]*(xy[0]-xy0[0]) + CD[0][1]*(xy[1]-xy0[1]);
	XY[1] = CD[1][0]*(xy[0]-xy0[0]) + CD[1][1]*(xy[1]-xy0[1]);
	xy += 2; XY += 2;
    }
}

static double *range(double *vector, int n)
/*++++++++++++++++
.PURPOSE  Compute min/max of an array
.RETURNS  (static) values of min/max
-----------------*/
{
  static double extremes[2], *e, *v ;
    v = vector; e = v + n ;
    extremes[0] = extremes[1] = *(v++) ;
    while (v < e) {
	if (*v < extremes[0]) extremes[0] = *v ;
	if (*v > extremes[1]) extremes[1] = *v ;
	v++ ;
    }
    return(extremes);
}

static void range2(double *xy, int n, double *xylim)
/*++++++++++++++++
.PURPOSE  Compute min/max of an array of (xy)
.RETURNS  ---
.REMARKS  xylim contains in output (Xmin, Ymin, Xmax, Ymax)
-----------------*/
{
    xylim[0] = xylim[2] = xy[0];
    xylim[1] = xylim[3] = xy[1];
    for (xy += 2; --n>0; xy += 2) {
       if (xy[0] < xylim[0]) xylim[0] = xy[0];
       if (xy[1] < xylim[1]) xylim[1] = xy[1];
       if (xy[0] > xylim[2]) xylim[2] = xy[0];
       if (xy[1] > xylim[3]) xylim[3] = xy[1];
    }
}

static double minv2(double a[2][2], double b[2][2])
/*++++++++++++++++
.PURPOSE  Inverse 2x2 matrix
.RETURNS  Determinant
.REMARKS  Straightforward
-----------------*/
{
  double d;
    d = a[0][0] * a[1][1] - a[0][1] * a[1][0] ;
    if(d) {
        b[0][0] =  a[1][1] / d ;
	b[1][1] =  a[0][0] / d ;
	b[0][1] = -a[0][1] / d ;
	b[1][0] = -a[1][0] / d ;
    }
    else memset(b, 0, sizeof(b)) ;
    return(d);
}

static double E2(double *x, double *e)
/*++++++++++++++++
.PURPOSE  Compute mathmatical expectation
.RETURNS  Its value
.REMARKS  In a 2-column matrix
-----------------*/
{
  double r, *p; int n ;
    for (p=x, r=0; p < e; p += 2)
      r += *p ;
    n = (p-x)/2 ;
    return(r/n);
}

static double E2p(double *x, double *e)
/*++++++++++++++++
.PURPOSE  Compute mathmatical expectation (periodic 360)
.RETURNS  Its value
.REMARKS  In a 2-column matrix
-----------------*/
{
  double r, d, o, *p; int n ;
    o = *x; p = x; p += 2 ;
    for (r=0, p += 2; p < e; p += 2) {
      d = *p - o ;
      if (d > 180.) d -= 360 ;
      if (d <-180.) d += 360 ;
      r += d ;
    }
    n = (p-x)/2 ;
    return(o + r/n);
}

static double E4(double *x, double *e)
/*++++++++++++++++
.PURPOSE  Compute mathmatical expectation
.RETURNS  Its value
.REMARKS  In a 4-column matrix
-----------------*/
{
  double r, *p; int n ;
    for (p=x, r=0; p < e; p += 4)
      r += *p ;
    n = (p-x)/4 ;
    return(r/n);
}

static double cov4(double *x, double *y, double *e)
/*++++++++++++++++
.PURPOSE  Compute covariance
.RETURNS  Its value
.REMARKS  In a 4-column matrix
-----------------*/
{
  double r, *p; int n ;
    for (p=x, r=0; p<e; p += 4, y += 4)
      r += *p * *y;
    n = (p-x)/4 ;
    return(r/n);
}

/*==================================================================
		Compute (exactly) the CD matrix in
 *==================================================================*/

#if 0
static void mat2(double *XY, double *xy, double CD[2][2])
/*++++++++++++++++
.PURPOSE  Compute the 2D matrix CD sych that
	XY = CD * (xy - xy0)
	(in complex: compute A and z0 | Z = A * (z - z0))
.RETURNS  ---
.REMARKS  xy = (x1 y1 x2 y2)  	XY = (X1 Y1 X2 Y2)
-----------------*/
{
  double dX, dY, dx, dy, D;
    dX = XY[2] - XY[0];
    dx = xy[2] - xy[0];
    dY = XY[3] - XY[1];
    dy = xy[3] - xy[1];
    D = dx*dx + dy*dy;
    CD[0][0] = (dX*dx + dY*dy) / D;
    CD[0][1] = (dX*dy - dY*dx) / D;
    CD[1][0] = -CD[0][1];
    CD[1][1] =  CD[0][0];
}
#endif

static double solve2(double *XY, double *xy, int k,
       double *xy0, double CD[2][2])
/*++++++++++++++++
.PURPOSE  Solve the 2-point problem: compute CD & xy0 where
	XY = CD * (xy - xy0)
	(in complex: compute A and z0 | Z = A * (z - z0))
.RETURNS  The scale (rad/pix)
.REMARKS  xy = (x1 y1 x2 y2)  	XY = (X1 Y1 X2 Y2)
	X2-X1 = scale*(cos(t)*(x2-x1) + sin(t)*(y2-y1))
   k=0,1,2,3 for the variations: k&1 ==> symmetry
				 k&2 ==> exchange points
-----------------*/
{
  double scale2, dX, dY, dx, dy, D;
    dX = XY[2] - XY[0];
    dY = XY[3] - XY[1];
    dx = xy[2] - xy[0];
    dy = xy[3] - xy[1];
    D = dx*dx + dy*dy;
    scale2 = (dX*dX + dY*dY) / D;	/* In fact, scale^2 */

    if (k&2) { dx = -dx; dy = -dy; }
    if (k&1) {				/* Symetry */
        CD[0][0] = (dX*dx - dY*dy) / D;
        CD[0][1] = (dX*dy + dY*dx) / D;
        CD[1][0] =  CD[0][1];
        CD[1][1] = -CD[0][0];
    }
    else {				/* No swap */
        CD[0][0] = (dX*dx + dY*dy) / D;
        CD[0][1] = (dX*dy - dY*dx) / D;
        CD[1][0] = -CD[0][1];
        CD[1][1] =  CD[0][0];
    }

    /* Compute the origin */
    D = CD[1][1]*CD[0][0]-CD[0][1]*CD[1][0];	/* Determinant */
    xy0[0] = xy[0] - ( CD[1][1]*XY[0] - CD[0][1]*XY[1])/D;
    xy0[1] = xy[1] - (-CD[1][0]*XY[0] + CD[0][0]*XY[1])/D;

    return(sqrt(scale2));
}

/*==================================================================
		Compute the Transformation
 *==================================================================*/

int ccd_setproj(int proj)
/*++++++++++++++++
.PURPOSE  Set the projection system
.RETURNS  The previous projection
.REMARKS  _proj_ system is set.
-----------------*/
{
  int o ;
    o = _proj_ ;
    _proj_ = proj ;
    return(o) ;
}

double ccd_sigma()
/*++++++++++++++++
.PURPOSE  Returns the last computed sigma (from ccd_compute, ccd_center
		or ccd_fromlist)
.RETURNS  last sigma in radians.
.REMARKS  -1. if no computation done.
-----------------*/
{
    return(sigma);
}

int ccd_compute(
   double *pos, double *xy, int n, 	/* IN: Positions / CCD xy */
   double center[2], double x0[2], 	/* MOD: Center definition */
   double CD[2][2]			/* OUT: the matrix */
   )
/*++++++++++++++++
.PURPOSE  Adjust the CD matrix and x0, projection center being FIXED
.RETURNS  Status 0=OK  >0 = Number of points with problems
.REMARKS  _proj_ system is used.
	The sigma can be retrieved via ccd_sigma.
-----------------*/
{
  double  *pX, *aX, *eX ;	/* Global --> local frame */
  double  den, vx, vy, cv, mx, my, epsilon ;
  double u[3], xc[2], Ex[4], CD1[2][2], dxy[2], *pp, *px ;
  char fxy[20] ;		/* Format to edit x,y */
  int i, k, goon ;

    sigma = -1.;

    if (n < 6) {
        fprintf(stderr,
	    "****Can't transform pixels to position with only %d point\n", n);
	return(-1) ;
    }

    /* How to edit x,y ? */
    pX = range(xy, n*2) ;
    epsilon = (pX[1] - pX[0])*1.e-4 ;
    vx = fabs(pX[0]); vy = fabs(pX[1]) ;
    if (vy > vx) vx = vy ;
    for (den = 1.e6, i=0; (vx < den) && (den >= 1.); den /= 10.) i++ ;
    sprintf(fxy, "%%9.%df %%9.%df", i, i) ;

    /* xc[0] = xc[1] = 0 ; */
    xc[0] = x0[0]; xc[1] = x0[1];
    /*
    fprintf(stderr, "----Center=%010.6f%+010.6f (", center[0], center[1]) ;
    fprintf(stderr, fxy, xc[0], xc[1]) ;
    fprintf(stderr, ") --\n");
    */

    /* Matrix X Y x y with Reduced Positions */
    aX = (double *)malloc(n*4*sizeof(double)) ;
    eX = aX + n*4 ;

    tr_oR(center, Rlocal) ;
    for (pX=aX, pp=pos; pX<eX; pX+=4, pp+=2) {
        tr_ou(pp, u) ;
	tr_uu(u, u, Rlocal) ;
	tr_up(_proj_, u, pX) ;
    }

    /* Set the reduced x,y */
    for (k = 0, goon=1 ; (k<50) && goon; k++) {
        for (pX=aX, px=xy; pX<eX; pX+=4, px+=2) {
            pX[2] = px[0] - xc[0];
            pX[3] = px[1] - xc[1];
        }

        /* Compute the CD matrix elements */
        vx = cov4(aX+2, aX+2, eX) ;
        vy = cov4(aX+3, aX+3, eX) ;
        cv = cov4(aX+2, aX+3, eX) ;
        den= vx*vy -cv*cv ;

        CD[0][0] = (vy*cov4(aX+0, aX+2, eX) - cv*cov4(aX+0, aX+3, eX)) / den ;
        CD[0][1] = (vx*cov4(aX+0, aX+3, eX) - cv*cov4(aX+0, aX+2, eX)) / den ;
        CD[1][0] = (vy*cov4(aX+1, aX+2, eX) - cv*cov4(aX+1, aX+3, eX)) / den ;
        CD[1][1] = (vx*cov4(aX+1, aX+3, eX) - cv*cov4(aX+1, aX+2, eX)) / den ;

        /* Recompute xc yc */
        for (i=0; i<4; i++) Ex[i] = E4(aX+i, eX) ;
        den = minv2(CD, CD1) ;
        dxy[0] = Ex[2] - (CD1[0][0]*Ex[0] + CD1[0][1]*Ex[1]) ;
        dxy[1] = Ex[3] - (CD1[1][0]*Ex[0] + CD1[1][1]*Ex[1]) ;
	goon =  (fabs(dxy[0]) > epsilon) || (fabs(dxy[1]) > epsilon) ;
        xc[0] += dxy[0] ;
        xc[1] += dxy[1] ;
	if (verbop) {
            fprintf(stderr, "----Center=%010.6f%+010.6f (",
		center[0], center[1]) ;
            fprintf(stderr, fxy, xc[0], xc[1]) ;
            fprintf(stderr, ") %2d iterations.\n", k+1) ;
	}
    }

    /* Look in the O-C */
    mx = my = vx = vy = 0 ; i = 0;
    for (pX=aX, px=xy; pX<eX; pX+=4, px+=2) {
        pX[2] = px[0] - xc[0];
        pX[3] = px[1] - xc[1];
	dxy[0] = CD[0][0]*pX[2] + CD[0][1]*pX[3] - pX[0] ;
	dxy[1] = CD[1][0]*pX[2] + CD[1][1]*pX[3] - pX[1] ;
	mx += dxy[0] ;
	my += dxy[1] ;
	vx += dxy[0]*dxy[0] ;
	vy += dxy[1]*dxy[1] ;
	if (verbop) {
	    fprintf(stderr, "....O-C%3d= %+9.2f %+9.2f (", ++i,
	        3600.*180./M_PI*dxy[0], 3600.*180./M_PI*dxy[1]) ;
	    fprintf(stderr, fxy, px[0],px[1]) ;
	    fprintf(stderr,")\n") ;
	}
    }
    vartot = vx + vy ;		/* Total Variance */
    mx /= n; my /= n;
    vx /= n; vy /= n;
    /* fprintf(stderr, "..........= %+9.2f %+9.2f (",
	3600.*180./M_PI*mx, 3600.*180./M_PI*my) ;
    fprintf(stderr, fxy, xc[0],xc[1]) ;
    fprintf(stderr,")\n") ;
    */
    sigma = sqrt(0.5*(vx + vy));
    vx = sqrt (vx) ;
    vy = sqrt (vy) ;
    if (verbop) fprintf(stderr, "....sigma = %9.2f %9.2f %20.3f\"/pix\n",
	3600.*180./M_PI*vx, 3600.*180./M_PI*vy,
	3600.*180./M_PI*sqrt(fabs(den))) ;

    cv = 3.*sqrt((vx*vx + vy*vy)/2.) ;
    for (pX=aX, px=xy, i=k=0; pX<eX; pX+=4, px+=2) {
	dxy[0] = CD[0][0]*pX[2] + CD[0][1]*pX[3] - pX[0] ;
	dxy[1] = CD[1][0]*pX[2] + CD[1][1]*pX[3] - pX[1] ;
	i++ ;
	if ((fabs(dxy[0]) > cv) || (fabs(dxy[1]) > cv)) {
	    if (verbop) {
	        fprintf(stderr, "++++O-C%3d= %+9.2f %+9.2f (", i,
		    3600.*180./M_PI*dxy[0], 3600.*180./M_PI*dxy[1]) ;
	        fprintf(stderr, fxy, px[0], px[1]) ;
	        fprintf(stderr, ")\n") ;
	    }
	    k++;
	}
    }
    x0[0] = xc[0]; x0[1] = xc[1] ;

    free(aX) ;
    return(k) ;
}

int ccd_center(
   double *pos, double *xy, int n, 	/* IN: Positions / CCD xy */
   double center[2], double x0[2], 	/* OUT: Center definition */
   double CD[2][2])
/*++++++++++++++++
.PURPOSE  Compute all elements (CD matrix, center, x0)
.RETURNS  Status 0=OK  >0 = Number of points with problems
.REMARKS  _proj_ system is used.
-----------------*/
{
  double varmin, dr, u[3], *eX ;
  double cenmin[2], cengrid[9][2], x0min[2], x0grid[9][2], CDmin[2][2] ;
  int i, trials, imin, goon, steps, old_verbop ;

    varmin = 9.e9 ;
    old_verbop = verbop; if (verbop) --verbop;
    dr = M_PI/180./60. ;  /* Displacement of projection center = 1arcmin */

    /* As a first approximation, take the barycenter */
    eX = pos + n*2 ;
    cenmin[0] = E2p(pos+0, eX) ;
    cenmin[1] = E2 (pos+1, eX) ;
    eX = xy + n*2 ;
    x0min[0] = E2 (xy+0, eX) ;
    x0min[1] = E2 (xy+1, eX) ;

    for (steps=trials=0, goon=1; goon; steps++) {
	tr_oR(cenmin, Rlocal) ;
        /* Compute the variance for the 9 projection centers */
	cengrid[0][0] = cengrid[1][0] = cengrid[2][0] = 0 ;
	cengrid[3][0] = cengrid[4][0] = cengrid[5][0] = -dr;
	cengrid[6][0] = cengrid[7][0] = cengrid[8][0] = dr ;
	cengrid[0][1] = cengrid[3][1] = cengrid[6][1] = 0 ;
	cengrid[1][1] = cengrid[4][1] = cengrid[7][1] = -dr;
	cengrid[2][1] = cengrid[5][1] = cengrid[8][1] = dr ;
	for (i=0; i<9; i++) {	/* Convert displacement to actual center */
	    tr_pu(_proj_, cengrid[i], u) ;
	    tr_uu1(u, u, Rlocal) ;
	    tr_uo(u, cengrid[i]) ;
	    x0grid[i][0] = x0min[0];
	    x0grid[i][1] = x0min[1] ;
	}
	for (i=imin=0; i<9; i++) {	/* Test the difference centers */
	    trials++;
	    ccd_compute(pos, xy, n, cengrid[i], x0grid[i], CD);
	    if (vartot < varmin) {
		varmin = vartot ;
		imin = i ;
		memcpy(CDmin, CD, sizeof(CDmin)) ;
		fprintf(stderr,
		    ".... Trial#%3d: c=%010.6f%+010.6f (sigma=%.3f\")\n",
		    trials, cengrid[i][0], cengrid[i][1],
		    180.*3600.*sqrt(vartot/(2.*n))/M_PI) ;
	    }
	}
	memcpy(x0min,  x0grid[imin],  sizeof(x0min)) ;
	memcpy(cenmin, cengrid[imin], sizeof(cenmin)) ;
	if (imin == 0) dr /= 3 ;
	goon = dr > M_PI/180./3600.0;
    }
    center[0] = cenmin[0]; center[1] = cenmin[1];
    x0[0] = x0min[0] ; x0[1] = x0min[1] ;
    verbop = old_verbop ;
    imin = ccd_compute(pos, xy, n, center, x0, CD);
    if (imin>0) {
	verbop = 1;
	ccd_compute(pos, xy, n, center, x0, CD);
	verbop = old_verbop ;
    }
    return(imin);
}

/*==================================================================
		From (xy) and estimate of center/scale,
		and catalog entries
 *==================================================================*/

static int markOutliers (
   double *XYobs, int no,	/* IN: XY observed     	*/
   double R2,			/* IN: Squared radius of catalogued sources */
   int *permutation		/* MOD: array of matches obs->catalogued */
)
/*++++++++++++++++
.PURPOSE  Update the permutation vector: set to
	-1=normally not there (outside XYcat limits)
	-2=could be there (within XYcat limits, but outside radius)
	-3=should be there (within radius),
.RETURNS  Number of outliers (those having permutation=-1/-2)
-----------------*/
{
  double *po, x2, y2;
  int io, m=0, n3=0;

    for (io=0; io<no; io++) {
	if (permutation[io]>=0) { m++; continue; }        /* Matching */
	po = XYobs + (io<<1);	/* Catalogued source */
	x2 = po[0]*po[0];
	y2 = po[1]*po[1];
	permutation[io] = -1;	/* cataloged source normally not found */
	if ((x2>R2)||(y2>R2)) continue;
	if ((x2+y2)>R2) permutation[io] = -2;
	else       { n3++; permutation[io] = -3; }
    }
    return(no-m-n3);
}

static double getScore(
   double *XYcat, int nc,	/* IN: XY from catalog 	*/
   double *XYobs, int no,	/* IN: XY observed     	*/
   double sigma,		/* IN: acceptable rms(XYobs) */
   double score_mini,		/* IN; Score to reach -- give up if 
				   too many mismatches */
   int *permutation		/* OUT: array of permutation
				   (obs => corresponding catalogued source */
)
/*++++++++++++++++
.PURPOSE  Estimate number of matching elements, and which permutation
	   of XYcat should be used to get it.
.RETURNS  Score (about number of matches)
.REMARKS  _proj_ system is used.
-----------------*/
{
  double score = 0.;
  double *pc, *po;
  double dxy[2], sigma2, diff2, score_max;
  char used[1000];	/* Set to '1' for used catalog entry */
  int ic, io, ic_max, missing=0, maxi_missing;

    if (nc>sizeof(used)) nc = sizeof(used);
    ic_max = MIN(nc, no);	/* Maximal possible score */
    if (score_mini>0) {
	if (score_mini > ic_max) return(0);	/* Impossible! */
	maxi_missing = ic_max - score_mini -2 ; /* Always 2 matching points */
    }
    else maxi_missing = ic_max;

    memset(permutation, -1, sizeof(int)*no);
    bzero(used, sizeof(used));	/* No catalog entry yet used */
    sigma2 = sigma*sigma;

    for (po=XYobs, io=0; io<no; io++, po+=2) {
	ic_max = -1;
	score_max = 0;
	for (pc=XYcat, ic=0; ic<nc; ic++, pc+=2) {
	    if (used[ic]) continue;
	    dxy[0] = po[0] - pc[0];
	    dxy[1] = po[1] - pc[1];
	    diff2 = (dxy[0]*dxy[0] + dxy[1]*dxy[1])/sigma2;
	    /* Take a score between 0 and 1, linear with squared distance
	       - distance > 5sigmas ==> impossible, ignore the point
	       - in range [5,2.2]: linear score between 0 and 1
	       - <2.2 (sqrt(5)): score=1
	    */
	    if (diff2 > 25.) continue;
	    else if (diff2 > 5.) diff2 = 1 - (diff2-5.)/20.;
	    else diff2 = 1.;
	    if (diff2 > score_max) { score_max = diff2; ic_max = ic; }
	    if ((verbop>1) && (score>2)) fprintf(stderr,
	      "....getScore: matching obs#%-2d with cat#%-2d => %+.1f\n",
	      io+1, ic+1, diff2);
	    if (diff2 > 0.92) break;
	}
	if (score_max>0) {
	    used[ic_max] = 1;
	    permutation[io] = ic_max;
	    score += score_max;
	}
	else {
	    missing++;
	    if (missing>=maxi_missing) break;
	}
    }
#if 0
    /* test */
    fprintf(stderr, "#...getScore(obs=%d,cat=%d) missing=%d/%d, score=%.1f\n",
	    no, nc, missing,no,score);
#endif
    return(score);
}

int ccd_fromlist(
   double *pos, int nc, 	    /* IN: Positions from Catalog */
   double *xy , int no, 	    /* IN: Observations from CCD  */
   double *center, 		    /* IN: Center definition      */
   double *x0, 				/* OUT: Center origin     */
   double CD[2][2])			/* OUT: Scaling matrix    */
/*++++++++++++++++
.PURPOSE  Compute all elements (CD matrix, center, x0)
.RETURNS  Status 0=OK  >0 = Number of points with problems, -1=err
.REMARKS  _proj_ system is used.
	Take 2 catalog positions + 2 observations, assume they coincide,
	and count the number of matches.
-----------------*/
{
static int show_progress;
static char cswap[] = ",/-\\*";
static char fmt_pair[] = "\
....obs(%02d,%02d)=cat(%02d,%02d)";
static char fmt_perm[] = "\
----obs(%02d%c%02d)=cat(%02d,%02d) => scale=%.2f\"/pix score=%.1f/%.1f,\
outliers=%d\n";
  double cd[2][2], xytest[4], XYtest[4], cxy[2], u[3], dxy[2];
  double dr2, dR2, r2max, R2max, f, scale=0, scale_max=0, sigma1;
  double score, score_min, score_max;
  double xyrange[4], XYrange[4];	/* Xmin, Ymin, Xmax, Ymax */
  double *XY, *XYobs, *eX, *pp, *px1, *px2, *pX1, *pX2, *ex, *ex1, *pmax[4];
  time_t started; int delay=0, s_elapsed;
  int io, np, no1, nc1, nm, nmc, *permutation, *perm_max;
  int npairs=0, outliers=0;
  int swaps=0;
  int print_perm=0;
  int scale_ok;

    if (nc<4) {
	fprintf(stderr, "****ccd_fromlist: only %d reference stars!\n", nc);
	return(-1);
    }
    started = time(0);
    show_progress = isatty(2) ? 2 : 0;	/* Indicates the permutations */

    /* Express the catalog as local X,Y */
    memset(pmax, 0, sizeof(pmax));
    XY = (double *)malloc(sizeof(double)*(nc+no)*2);
    eX = XY + (nc*2);
    XYobs = eX;			/* xy transformed in standard coordinates */
    tr_oR(center, Rlocal);
    for (pX1=XY, pp=pos; pX1<eX; pX1+=2, pp+=2) {
        tr_ou(pp, u) ;
	tr_uu(u, u, Rlocal) ;
	tr_up(_proj_, u, pX1) ;
    }

    /* Need a permutation vector + second one to save max.score permutation */
    permutation = (int *)malloc(nm=sizeof(int)*2*no);
    memset(permutation, -1, nm);
    perm_max = permutation + no;
    fmt_perm[0] = fmt_perm[1] = fmt_perm[2] = fmt_perm[3] = '-';

    /* Set the x/y extrema */
    range2(xy, no, xyrange);
    range2(XY, nc, XYrange);

    /* Compute approximative radius (squared);
       the choice of the 2 points used for calibration
       are required to be at 1/sqrt(20) =~ 0.22 of the maximal distance
       between 2 points.
    */
    dxy[0] = xyrange[2] - xyrange[0]; dxy[1] = xyrange[3] - xyrange[1];
    r2max  = (dxy[0]*dxy[0] + dxy[1]*dxy[1])/4.;
    dxy[0] = XYrange[2] - XYrange[0]; dxy[1] = XYrange[3] - XYrange[1];
    R2max  = (dxy[0]*dxy[0] + dxy[1]*dxy[1])/4.;

    /* Choose 2 points in the observations, at sufficient distance .
       For each set of 2 points, try to match as many stars as possible.
       Test a max. number of 10 pairs.
       The score (number of matching points) must be at least half of the points
    */
    no1 = MIN(no,test_points); ex = xy + no1*2;
    nc1 = MIN(nc,test_points); eX = XY + nc1*2;
    nm  = MIN(nc, no);   /* Number of points which can be tested
                           (generally the number of observed points) */
    /* Minimal score required: take a geometrical serie */
    score_min = ascore[0];
    if (score_min < 1.) {
        for (f=1, io=nm; --io>0; f *= 0.95);
        score_min = (1.-f)/(1.-0.95);
    }
    score_max = 2.;	/* Minimal value of score_max */
    if (score_min <3.0) score_min = 3.;
    score = 0;
    /* if (verbop) */ {
	fprintf(stderr,
          "....ccd_fromlist: keep obs=%d cat=%d points; need score>=%.1f\n",
          no1, nc1, score_min);
    }
    /* Limit the scan: take between Np/3 */
    np = no1>=60 ? no1/3 : no1>40 ? no1/2 : no1-2;
    ex1 = xy + 2*np;
    fprintf(stderr, "#...testing %d primary observed points\n", np);
    /* nmc = current number of matching points */
    for (px1=xy, nmc=nm; px1<ex1; px1+=2, nmc--) {
	s_elapsed = time(0) - started;
	/* Stop if no match could be found */
	if (s_elapsed >delay) {
	    if (delay) {
		fprintf(stderr, "#+++toolong+++\n");
		break;
	    }
	    if ((s_elapsed>5) && ((px1-xy)>5)) {
		delay = 2*(time(0) - started);
	        if (verbop) fprintf(stderr,
		    "#...ccd_fromlist: delay set to %ds.\n",
		    (int)delay);
	    }
	}
	for (px2=px1+2; px2<ex; px2+=2) {
	    print_perm = show_progress;  /* Details only if a tty behind */
	    dxy[0] = px2[0]-px1[0];
	    dxy[1] = px2[1]-px1[1];
	    dr2 = dxy[0]*dxy[0] + dxy[1]*dxy[1];
	    if ((dr2/r2max) < 0.05) {	/* At least sqrt(0.05)=22% of max */
		if (verbop>1) fprintf(stderr,
		  "....ccd_fromlist: obs. #%d and #%d too close (%.1fpix)\n",
		  1+index(px1-xy)/2, 1+index(px2-xy)/2, sqrt(dr2));
		continue;
	    }
	    xytest[0] = px1[0]; xytest[1] = px1[1];
	    xytest[2] = px2[0]; xytest[3] = px2[1];
	    npairs++;			/* Number of pairs tested */
	    /*
		I've chosen a pair of observed positions, saved in xytest
		Find another pair of catalogued positions, save in XYtest
		The "swaps" loop tries the 4 combinations resulting of
		    - Echanging the 2 pairs (obs#1=cat#1 <--> obs#1=cat#2)
		    - Swapping axes x<-->y
	    */
	    for (pX1=XY; pX1<eX; pX1+=2) {
		for (pX2=pX1+2; pX2<eX; pX2+=2) {
	            if (verbop>1) fprintf(stderr, fmt_pair,
		      1+(px1-xy)/2, 1+(px2-xy)/2,
		      1+(pX1-XY)/2, 1+(pX2-XY)/2);
		    dxy[0] = pX2[0]-pX1[0];
		    dxy[1] = pX2[1]-pX1[1];
		    dR2 = dxy[0]*dxy[0] + dxy[1]*dxy[1];
		    /*
		       If we assume that the catalogued positions
		       occupies the same area as the observed positions,
		       the 2 pairs should have proportional lengths.
		       Therefeore we require dR2/R2max {about} dr2/r2max
		       (our choice is sqrt(10) ~ 3 times smaller/larger
		    */
		    f = (dR2/dr2) * (r2max/R2max);
		    if ((f<0.1) || (f>10.)) {
			if (verbop>1) fprintf(stderr,
			  " .. scale=%.1f(ratio)\n", sqrt(f));
			continue;
		    }
		    XYtest[0] = pX1[0]; XYtest[1] = pX1[1];
		    XYtest[2] = pX2[0]; XYtest[3] = pX2[1];
		    /* Compute the estimations for the 4 swaps */
		    for (swaps=0, scale_ok=1; (swaps<4)&&scale_ok; swaps++) {
		        scale =  solve2(XYtest, xytest, swaps, cxy, cd);
		        if ((verbop>1) && (swaps == 0)) fprintf(stderr,
		            " => scale=%.2f\"/pix", scale*(180.*3600./M_PI));
			if (ascale[0] /* Restriction on possible scales */
			  && ((scale<ascale[0]) || (scale>ascale[1]))) {
			    scale_ok=0;
			    if (verbop>1) fprintf(stderr, " (impossible)\n");
			    continue;  /* Will stop this loop since scale_ok=0*/
		        }
		        /*computeXY(XYobs, px1, nmc, cxy, cd);*/
		        computeXY(XYobs, xy, no, cxy, cd);
		        /* sigma1 represents the acceptable rms.
			   If not given in arguments,
			   take a pixel size + 20%
		        */
			sigma1 = asigma[1];
			if (sigma1 <= 0) sigma1 = 1.2*scale;
			if (verbop>1) fprintf(stderr, "\n");
		        score = getScore(XY, nc /*nm*/, XYobs, no,
			        sigma1, score_min, permutation);
		        if (verbop>1) fprintf(stderr, "----getScore: %.1f%c",
			  score, cswap[swaps]);
		        if (score>score_max) {
			    if (verbop>1) fprintf(stderr, "\b*");
		            else if (verbop) print_perm = 1;
			    score_max = score; scale_max = scale;
			    /* Save the best match */
			    outliers = markOutliers(XYobs, nmc, R2max,
				    permutation);
			    memcpy(perm_max, permutation, sizeof(int)*no);
			    pmax[0] = pX1; pmax[1] = pX2;
			    pmax[2] = px1; pmax[3] = px2;
			    if (delay) started = time(0);
		        }
			if (print_perm&1) fprintf(stderr, fmt_perm,
		              1+(px1-xy)/2, cswap[swaps], 1+(px2-xy)/2,
			      1+(pX1-XY)/2, 1+(pX2-XY)/2,
		              (180.*3600./M_PI)*scale, score, score_min,
			      outliers);
			else if (print_perm&2) { int len;
			    len = fprintf(stderr, fmt_pair,
		              1+(px1-xy)/2, 1+(px2-xy)/2,
		              1+(pX1-XY)/2, 1+(pX2-XY)/2);
			    while (--len>=0) fputc('\b', stderr);
			}
			print_perm = 0;
		        if (score>=score_min) break;
		    }
		    if (score>=score_min) break;
		}
		if (score>=score_min) break;
	    }
	    if (score>=score_min) break;
	}
	if (score>=score_min) break;
    }

    /* When not a single pair found, it's impossible ! */
    if (npairs == 0) {
	fprintf(stderr,
 	  "****Not a single pair with distance>%.1fpix could be found\n",
	  sqrt(0.05*r2max));
	fprintf(stderr, "    (%s)\n",
	  "maybe the (x,y) list was not sorted by increasing magnitude?");
	return(-1);
    }

    /* Try in all cases to improve the found solution:
       in "perm_max", we find the correct order
    */
    /* Write out the points */
    if ((swaps==4) || (score_max < score_min)) 	/* No satisfactory solution */
         fmt_perm[0] = score_max <=0.5*score_min ? '*' :'+';
    else fmt_perm[0] = '=';
    fmt_perm[1] = fmt_perm[2] = fmt_perm[3] = fmt_perm[0];
    pX1 = pmax[0]; pX2 = pmax[1];  /* Catalogued first pair */
    px1 = pmax[2]; px2 = pmax[3];  /* Corresponding observed pair */
    fprintf(stderr, fmt_perm, 1+(px1-xy)/2, cswap[swaps], 1+(px2-xy)/2,
			      1+(pX1-XY)/2, 1+(pX2-XY)/2,
		              (180.*3600./M_PI)*scale_max,
			      score_max, score_min, outliers);
    if (verbop) {
	fprintf(stderr, "#===List of Matching points:\n"); fflush(stderr);
    }
    /*for (io=0; io<nmax; io++, px1+=2) */
    for (io=0, pX1=XY, pX2=XYobs, px1=xy; io<nm; io++, px1+=2) {
	if (perm_max[io] < 0) continue;
	np = perm_max[io]*2;
	pp = pos+np;
	pX1[0] = pp[0];
	pX1[1] = pp[1];
	pX2[0] = px1[0];
	pX2[1] = px1[1];
	if (verbop) {
	    if ((io < no1) || verbop>1) fprintf(stderr,
	      " %09.5f %+09.5f = %9.2f %9.2f ; obs#%02d=cat#%02d\n",
	      pp[0], pp[1], px1[0], px1[1], io+1, perm_max[io]+1);
	}
	pX1 += 2; pX2 += 2;
    }
    np = (pX1-XY)/2;

    memcpy(x0, cxy, sizeof(double)*2);
    memcpy(CD, cd, sizeof(cd));
    if ((np<6) || (fmt_perm[0] == '*')) {
	fprintf(stderr, "++++Only %d points, or score not reached\n", np);
	sigma = -1.;
	np = 0;
    }
    else {
        pX1 = pmax[0]; /* Catalogued first pair */
        px1 = pmax[2]; /* Corresponding observed pair */
	/*np = ccd_center(pX1, px1, nmax, center, x0, CD);*/
	np = ccd_center(XY, XYobs, np, center, x0, CD); 
	/* It may happen (in crowded regions) that the apparent
	   large score leads to an erroneous result.
	   Give suggestion to solve this problem: require that
	   the sigma is not large than some predefined amount.
	*/
	scale = sqrt(fabs(CD[0][0]*CD[1][1]-CD[0][1]*CD[1][0]));
	if (((ascale[0]) && (scale<ascale[0])) || (scale>ascale[1])) {
	    fprintf(stderr, "++++Scale changed !\n");
	    fprintf(stderr, "    If necessary, rerun with option\n");
	    fprintf(stderr, "       sigma=%.2f    (required sigma in \"/pix)\n",
		0.5*3600.*180./M_PI*sigma);
	    fprintf(stderr, "    to force a better adjustment\n");
	}
    }


    free(XY);
    free(permutation);

    return(np);
}


/*==================================================================
		Program
 *==================================================================*/
#ifdef TEST

static int optp   = 0 ;			/* -p option */
static int  omag = -1;			/* Word with magnitude 	*/
static float mag_min = 0;		/* Minimal magnitude to look at */
static int colsep;

static double CD[2][2] ;		/* (x,y) = CD * (pixels-pix0)	*/
static double pix0[2] ;			/* Pixel position of center	*/
static double  Rlocal[3][3] ;		/* Rotation Matrix XY-->RA,Dec	*/

#define iscomment(c)	((c=='#')||(c=='%'))
#define isasep(c)	((c=='\t')||(c==';')||(c== '|')||(c==','))

static char *param[] = {		/* Table of parameters + values */
   "refcat=", (char *)0,		/* [0] */
   "refmag=", (char *)0,		/* [2] */
   "-c=", (char *)0,			/* [4] */
   "scale=", (char *)0,			/* [6] */
   "-out=", (char *)0,			/* [8] */
   "-out.max=", (char *)0,		/*[10] */
   "score=", (char *)0,			/*[12] */
   "sigma=", (char *)0,			/*[14] */
   (char *)0, (char *)0,
   (char *)0, (char *)0,
   (char *)0, (char *)0,
   (char *)0, (char *)0,
   (char *)0, (char *)0,
   (char *)0, (char *)0,
   (char *)0, (char *)0,
   (char *)0, (char *)0,
   (char *)0, (char *)0,
   (char *)0, (char *)0,
   (char *)0, (char *)0,
   (char *)(-1)				/* Marker END */
};

static char usage[] = "\
Usage: ccd [-v] [-p] [-f reference_file ] [-t#] [refcat=...  refmag=... ]\n\
       [scale=min,max] [score=min] [sigma=min,max] [-c=...] [-c.r[b]=...]\n\
       [-mag[=2]] [asu-constraints...]\
" ;

static char help[] = "\
Find the position of sources defined by (x,y) pairs. 2 modes of reduction:\n\
\n\
1. From a list of reference sources, i.e. sources having RA, Dec, X and Y.\n\
   This is the default mode. The reference stars are specified via a list:\n\
   RA Dec = x y ; comments\n\
   Lines missing the '=' sign are assumed to contain only X and Y -- \n\
   the positions are computed for these lines.\n\
\n\
2. From a list of positions taken or retrieved from some reference catalog\n\
   (refcat must be specified). The list of (RA,Dec) is either taken from a\n\
   catalog (e.g. refcat=I/289 for UCAC2), or from the input if refcat is\n\
   not followed by the = sign. The list of (x,y) pairs is given in the\n\
   input file (-f argument, default=stdin); this list is separated from the\n\
   (RA,Dec) list by a blank line. A header line may precede the (x,y) list,\n\
   to specify the columns (X Y mag and Vmag, Rmag, etc are recognized). \n\
   In this option:\n\
   - the (x,y) pairs are assumed to be ordered by increasing magnitude;\n\
     (only when the header line is not given in the input)\n\
   - refmag specifies which magnitude to use in refcat (leftmost mag)\n\
   - the center of the field (-c=) is required.\n\
   - the optional scale or scale range is in arcsec/pix.\n\
   - the optional score gives the minimal score expected\n\
   - the optional sigma gives the maximal sigma expected, in arcsec/pix\n\
   - t specifies the number of reference positions and (x,y) pairs to examine\n\
   - the other options represent ASU constraints to specify the radius (-c.r)\n\
     or box (-c.b), the equinox (-c.eq=), the max. number of reference stars\n\
     (-out.max), or catalog-specific constraints (otype='*')\n\
\n\
Options:\n\
  -v: verbose option (details of tentative identifications)\n\
  -f: the list of corresponding sources is in a file (default stdin).\n\
  -p: print out the correspondances ( RA Dec = x y)\n\
  -t: maximal number of points to use in the test phase (default 24)\n\
  -mag: specifies the column with a magnitude, if any (separator=semi-colon)\n\
\n\
If the -f option is absent, the list of reference positions is given in the\n\
standard input, and is terminated by a blank line. Lines following this blank\n\
line are assumed to contain just (x,y) pairs for which positions are computed\n\
\n\
Example of a full program to compare a table with the UCAC2 (I/289) catalog:\n\
\n\
tabmap -ascii -data -map '${Xpos} ${Ypos} ; ${Vmag}' table.dat \\\n\
| sort -t\\; +1n | ccd refcat=I/289 -c=236.51433-37.78614 \n\
";

static char **dic_lookup(char **table, char *param)
/*++++++++++++++++
.PURPOSE  Find in dictionary a parameter
.RETURNS  location in table / NULL
.REMARKS  param must NOT contain the = sign !! (Mod. V1.56)
-----------------*/
{
  char **a, *p; int len, c=0;
    p = strchr(param, '=');
    /* if (!p) return((char **)0); */
    if (p) len = 1 + (p-param);
    else { len = strlen(param); c = '='; }
    for (a=table; *a; a += 2) {
	if (strncmp(*a, param, len) == 0) {
	    if (!c) return(a);
	    if ((*a)[len] == c) return(a);
	}
	if (*a == (char *)(-1)) break;
    }
    return((char **)0);
}

static char **dic_insert(char **table, char *text)
/*++++++++++++++++
.PURPOSE  Add the value in dictionary
.RETURNS  Where it was added / NULL
.REMARKS  param MUST contain the = sign !!
-----------------*/
{
  char **a, *p; int len;
    /* No = sign ==> ignored */
    p = strchr(text, '=');
    if (!p) return((char **)0);
    len =  1 + (p-text);
    if ((a = dic_lookup(table, text))) {	/* Replace existing parameter */
	if (a[1]) {
	    free(a[1]);
	    fprintf(stderr, "++++Replace symbol: %s%s\n", a[0], a[1]);
	}
	a[1] = (char *)0;
    }
    else {				/* Find a free slot */
	for (a=table; *a; a += 2) ;
	if (*a == (char *)(-1)) {
	    fprintf(stderr, "++++Too many symbols, ignore: %s\n", text);
	    return((char **)0);
	}
	a[0] = strdup(text);
	a[0][len] = 0;
    }
    a[1] = strdup(p+1);
    return(a);
}

static double get_mag(char *text)
/*++++++++++++++++
.PURPOSE  Find the magnitude
.RETURNS  Value of magnitude
-----------------*/
{
  /*double  mag; */
  int j;
    while (*text == ' ') text++;
    j = omag;
    while (--j>=0) {
	if (ispunct(colsep))  {			/* Skip field */
	    while (*text && (*text != colsep)) text++;
	    if (*text) text++;
	}
	else 				/* Blank separator: skip field */
	    while (isgraph(*text) && (*text != colsep)) text++;
        while (*text == ' ') text++;
    }
    return(atof(text));
}

static int cmp_mag(char **a, char **b)
/*++++++++++++++++
.PURPOSE  Compare 2 magnitudes
.RETURNS  Value of magnitude
-----------------*/
{
  double d;
    d = get_mag(*a) - get_mag(*b);
    if (d<0) return(-1);
    else if (d>0) return(1);
    else return(0);
}

static void xy2pos(double xy[2], double o[2])
/*++++++++++++++++
.PURPOSE  Compute the Position from the x,y
.RETURNS  ---
-----------------*/
{
  double  XY[2], u[3], pix[2] ;
    pix[0] = xy[0] - pix0[0] ;
    pix[1] = xy[1] - pix0[1] ;
    XY[0] = CD[0][0] * pix[0] + CD[0][1] * pix[1] ;
    XY[1] = CD[1][0] * pix[0] + CD[1][1] * pix[1] ;
    tr_pu(_proj_, XY, u) ;
    tr_uu1(u, u, Rlocal) ;
    tr_uo(u, o) ;
}

static int isDataLine(char *text)
/*++++++++++++++++
.PURPOSE  Check a text, and decide it's data
.RETURNS  1<<(numbers-1): 1=1num, 2=2num, 4=3num, 8=4num, etc
.REMARKS  Yes if only digits +-/ : .
-----------------*/
{
  char *p; int nd=0, id=0, sep=1;
    if (iscomment(text[0])) return(0);
    for (p=text; *p; p++) {
	if (isdigit(*p)) { if(sep) id++, sep=0; nd|=id; continue; }
	if (isspace(*p)) { sep=1; continue; }
	if (*p == '-')   { sep=1; continue; }
	if (*p == '+')   { sep=1; continue; }
	if (*p == '.') continue;
	if (*p == ';') break;
	if (*p == ':') continue;
    }
    return(nd&2);	/* At least 2 numbers */
}

static int isHeadLine(char *text, int *pow)
/*++++++++++++++++
.PURPOSE  Check column headers -- must find xy column(s)
.RETURNS  Separator between columns (0 when can't be a header)
.REMARKS  x|y = (xy) position.
	When one column contains contains a magnitude, its value is
	stored in omag.
-----------------*/
{
  static char photo[] = "UBVRIJHKPpuvbygrizm" ;
  char *p, sep=0;
  int owx = -1;
  int oph = -1;		/* Photometric column */
  int ow  = 0;

    p = text; if (iscomment(*p)) p++;
    while (*p == ' ') p++;
    while (*p) {
	if (sep) {
	    while (*p == ' ') p++;
	    if (*p == sep) { p++; ow++; }
	}
	else if (*p == ' ') ow++;
	while (*p == ' ')   p++;
	if (isalpha(*p)) {	/* Could be a column header */
	    if ((owx<0) && (tolower(*p) == 'x')) {
		owx = ow;
	    }
	    else if ((oph<0) && strchr(photo, *p)) {
		/* could be a magnitude ? */
		if (*p != 'm') p++;
		if ((*p == '\'') || (*p == '*')) p++;
		if (strncasecmp(p, "mag", 3) == 0) p += 3;
		if ((isasep(*p)) || isspace(*p) || isdigit(*p))
		    oph = ow;
	    }
	    while (isgraph(*p) && (!(isasep(*p)))) p++;
	}
	else if (isasep(*p)) {
	    if (sep == 0) { sep = *p; ow = 0; }
	}
	else break;		/* Not a letter, not a separator */
    }
    if (owx>=0) {
	*pow = owx;
	if (!sep) sep = ' ';
	if (omag<0) omag = oph;	/* When a magnitude could be found */
    }
    return(sep);
}

static int fget_xy(FILE *fcat, int sep, int oxy, /* V1.53: sep & oxy added */
	double *xy, char **line, int omax)
/*++++++++++++++++
.PURPOSE  Read the file containing (x,y) values
.RETURNS  Number of (xy) pairs. Separator is sep; word position of xy is oxy
.REMARKS  If line[0] is not null, first line already read.
	An empty line indicates the end of the list.
-----------------*/
{
  char buf[BUFSIZ];
  char *p, *o;
  double *adouble;
  int no = 0, ow;
  int i, j, len, n;

    adouble = xy;
    /* Read First */
    if (line[0]) no = 1;

    /* Read input as character strings; interpret later */
    while((no < omax) && fgets(buf, sizeof(buf), fcat)) {
        /* Remove trailing blanks */
        for (p = buf+strlen(buf)-1; (p>=buf) && isspace(*p); --p) *p=0;
        if (!buf[0]) break;		/* Empty line */
	if ((no<1) && (oxy<0) && iscomment(buf[0])) {
	    sep = isHeadLine(buf+1, &ow);
	    if (sep) oxy = ow;
	}
        if (isDataLine(buf)) {
	    if (line[no]) free(line[no]);	/* V1.53 */
            line[no++] = strdup(buf);
	}
	else if ((no<1) && (oxy<0) && (i = isHeadLine(buf, &ow))) {
	    sep = i;
	    oxy = ow;
	}
    }

    /* Interpret the (x,y) lines */
    /* Default separator is ; */
    if (sep == 0) {
	sep = ';';
	oxy = 0;
    }
    mag_min = 0;
    if ((omag>0) && (omag!=oxy) && sep) {	/* SORT */
      double mag1=-27., mag;
        colsep = sep;
        for (i=0; i<no; i++) {
	    mag = get_mag(line[i]) ;
	    /* Keep in mag_min the brightest non-null magnitude */
	    if ((mag_min == 0) && (mag == mag)) mag_min = mag;
	    if (mag >= mag1) { mag1 = mag; continue; }
	    if (mag == 0) continue;
	    fprintf(stderr,
		    "++++Input (xy) list not sorted by magnitude, do it!\n");
	    qsort(line, no, sizeof(char *), (SORT_FCT)cmp_mag);
	    break;
	}
    }

    for (i=0; i<no; i++) {
        /* if(verbop) fprintf(stderr, "#...%s\n", line[i]); */
	p = line[i];
	if (oxy>0) {	/* Move to the required word */
	    for (j=oxy; j>0; --j) {
		if (sep == ' ') {
		    while (*p == ' ') p++;
		    while (isgraph(*p)) p++;
		}
		else {
		    while (*p && (*p != sep)) p++;
		    if (*p) p++;
		}
	    }
	}
	/* Compute the length of the 2 numbers vector */
	o = p;
	if (sep == ' ') {
	    while (*p == ' ') p++;
	    while (isgraph(*p)) p++;
	    while (*p == ' ') p++;
	    while (isgraph(*p)) p++;
	}
	else
	    p = strchr(p, sep);
	len = p ? (p-o) : strlen(o);
	n = tr_af1(o, len, adouble, 2);
        /* if (p = strchr(line[i], ';')) *p = 0;
        n = tr_af1(line[i], strlen(line[i]), adouble, 2);
        if (p) *p = ';';
	*/
        if (n != 2) {
    	    fprintf(stderr, "++++Bad (x,y), ignored: %s\n", line[i]);
    	    free(line[i]);
    	    no--;
    	    for (n=i; n<no; n++) line[n] = line[n+1];
    	    i--;
        }
        else adouble += 2;
    }

    return(no);
}

int main (int argc, char **argv)
/*++++++++++++++++
.PURPOSE  Find the elements of CD matrix
.REMARKS  Assume always a TAN projection
-----------------*/
{
  static char fmt_x0[] = " x0=%.2f";	/* Added V1.54 */
  double center[2], apos[MAX_REFSTARS*2], xy[MAX_REFSTARS*2];
  double xyrange[4], scale, radius, sigmasec, scalesec;
  char buf[BUFSIZ], acoo[32], *line[MAX_REFSTARS], *linxy[MAX_REFSTARS];
  int i, n=0, nc=0, no=0, nxy=0, hcenter, badrefs, coprec;
  char *p, *pv, *status, *source, **a, *tmpfile;
  double *adouble, *edouble;
  int oxy = -1, ow;
  char sep = 0;
  FILE *fcat, *fxy;

    memset(line, 0, sizeof(line));
    memset(linxy, 0, sizeof(linxy));

    /* Process the options */
    fcat = stdin;
    fxy  = (FILE *)0;
    while ((argc>1) && (argv[1][0] == '-')) {
	p = *++argv; --argc;
	if (*p == '-') switch(p[1]) {
	  case 'v':	/* Verbose */
	    if (strncmp(*argv, "-ver", 4) == 0) {
	         printf("ccd  (CDS) Version %s\n", VERSION);
	         exit(0);
	    }
	    /* Accept -v2 or -vv for verbosity level */
	    if (isdigit(p[2])) verbop = atoi(p+2);
	    else for (i=1; p[i] == 'v'; i++) verbop++;
	    continue;
	  case 'h':	/* Help */
	    printf("%s\n\n%s", usage, help);
	    exit(0);
	  case 'f':	/* Input file with references */
	    if (p[2]) p += 2;
	    else { p = *++argv; --argc; }
	    fcat = fopen(p, "r");
	    if (!fcat) {
		fprintf(stderr, "****");
		perror(p);
		exit(1);
	    }
	    continue;
	  case 'p':	/* Print the positions   */
	    optp = 1;
	    continue;
	  case 't':	/* Number of test points */
	    if (p[2]) p += 2;
	    else { p = *++argv; --argc; }
	    test_points = atoi(p);
	    if (test_points < 6) {
		fprintf(stderr, "++++ccd: test points must be >= 6\n");
		test_points = 6;
	    }
	    continue;
	  case 'm':	/* Magnitude */
	    if (strncmp(p, "-mag", 4) == 0) {
		omag = 1;
		if (p[4] == '=') omag = atoi(p+5)-1;
		continue;
	    }
	    /* NO BREAK (error) */
	  default:
	    if (dic_lookup(param, p)) {
		dic_insert(param, p);
		continue;
	    }
	    fprintf(stderr, "****Can't interpret option: %s\n", p);
	    exit(1);
	}
    }

    if (argv[1] && strstr(argv[1], "cat")) {	/* Use ccd_fromlist */
	/* Find details from the argument list */
	if (strchr(argv[1], '='))
	  dic_insert(param, argv[1]);
	++argv; --argc;
	while (argc>1) {
	    p = *++argv; --argc;
	    if (dic_insert(param, p));
	    else {	/* Assume a file name, if last argument  */
	        if (argc == 1) fxy = fopen(p, "r");
		if (!fxy) fprintf(stderr,
	      	    "++++Argument not taken into account: %s\n", p);
	        if (fcat == stdin) fcat = fxy;
	    }
	}
	/*
	   Read the parameters (center etc) and catalogued positions
	   (catalogued positions only if source not specified,
	   i.e. source is NULL)
	*/
	source = param[1];
	line[0] = (char *)0;
	while (fgets(buf, sizeof(buf), fcat)) {
	    if (iscomment(buf[0])) {	/* Comment */
		if (oxy<0) sep = isHeadLine(buf+1, &oxy) ;
	    }
	    /* Remove trailing blanks */
	    for (p = buf+strlen(buf)-1; (p>=buf) && isspace(*p); --p) *p=0;
	    if (!buf[0]) break;			/* Empty line */
	    if (isDataLine(buf)) 		/* (x,y) */
		break;

	    /* Ignore line made of a single character like -----  */
	    for (i=1; buf[i] == buf[0]; i++) ;
	    if (!buf[i]) continue;
	    if (iscomment(buf[0])) continue;
	    a = dic_insert(param, buf+i);
	    if ((!a) && (i = isHeadLine(buf, &ow))) {
		/* Is a header line.... */
		sep = i;
		oxy = ow;
	    }
	    else fprintf(stderr, "++++Argument ignored: %s\n", buf);
	}
	if (verbop && sep) {
	    fprintf(stderr, "----XY-column in input stream is word #%d", oxy+1);
	    if (sep != ' ') fprintf(stderr, " (separator=%c)", sep);
	    fprintf(stderr, "\n");
	}

	/* Interpret the Center */
	if (!param[5]) {
	    fprintf(stderr,
	      "****ccd: center not specified!\n%s\n", usage);
	      exit(2);
	}
	hcenter = tr_aq(param[5], strlen(param[5]), center);

	/* The catalog to look may be defined via a -source */
	source = param[1];
	if (!source) {
	    a = dic_lookup(param, "-source=");
	    if (a) source = a[1];
	}

	/* Get the reference positions . Default radius is 6' */
	if (source) { 		/* Get position from a Query  */
	    if (buf[0]) {	/* Get first all (x,y) pairs  */
	        if (isDataLine(buf)) line[0] = strdup(buf);
	        no = fget_xy(fcat, sep, oxy, xy, line, MAX_REFSTARS);
	        if (fcat != stdin) fclose(fcat);
	    }
	    sprintf(buf, "/tmp/ccd.%d", getpid());
	    tmpfile = strdup(buf);
	    if ((fcat != stdin) && (fcat != fxy)) fclose(fcat);
	    fcat = fopen(tmpfile, "w");
	    fprintf(fcat, "-oc=deg\n-out.max=%s\n-source=%s\n",
	      param[11] ? param[11] : "100", source);
	    if (param[5]) fprintf(fcat, "-c=%s\n", param[5]);
	    fprintf(fcat, "-sort=%s\n", param[3]? param[3] : "*PHOT*");
	    /* Ignore too bright objects */
	    if ((omag>0) && (mag_min>2)) fprintf(fcat, "%s=>%g\n",
		    param[3]? param[3] : "*PHOT*", mag_min-1.);
	    /* V1.52: non-NULL magnitude */
	    else if (param[3] && param[3][0] != '*'
		&& (!dic_lookup(param, param[3]))) /* Argument not used */
		fprintf(fcat, "%s=0.1..25\n", param[3]);
	    fprintf(fcat, "-out=%s\n", param[9] ? param[9] :
			"_RAJ,_DEJ, *PHOT*");
	    /* Get the estimated scale */
	    ascale[0] = ascale[1] = 0;
	    if (param[7]) {
		i = tr_af1(param[7], strlen(param[7]), ascale, 2);
		if (i < 1) {
		    fprintf(stderr, "****Bad scale: %s\n", param[7]);
		    exit(1);
		}
		if (i == 1) {	/* Limits in scale: +/-25% */
		    ascale[1] = ascale[0] * 1.25;
		    ascale[0] *= 0.80 ;
		}
		/* Convert to radians/pixel */
		ascale[0] *= M_PI/(180.*3600.);
		ascale[1] *= M_PI/(180.*3600.);
	    }
	    /* Get the radius. If not specified, use circle of 6'. */
	    if (dic_lookup(param, "-c.r") || dic_lookup(param, "-c.b")) ;
	    else if (param[7] && (no>0)) {
		scale = sqrt(ascale[0]*ascale[1]);
		range2(xy, no, xyrange);
		xyrange[0] = xyrange[2] - xyrange[0];	/* Interval */
		xyrange[1] = xyrange[3] - xyrange[1];
		radius = 0.5 * scale			/* Radius in radians */
		       * sqrt(xyrange[0]*xyrange[0] + xyrange[1]*xyrange[1]);
		radius *= (180.*60.)/M_PI; 	/* Convert to arcmin */
		fprintf(fcat, "-c.rm=%.3f\n", radius);
	    }
	    else fprintf(fcat, "-c.rm=%d\n", 6);
	    /* Scores & Sigmas */
	    ascore[0] = ascore[1] = 0;
	    a = dic_lookup(param, "score=");
	    if (a[1]) {
		i = tr_af1(a[1], strlen(a[1]), ascore, 2);
		if (i < 1) {
		    fprintf(stderr, "****Bad %s %s\n", a[0], a[1]);
		    exit(1);
		}
		if (i == 1) ascore[1] = 10*ascore[0];
	    }
	    asigma[0] = asigma[1] = 0;
	    a = dic_lookup(param, "sigma=");
	    if (a[1]) {
		i = tr_af1(a[1], strlen(a[1]), asigma, 2);
		if (i < 1) {
		    fprintf(stderr, "****Bad %s %s\n", a[0], a[1]);
		    exit(1);
		}
		if (i == 1) { asigma[1] = asigma[0]; asigma[0] = 0; }
		/* sigma must be expressed in rad/pix */
		asigma[0] *= M_PI/(180.*3600.);
		asigma[1] *= M_PI/(180.*3600.);
	    }

	    /* Other criteria */
	    for (a=param+16; *a; a += 2) {
		if (*a == (char *)(-1)) break;
		fprintf(fcat, "%s%s\n", a[0], a[1]);
	    }
	    fclose(fcat);
	    sprintf(buf, "asu < %s", tmpfile);
	    if (verbop) { char abuf[80];
		fprintf(stderr, "----Launch %s:\n", buf);
		fcat = fopen(tmpfile, "r");
		while (fgets(abuf, sizeof(abuf), fcat))
		  fprintf(stderr, "%s", abuf);
		fclose(fcat);
		fprintf(stderr, "#\n#---Reading reference stars:\n");
	    }
	    fcat = popen(buf, "r");
	    if (!fcat) {
		fprintf(stderr, "****Launch ");
		perror(buf);
		exit(1);
	    }
	    /* Skip the headings until data */
	    while (fgets(buf, sizeof(buf), fcat)) {
		if(verbop>1) fprintf(stderr, "#...%s", buf);
		if (buf[0] != '-') continue;
		for (i=0; (buf[i]=='-')||(buf[i]=='\t'); i++) ;
		if (iscntrl(buf[i])) break;
	    }
	    /* Here the line is ---\t---\t */
	    if (!fgets(buf, sizeof(buf), fcat)) fprintf(stderr,
		    "#***No data found after header/data separator.\n");
	    if(verbop) fprintf(stderr, "#...%s", buf);
	}
	else {
	    tmpfile = (char *)0;
	}

	/* Read the reference positions (from launched script, or input) */
	adouble = apos;
	edouble = &apos[ITEMS(apos)-2];
	nc = 0;
	while(1) {
	    /* Remove trailing blanks */
	    for (p = buf+strlen(buf)-1; (p>=buf) && isspace(*p); --p) *p=0;
	    if (!buf[0]) break;		/* Empty line */
	    if (iscomment(buf[0])) ;	/* Comment */
	    else {			/* --- Interpret Catalogued Position */
	        if ((p = strchr(buf, ';'))) *p = 0;
	        if ((p = strchr(buf, '\t'))) {	/* Keep only left 2 cols */
		    *p = ' ';
		    if ((p = strchr(buf, '\t'))) *p = 0;
	        }
	        i = tr_aq(buf, strlen(buf), adouble);
	        if (i < 3)
		    fprintf(stderr, "++++Bad position, ignored: %s\n", buf);
	        else if (adouble >= edouble) fprintf(stderr,
		    "++++Too many positions, ignored: %s\n", buf);
		else {
		    nc++, adouble += 2;
	    	    /* if(verbop) fprintf(stderr,
		      "#...catalogued pos: %s\n", buf);
		    */
		}
	    }
	    if (!fgets(buf, sizeof(buf), fcat)) break;
	    if(verbop) fprintf(stderr, "#...%s", buf);
	}

	/* All catalogued data read. Close files */
	if (tmpfile) {
	    unlink(tmpfile);
	    free(tmpfile);
	}
	if (fcat != stdin) fclose(fcat);
	if (nc < 4) {
	    fprintf(stderr, "++++Only %d reference stars, can't work\n", nc);
	    exit(1);
	}

	/* File with x,y */
	if (!fxy) fxy = stdin;

	adouble = xy;
	edouble = &xy[ITEMS(xy)-2];
	/* Read up to MAX (x,y) pairs from standard input */
	if (no == 0) {
	    no = fget_xy(fxy, sep, oxy, xy, line, MAX_REFSTARS);
	}
	if(verbop) fprintf(stderr,
	  "#...catalogued pos=%-3d  (x,y) pairs=%d\n", nc, no);
	badrefs  = ccd_fromlist(apos, nc, xy, no, center, pix0, CD);
	if (badrefs<0) {
	    fprintf(stderr, "****Problem in ccd_fromlist, terminated\n");
	    exit(1);
	}
	sigmasec = sigma*180.*3600./M_PI;
	scalesec = sqrt(fabs(CD[0][0]*CD[1][1]-CD[0][1]*CD[1][0]))
	         *180.*3600./M_PI;
	if (sigma<=0) {
	    coprec = 4;
	    printf("#***Matching %3d observed with %3d catalogued FAILED***\n",
		    no, nc);
	}
	else {
	    printf("#===Matching %3d observed with %3d catalogued:\n", no, nc);
	    printf("#   scale=%garcsec/pix, sigma=%.2garcsec\n",
		    scalesec, sigmasec);
	    for (coprec=3, scale=sigmasec; scale<90.; coprec++)
	        scale *= 10.;
	    /* Number of decimals for edition of x0 and y0 (V1.54)
	     * depends on scale (precision up to 0.1arcsec)
	     */
	    fmt_x0[1] = 'x';
	    p = strchr(fmt_x0, '='); p[3] = '1';
	    for (scale=scalesec; scale>0.1; scale /= 10.)
		p[3] += 1;
	}
        printf(" c=%09.5f%+09.5f precoo=%d Eq=J2000",
		 center[0], center[1], coprec);
	if ((p = getenv("Xpos"))) printf(" x=%s", p);
	if ((p = getenv("Ypos"))) printf(" y=%s", p);
        printf(" \\\n error=%.2farcsec", sigmasec);
	printf(fmt_x0, pix0[0]); fmt_x0[1] = 'y';
	printf(fmt_x0, pix0[1]);
	printf(" CD=\\\n%.7g,%.7g,%.7g,%.7g}{ Position computed \\\n",
	    CD[0][0], CD[0][1], CD[1][0], CD[1][1]) ;
	printf("from a comparison with catalog %s", param[1]);
	printf(" ({sigma}=%.2f'')}\n", sigmasec);
	if (badrefs) printf("#***There are %d bad reference stars\n", badrefs);

	/* Now edit the result as RA, Dec = (original data) */
	if (optp) {
	    printf("\n\n#Positions of each point:\n");
	    for (i=0, adouble = xy; i<no; i++, adouble += 2) {
	        xy2pos(adouble, center);
	        ed_q(acoo, center, coprec);
	        printf(" %s = %s\n", acoo, line[i]);
	    }
	}
	if (isatty(0) && isatty(1) && optp) {
	    printf("#...(Hit Return to transform the input) ");
	    getchar();
	}
	adouble = xy;
	while (fgets(buf, sizeof(buf), stdin)) {
	    if (!optp) continue;
	    /* Remove trailing blanks */
	    for (p = buf+strlen(buf)-1; (p>=buf) && isspace(*p); --p) *p=0;
	    if (isDataLine(buf)) {
	        if ((p = strchr(buf, ';'))) *p = 0;
	        n = tr_af(buf, strlen(buf), adouble);
	        if (p) *p = ';';
	    }
	    else n = 0;
	    if (n != 2) { printf("%s\n", buf); continue; }
	    xy2pos(adouble, center);
	    ed_q(acoo, center, coprec);
	    printf(" %s = %s\n", acoo, buf);
	}
	exit(0);
    }
    status = buf ;
    while(status) {
        /* Get the RA Dec Position */
        if (isatty(0)) {
            printf("----Give the reference positions as  RA  Dec  =  x  y\n");
	    printf("    (terminate with a blank line)\n") ;
        }
        for (n=0; n<MAX_REFSTARS; ) {
	    line[n] = (char *)0;
            if (isatty(0)) printf("%4d> ", n+1) ;
	    status = fgets(buf, sizeof(buf), stdin);
	    if (!status) break ;
	    /* Remove trailing blanks */
	    for (p = buf+strlen(buf)-1; (p>=buf) && isspace(*p); --p) *p=0;
            if (!buf[0]) break ;	/* Blank line ! */
	    for (p=buf; isspace(*p); p++) ;
	    if (p[0] == '#') continue ;
	    if (p[0] == '%') continue ;
	    if (line[n]) free(line[n]);	/* Was already allocated */
	    line[n] = strdup(buf);
	    if ((p = strchr(buf, '='))) {
	        i = tr_aq(buf, p-buf, apos+n*2) ;
	        if (i<=0) fprintf(stderr,
	            "++++Line ignored (no coo): %s\n", buf) ;
	        else {
	            p++ ;
		    if ((pv = strchr(p, ';'))) *pv = 0;
		    i = tr_af1(p, strlen(p), xy+n*2, 2) ;
		    if (i != 2) fprintf(stderr,
		        "++++Line ignored (no x,y): %s\n", buf) ;
		    else n++ ;
		    if (pv) *pv = ';';
	        }
	    }
	    else {
		linxy[nxy++] = line[n];	/* It's an observation! */
		line[n] = (char *)0;
		/* fprintf(stderr, "++++Line ignored (no  = ): %s\n", buf) ; */
	    }
        }

        badrefs = ccd_center(apos, xy, n, center, pix0, CD) ;
        scale = CD[0][0]*CD[1][1] - CD[0][1]*CD[1][0];
        memset(Rlocal, 0, sizeof(Rlocal)) ;
        tr_oR(center, Rlocal) ;

        printf("-c=%08.4f%+08.4f x0=%f y0=%f \\\nCD=%g,%g,%g,%g\n",
            center[0], center[1], pix0[0], pix0[1],
	    CD[0][0], CD[0][1], CD[1][0], CD[1][1]) ;
	if (badrefs) printf("****There are %d bad reference stars\n", badrefs);

	/* Now get just the (x,y) and compute RA, Dec */
        if (isatty(0)) {
            printf("----Give the (x,y), terminate with a blank line\n");
        }
	a = linxy;
	if (nxy>0) status = linxy[0];
	while(status) {
	    if (nxy>0) { status = *(a++); nxy--; }
	    else {
	        if (isatty(0)) printf("%4d> ", n);
	        status = fgets(buf, sizeof(buf), stdin);
	        /* Remove trailing blanks */
	        for (p = buf+strlen(buf)-1; (p>=buf) && isspace(*p); --p) *p=0;
	    }
	    if (!status) break ;
	    if (!status[0]) break ;
	    for (p=status; isspace(*p); p++) ;
	    if (p[0] == '#') continue ;
	    if (p[0] == '%') continue ;
	    if ((pv = strchr(status, ';'))) *pv = 0;
	    i = tr_af1(status, strlen(status), xy, 2);
	    if (pv) *pv = ';';
	    if (i != 2) {
	        fprintf(stderr, "++++Line ignored (no x,y): %s\n", status);
	        continue;
	    }
	    xy2pos(xy, center);
	    /*printf("%08.4f%+08.4f = %s\n", center[0], center[1], status);*/
	    ed_q(acoo, center, 6);
	    printf("%08.4f%+08.4f = %s = %s\n",
	       center[0], center[1], acoo, status);
	    n++;
	}
    }
    exit(0);
}
#endif
