/*++++++++++++++++++++++
.TYPE           Module
.IDENTIFICATION coocube.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [ESO-IPG]  
.KEYWORDS       Coordinate transformations

.VERSION 1.0	24-May-1989: Adapted from TRCOO
.VERSION 2.0	06-Apr-1994: New routine names.

.ENVIRONMENT	Any

.COMMENTS
\begin{TeX}
	coocube is a method of gnomonic projection on the
	faces of a cube. Dpending on the position on the
	celestial sphere, one of the 6 faces is chosen such
	that the projections are in the range $[-1, +1[$.

	The faces are numbered from 1 to 6; the axises for
	each face are defined in the following table:
	
	$$\begin{tabular}{| c | c | r r |}\hline
	\# & Face & $X$ & $Y$		\\ \hline
	1 & $z= 1$ 	& $ y$  & $-x$\\
	2 & $y= 1$ 	& $-z$  & $-x$\\
	3 & $x= 1$ 	& $ y$  & $ z$\\
	4 & $x=-1$	& $-z$  & $-y$\\
	5 & $y=-1$	& $ x$  & $ z$\\
	6 & $z=-1$ 	& $ x$  & $-y$\\
        \hline\end{tabular}$$
	

A representation of the faces of the cube is:
\begin{verbatim}

           +---------+
           |    ^    |
           |    |    |
           |   1+--->|
           |      x  |
           |         |
 +---------+---------+---------+---------+
 |    ^    |    ^    |         |         |
 |    |    |    |    |         |         |
 |   5+--->|  3 +--->|   2+--->|   4+--->|
 |      x  |      x  |   x|    |   x|    |
 |         |         |    v    |    v    |
 +---------+---------+---------+---------+
           |    ^    |
           |    |x   |
           |<---+    |
           |   6     |
           |         |
           +---------+
\end{verbatim}
\end{TeX}

-----------------------------------------------------------------------*/
 
#include <trigo.h>		/* Trigo functions	*/

static 	char axis6[3][8] = {	/* Definitions of Projections: 	*/
		{0,2,6,2,6,3,3,0},	/* Along X-axis	*/
		{0,4,4,1,5,1,5,0}, 	/* Along Y-axis	*/
		{0,1,2,3,4,5,6,0} 	/* Along z-axis	*/
   };
static  char adj_face[2][8] = {		/* Adjacent faces:	*/
		{0,2,6,2,6,3,3,0}, 	/* ... when x >= 1 	*/
		{0,4,4,1,5,1,5,0}	/* ... when y >= 1 	*/
   };
static double z;	/* Denominator for projection */

	/* u6 macro tranforms number 1--6 into correct direction cosine */
#define u6(i)		(i&4 ? -u[4^i] : u[3^i])
/* u6(i) gives the direction perpendicular to the face:
 --------------
   i   u6(i)
 --------------
   1   [2]=+z
   2   [1]=+y
   3   [0]=+x
   4   [0]=-x
   5   [1]=-y
   6   [2]=-z
 ------------- */

/*============================================================================*/
int face6(u)
/*+++
.PURPOSE Compute face number from unit cosines
.RETURNS Face number (1 to 6) if OK, 0 otherwise.
.ALGORITHM Face number is defined by:
\begin{TeX}
$$ 	|z| \geq \max(|x|,|y|) \Rightarrow 1 or 6 \qquad
	|x| \geq \max(|y|,|z|) \Rightarrow 3 or 4 \qquad
	|y| \geq \max(|z|,|y|) \Rightarrow 2 or 5 $$
\end{TeX}
---*/
  double u[3]; 	/* IN: dir cosines */
{ 
  int	k, n;	/* Face number	*/
  int	s[3];	/* Sorted faces	*/
	
	/* Sort the components of the unit vector to guess the face number */
    s[0] = 0; s[1] = 1; s[2] = 2;
    if (u[1] < u[0])	    s[0] = 1, s[1] = 0 ;
    if (u[s[2]] < u[s[0]])  k = s[2], s[2] = s[0], s[0] = k ;
    if (u[s[2]] < u[s[1]])  k = s[2], s[2] = s[1], s[1] = k ;
#ifdef DEBUG
printf(".... Ordered components: %d %d %d", s[0], s[1], s[2]);
#endif
    if (u[s[2]] > -u[s[0]]) n = 3 - s[2];
    else		    n = 4 + s[0];
#ifdef DEBUG
printf(" =%d=\n", n);
#endif

    return(n);
}

int aface6 ( n , dir )
/*+++		
.PURPOSE  Find out the adjacent face number
.RETURNS  The number of the adjacent face found if we move toward
	the direction specified by dir
.REMARKS  A direction is specified by a sign (+ or -) followed by x or y;
		other letters are ignored.
---*/
  int    n;	/* IN: plane number 1 to 6 */
  char *dir;	/* IN: Direction specification as +/- x/y */
{
  char  *p;
  int	i, k, face;

    for (face = n&7, k=0, p=dir; *p; p++)  switch(*p) {
    	case '+': k=0; continue;
    	case '-': k=1; continue;
    	case 'x': case 'X': case '0': i=0; goto compute_face;
    	case 'y': case 'Y': case '1': i=1; goto compute_face;
    	default:  k=0; continue;
    	compute_face: 
    	    face = adj_face[i][face];	/* Increasing x or y */
    	    if(k)  face = 7 - face;	/* Decreasing x or y */
    	    k=0; continue;
    }
    return(face);
}

/*============================================================================*/
int tr_ux6 ( u , p , n ) 
/*+++
.PURPOSE  Extract the significant xy values from xyz
.RETURNS  Face number (1 to 6) if OK, 0 otherwise.
.REMARKS  n == 0 ==> choose correct face number
---*/
  double u[3]; 	/* IN: dir cosines */
  double p[2]; 	/* OUT: value of projections */
  int    n;	/* IN: plane number 0 (derive from position), 1 to 6 */
{ 
  int	face, i;

    face = n == 0 ? face6(u) : n&7;
    if (! face) return (0);		/* Bad face number */
    z = u6(face);	/* Component perpendicular to face */
    if (z <= 1.e-6) return(0);	/* It's bad ! */
  
#ifdef DEBUG
printf(".... ux6: Face=%d =%d=", n, face);
#endif
		/* Along x-axis */
    i = axis6[0][face];
    p[0] = u6(i);
		/* Along y-axis */
    i = axis6[1][face];
    p[1] = u6(i);
#ifdef DEBUG
printf(" (%.6lf,%.6lf)\n", p[0], p[1]);
#endif

    return(face);
}

/*============================================================================*/
int tr_xu6 ( p , u , n )
/*+++		
.PURPOSE  Compute dir cos (unit vector) from xy values on plane n (1 to 6)
.RETURNS  n (plane number);
	  0   if n is outside [1,6] with  returned u[i] set to 0
---*/
  double p[2]; 	/* IN: value of projections */
  double u[3]; 	/* OUT: dir cosines 	*/
  int    n;	/* IN: plane number 1 to 6 */
{
  int	i, face;

    if ((face = axis6[2][n&7])) ;
    else { u[0] = u[1] = u[2] = 0.e0; return(0); }
  
  	/* Component perpendicular to face */
    z = sqrt(1.e0 - p[0]*p[0] - p[1]*p[1]);
    if (face&4)	u[4^face] = -z;
    else	u[3^face] = z;

	/* Component along x-axis */
    i = axis6[0][face];
    if (i&4)	u[4^i] = -p[0];
    else	u[3^i] =  p[0];
	/* Component along y-axis */
    i = axis6[1][face];
    if (i&4)	u[4^i] = -p[1];
    else	u[3^i] =  p[1];

    return(face);
}

/*============================================================================*/
int tr_pu6 ( p , u , n )
/*+++		
.PURPOSE   Compute dir cos (unit vector) from projection on plane n (1 to 6)
.RETURNS  n (plane number);
	  0   if n is outside [1,6] with  returned u[i] set to 0
---*/
  double p[2]; 	/* IN: value of projections */
  double u[3]; 	/* OUT: dir cosines 	*/
  int    n;	/* IN: plane number 1 to 6 */
{
  double factor;
  int	i, face;

    if ((face = axis6[2][n&7])) ;
    else { u[0] = u[1] = u[2] = 0.e0; return(0); }
  
  	/* Component perpendicular to face */
    factor = 1./sqrt(1.e0 + p[0]*p[0] + p[1]*p[1]);
    if (face&4)	u[4^face] = -factor;
    else	u[3^face] = factor;

	/* Component along x-axis */
    i = axis6[0][face];
    if (i&4)	u[4^i] = -p[0] * factor;
    else	u[3^i] =  p[0] * factor;
	/* Component along y-axis */
    i = axis6[1][face];
    if (i&4)	u[4^i] = -p[1] * factor;
    else	u[3^i] =  p[1] * factor;

    return(face);
}

/*============================================================================*/
int tr_up6 ( u , p , n ) 
/*+++
.PURPOSE  Projection on plane #n
.RETURNS  Face number (1 to 6) if OK, 0 otherwise.
.REMARKS  n == 0 ==> choose correct face number
---*/
  double u[3]; 	/* IN: dir cosines */
  double p[2]; 	/* OUT: value of projections */
  int    n;	/* IN: plane number 0 (derive from position), 1 to 6 */
{ 
  int	face;

    face = tr_ux6(u, p, n);		/* Find xy + z     */
    if (! face) return (0);		/* Bad face number */
  
		/* Along x-axis */
    p[0] /= z;
		/* Along y-axis */
    p[1] /= z;
#ifdef DEBUG
printf("....tr_up6(%d) => Face=%d (%.6lf,%.6lf)\n", n, face, p[0], p[1]);
#endif

		/* Test if values too large, use adjacent face */
    if (n == 0) {
    	if (p[0] >= 1.e0)  face = tr_up6(u, p, adj_face[0][face]); 
    	if (p[1] >= 1.e0)  face = tr_up6(u, p, adj_face[1][face]); 
    }
    return(face);
}

/*============================================================================*/
#ifdef DEBUG
#include <astro.h>
main()
/*+++++
.PURPOSE Test demonstration
.RETURN To OS
-----*/
{
	char buf[80];
	double o[2], u[3], p[2];
	int n;

  while(1)
  {
	printf("Give position: ");
  	if(!gets(buf))	break;
  	tr_ag(buf, strlen(buf), o);
  	printf("Input coordinates: %12.6f, %12.6f\n", o[0], o[1]);
  	tr_ou(o, u);
  	printf("Input unit vector: %12.9f, %12.9f, %12.9f\n", 
  		u[0], u[1], u[2]);
  	n = tr_up6(u, p, 0);
  	printf("Cube projectin #%d: %12.6f, %12.6f", n, p[0], p[1]);
  	printf("; adj -x=%d, +x=%d, -y=%d, +y=%d\n", 
  		aface6(n, "-x"),  aface6(n, "+x"), 
  		aface6(n, "-y"),  aface6(n, "+y"));
  	tr_pu6(p, u, n);
  	printf("      init vector: %12.9f, %12.9f, %12.9f\n", 
  		u[0], u[1], u[2]);
  	tr_uo(u, o);
  	printf("      Coordinates: %12.6f, %12.6f\n", o[0], o[1]);
    }
}
#endif
