/*+++++++++  
.MODULE    ednum.c
.AUTHOR    Francois Ochsenbein [ESO]
.LANGUAGE  C
.CATEGORY  Edition of numbers with variable number of decimals
	and of spherical coordinates
.COMMENTS
Functions in this module edit coordinates, either as 
decimal degrees ("G" type), or as 
equatorial coordinates ("Q" type).

Routines starting with ed2_ use the size of the output buffer
as the second parameter.

.VERSION 1.0 	26-Feb-1987: Creation 
.VERSION 1.1 	02-Mar-1987: Declinations are properly aligned.
.VERSION 1.2 	24-Mar-1987: Fill with blank if necessary
.VERSION 1.3 	14-Oct-1988: Cosmetic modifications
.VERSION 2.0	22-Feb-1991: Tested on PCs with TuboC
.VERSION 3.0	22-Nov-1993: Merged traq /etc
.VERSION 3.1	05-Feb-2001: Be sure to be within acceptable limits 
				for precision
.VERSION 3.2	03-Dec-2002: cooname function
.VERSION 3.3	18-Dec-2004: Fill with blanks / * in ed2_o / ed2_q
.VERSION 3.4	29-Sep-2008: 64-bit
.VERSION 3.5	25-Aug-2009: Edit blanks for NaN values
----------------------------------------------------------------------------*/
 
#include <math.h>		/* System	*/
#include <astropos.h>
#include <stdio.h>		/* For sprintf	*/
#define MAX(a,b)		((a) >= (b) ? (a) : (b))
#define MIN(a,b)		((a) <= (b) ? (a) : (b))
#define ITEMS(a)		sizeof(a)/sizeof(a[0])

#ifndef int4
#define int4 int
#define int8 long long
#endif

static double gf[] = {1., 10., 100., 1.e3,  1.e4,  1.e5,  1.e6,  1.e7,
	1.e8, 1.e9};
static double qf[] = {1., 10.,  60., 600., .36e4, .36e5, .36e6, .36e7, 
	.36e8, .36e9 };
			/* Bytes required */
static char gb[] = { 6, 10, 12, 14, 16, 18, 20, 22, 24, 26 };
static char qb[] = { 7, 10, 13, 16, 19, 22, 24, 26, 28, 30 };
/*===========================================================================*/
static int ed(char *dest, int4 value, int nd)
/*++++++++++
.PURPOSE Edit number `value' with `nd' digits
.RETURNS 0 for OK, remainder if number not edited.
-------------*/
  /* char  *dest :OUT: pointer to destination string 	*/
  /* int4 value   :IN: The number edit	*/
  /* int   nd     :IN: Number of digits	*/
{
  int4 	n;
  char 	*p;
  int	i;
    for (n = value, i = nd, p = dest + nd; --i >= 0; n /= 10)
  	*(--p) = '0' + n%10;
    i = n;
    return(i);
}

/*===========================================================================*/
int ed_o1(char *dest, double *value, int precision, int op)
/*++++++++++
.PURPOSE	Edit Single coordinates in decimal degrees
		(longitude or declination)
		The precision (number of decimals) must be between
		1 and 10.
.METHOD		Coordinates are rounded.
.RETURNS	Length of the nul-terminated edited string, -1
		for errors.
-------------*/
  /* char   *dest   :OUT: pointer to destination string 	*/
  /* double value[2] :IN: The coordinates to edit		*/
  /* int precision   :IN: Precision for edition		*/
  /* int    op       :IN: 0 = long, 1 = lat		*/
{
  int4 	n;
  char 	*p;
  double f;
  int	nd;	/* Number of Decimals 	*/

	/* Test bad coordinates	or not enough space	*/

    p = dest; *p = 0;

    if ((precision <= 0) || isnan(value[1]))
	return(0);
    if (precision > 10)	precision = 10;
    nd = precision - 1;

	/* Round the Longitude / Latitude	*/
	
    if (op == 0) {
   	f = value[0] + .5e0/gf[nd] ;
  	if (f < 0.0e0)		f += 360.0e0;
  	if (f >= 360.0e0)	f -= 360.0e0;
  	n = f;			/* Integer part */
  	ed(p, n, 3);
  	p += 3;
    }
    else {
   	*(p++) = (value[1] >= 0 ? '+' : '-');
  	f = fabs(value[1]) + .5e0/gf[nd] ;
  	n = f;			/* Integer part */
  	ed(p, n, 2);
  	p += 2;
    }  
  
	/* Edit Fractions */

    if (nd > 0) {
   	*(p++) = '.';
  	f = (f - n) * gf[nd];
  	n = f;
  	ed(p, n, nd);
  	p += nd;
    }

    *p = 0;	/* Terminate String */
    return(p - dest);
}
  
/*===========================================================================*/
int ed_o(char *dest, double *value, int precision)
/*++++++++++
.PURPOSE	Edit coordinates in decimal degrees.
		The precision (number of decimals) msut be between
		1 and 6.
.METHOD		Coordinates are rounded
.RETURNS	Length of the nul-terminated edited string, -1
		for errors.
-------------*/
  /* char *dest     :OUT: pointer to destination string 	*/
  /* double value[2] :IN: The 2 coordinates to edit	*/
  /* int  precision  :IN: Precision for edition		*/
{
  char 	*p;

#if DEBUG
    printf("ed_o: coordinates=(%12.6f,%12.6f)=%d=\n",
	value[0], value[1], precision);
#endif
  
  p = dest;

	/* Edit the Longitude	*/
  p += ed_o1(p, value, precision, 0);

  *(p++) = ' ';		/* Add a blank	*/
  
	/* Edit the Latitude	*/
  p += ed_o1(p, value, precision, 1);
  
  return(p - dest);
}
  
/*===========================================================================*/
int ed_q1(char *dest, double *value, int precision, int op)
/*++++++++++
.PURPOSE	Edit equatorial coordinates in the astronomical form.
		The precision ranges from 0 to 10, for degrees to 
		arc.sec/1000.
.METHOD		Coordinates are rounded
.RETURNS	The length (without the trailing nul) / -1 if error
.REMARKS	Right ascension with at least minutes.
-------------*/
  /* char  *dest     :OUT: pointer to destination string 	*/
  /* double value[2]  :IN: The 2 coordinates to edit	*/
  /* int  precision   :IN: The precision, 0 to 10		*/
  /* int   op         :IN: 0 = RA, 1 = dec	*/
{
  int4 	n;
  char 	*p;
  double f;
  int	nd, i;
  static char sep[3] = {0, '.', ' '};
  
	/* Test bad coordinates	or not enough space	*/
    p = dest; *p = 0;

    if ((precision <= 0) || isnan(value[1]))
	return(0);
    if (precision >= ITEMS(qf)) precision = ITEMS(qf)-1 ;

	/* Round and Edit the Right Ascension	*/
    if (op == 0) {
  	nd = precision; if (nd < 2) nd = 2;
  	f = value[0]/15.0e0 + 0.5e0/qf[nd] ;
  	if (f < 0.0e0)		f += 24.0e0;
  	if (f >= 24.0e0)	f -= 24.0e0;
    }
    else {
  	*(p++) = (value[1] >= 0 ? '+' : '-');
  	nd = precision - 1;
  	f = fabs(value[1]) + .5e0/qf[nd] ;
    }  

    n = f;			/* Integer part */
    ed(p, n, 2), p += 2;
  
	/* Edit mn	*/
    i = MIN(nd, 2);
    if (i == 0)	goto terminate;
    *(p++) = sep[i];
    f = (f-n) * qf[i];
    n = f;
    ed(p, n, i);
    p += i, nd -= i;
  
	/* Edit seconds	*/
    i = MIN(nd, 2);
    if (i == 0)	goto terminate;
    *(p++) = sep[i];
    f = (f-n) * qf[i];
    n = f;
    ed(p, n, i);
    p += i, nd -= i;
  
	/* Edit Fractions of seconds	*/
    if (nd == 0)	goto terminate;
    *(p++) = '.';
    f = (f-n) * gf[nd];
    n = f;
    ed(p, n, nd);
    p += nd;

  terminate:
    *p = 0;		/* Terminate String */
    return(p-dest);
}

/*===========================================================================*/
int ed_q(char *dest, double *value, int precision)
/*++++++++++
.PURPOSE	Edit equatorial coordinates in the astronomical form.
		The precision ranges from 0 to 10, for degrees to 
		arc.sec/1000.
.METHOD		Coordinates are rounded
.RETURNS	The length (without the trailing nul) / -1 if error
.REMARKS	The string is filled with * if the coordinates cannot be edited .
-------------*/
  /* char   *dest   :OUT: pointer to destination string 	*/
  /* double value[2] :IN: The 2 coordinates to edit	*/
  /* int  precision  :IN: The precision, 0 to 10		*/
{
  char 	*p;

#if DEBUG
    printf("ed_q: coordinates=(%12.6f,%12.6f)=%d=\n",
	value[0], value[1], precision);
#endif

    p = dest;

	/* Edit the Right Ascension	*/
    p += ed_q1(p, value, precision, 0);

    *(p++) = ' ';		/* Add a blank	*/
  
	/* Edit the Declination	*/
    p += ed_q1(p, value, precision, 1);
  
    return(p - dest);
}

/*===========================================================================
		Edition with a buffer
 *===========================================================================*/

int ed2_o(char *buf, int size, double *value, int precision)
/*++++++++++
.PURPOSE	Edit Galactic coordinates in the astronomical form.
		The precision ranges from 0 to 10, for degrees to 
		fractions of mas
.RETURNS	The length (without the trailing nul) / -1 if error
-------------*/
  /* char    *buf   :OUT: pointer to destination string 	*/
  /* int    size     :IN: Size of buffer			*/
  /* double value[2] :IN: The 2 coordinates to edit	*/
  /* int   precision :IN: The precision, 0 to 10		*/
{
  char 	*p, *pd;
  int   i, prec;

#if DEBUG
    printf("ed2_o: coordinates=(%12.6f,%12.6f)=%d=\n",
	value[0], value[1], precision);
#endif

    p  = buf; *p = 0;
    if ((precision <= 0) || isnan(value[1])) {
        /* No value -- fill with blanks */
	while (--size > 0) *(p++) = ' ';
	*p = 0;
	return(0);
    }
    if (precision > ITEMS(gb)) precision = ITEMS(gb);
    for (prec = precision-1; gb[prec] >= size; prec--) ;
    if (prec < 0) {	/* Fill with * */
	while (--size > 0) *(p++) = '*';
	*p = 0;
	return(0);
    }

	/* Compute where the delta part must start */
    for (i = ITEMS(gb)-1; gb[i] >= size; i-- ) ;
    pd = buf + (size-1) - (gb[i]>>1);
    
	/* Edit the Longitude		*/
    p += ed_o1(p, value, prec+1, 0);
    	/* Insert blanks to adjust left+right */
    while (p < pd) *(p++) = ' ';

	/* Edit the Declination	*/
    p += ed_o1(p, value, prec+1, 1);
    	/* Insert blanks to adjust left+right */
    for (pd = buf + (size-1); p < pd; p++) *p = ' ';
    *p = 0;
  
    return(p - buf);
}

/*===========================================================================*/
int ed2_q(char *buf, int size, double *value, int precision)
/*++++++++++
.PURPOSE	Edit equatorial coordinates in the astronomical form.
		The precision ranges from 0 to 10, for degrees to 
		fractions of mas
.RETURNS	The length (without the trailing nul) / -1 if error
.REMARKS	The string is filled with * if the coordinates cannot be edited .
-------------*/
  /* char    *buf   :OUT: pointer to destination string 	*/
  /* int    size     :IN: Size of buffer			*/
  /* double value[2] :IN: The 2 coordinates to edit	*/
  /* int   precision :IN: The precision, 0 to 10		*/
{
  char 	*p, *pd;
  int   i, prec;

#if DEBUG
    printf("ed2_q: coordinates=(%12.6f,%12.6f)=%d=\n",
	value[0], value[1], precision);
#endif

    p  = buf; *p = 0;
    if ((precision <= 0) || isnan(value[1])) {
    	/* No value -- fill with blanks */
	while (--size > 0) *(p++) = ' ';
	*p = 0;
	return(0);
    }
    if (precision > ITEMS(qb)) precision = ITEMS(qb);
    for (prec = precision-1; qb[prec] >= size; prec--) ;
    if (prec < 0) {	/* Fill with * */
	while (--size > 0) *(p++) = '*';
	*p = 0;
	return(0);
    }

	/* Compute where the delta part must start */
    for (i = ITEMS(qb)-1; qb[i] >= size; i-- ) ;
    pd = buf + (size-1) - (qb[i]>>1);
    
	/* Edit the Right Ascension	*/
    p += ed_q1(p, value, prec+1, 0);
    	/* Insert blanks to adjust left+right */
    while (p < pd) *(p++) = ' ';

	/* Edit the Declination	*/
    p += ed_q1(p, value, prec+1, 1);
    	/* Insert blanks to adjust left+right */
    for (pd = buf + (size-1); p < pd; p++) *p = ' ';
    *p = 0;
  
    return(p - buf);
}

/*===========================================================================*/

char *cooname(int type, double equinox)
/*++++++++++
.PURPOSE Get the standard naming of the coo frame
.RETURNS Internal buffer
-----------*/
  /* int type       :IN: Type of grame in _COO_xxx set */
  /* double equinox :IN: The equinox (relevqnt only for Equatorial) M*/
{
  static char buffer[32];
  char *p;
    p = (char *)0;	/* Set to non-zero if equinox to be edited */
    switch(type) {
      case _COO_GAL_:		return("G");
      case _COO_SUPERGAL_:	return("S");
      case _COO_ECL_:	p = buffer; *(p++) = 'E';	break;
      case _COO_FK4_:	p = buffer; *(p++) = 'B';	break;
      case _COO_FK5_:	p = buffer; *(p++) = 'J';	break;
      case _COO_ICRS_:		return("ICRS");
      default : 
	sprintf(buffer, "***Bad_cootype=%d\n", type);
    }
    if (p) sprintf(p, "%.0f", equinox);
    return(buffer);
}


#ifdef TEST
/*===========================================================================*/
#define ASIZE 40	/* Maxi accuracy */
#include <stdlib.h>
#include <string.h>

int main()
/*++++++++++++++++
.PURPOSE  Test program (for functions in this module)
.RETURNS  ---
-----------------*/
{ 
  char msg[133];
  double coo[2];
  int  n, i, asize;

    puts("Testing coordinates ascii/bin transformations\n");
    for(;;) {
  	printf("----Specify width in bytes for edition: ");
    	if(!fgets(msg, sizeof(msg), stdin)) break;
    	if (! *msg) break;
    	asize = atoi(msg);
	if (asize <= 1) asize = 0;
	if (asize > ASIZE) {
	    printf("++++Size restricted to %d bytes\n", ASIZE);
	    asize = ASIZE;
	}
    	for(;;) {
    	    printf("----Your input: "); 	
	    if (!fgets(msg, sizeof(msg), stdin)) break;
	    n = tr_aq(msg, strlen(msg), coo);
	    printf("====Returned status %d\n", n);
	    printf("\t%.16lg %.16lg\n", coo[0], coo[1]);
	    i = ed_q(msg, coo, n);
	    printf("\tEdited <%s> stat=%d\n", msg, i);
	    i = ed2_q(msg, asize, coo, n);
	    printf("\tEdited <%s> stat=%d\n", msg, i);
    	}
    }

    puts("Testing Galactic coordinates ascii/bin transformations\n");
    for(;;) {
  	printf("----Specify width in bytes for edition: ");
    	if(!fgets(msg, sizeof(msg), stdin)) break;
    	if (! *msg) break;
    	asize = atoi(msg);
	if (asize <= 1) asize = 0;
	if (asize > ASIZE) {
	    printf("++++Size restricted to %d bytes\n", ASIZE);
	    asize = ASIZE;
	}
    	for(;;) {
    	    printf("----Your input: "); 	
	    if (!fgets(msg, sizeof(msg), stdin)) break;
	    n = tr_aq(msg, strlen(msg), coo);
	    printf("====Returned status %d\n", n);
	    printf("\t%.16lg %.16lg\n", coo[0], coo[1]);
	    i = ed_o(msg, coo, n);
	    printf("\tEdited <%s> stat=%d\n", msg, i);
	    i = ed2_o(msg, asize, coo, n);
	    printf("\tEdited <%s> stat=%d\n", msg, i);
    	}
    }
    exit(0);
}
#endif
