/*++++++++
.MODULE    trtime.c
.AUTHOR    Francois Ochsenbein [ESO]
.LANGUAGE  C
.CATEGORY  Conversion from character to date structure.
.COMMENTS
\begin{TeX}
\begin{itemize}
\item	integer (mnemonic {\tt i}): the number of seconds elapsed
		since Jan 1, 1970, at 00:00:00.
		The available period is roughly 1902-2037.
\item	double (mnemonic {\tt jd}): the Julian Date
\item	structure (mnemonic {\tt tm}): a structure containing, as integers:
	\begin{itemize}
	\item	{\tt year} number
	\item 	{\tt mon} the month number in range [0-11]
	\item 	{\tt mday} the day number within month, in range [1-31]
	\item 	{\tt yday} the day number within year, in range [0-355]
	\item 	{\tt wday} the day number within week, in range [0-6]
				(0 is Sunday)
	\item 	{\tt hour} in range [0-23]
	\item 	{\tt min} in range [0-59]
	\item 	{\tt sec} in range [0-59]
	\end{itemize}
\end{itemize}
This routine does only include transformations between structure tm and
integer.
\end{TeX}

.VERSION   1.0	04-Mar-1987: Creation
.VERSION   1.1	21-Oct-1988: Cosmetic modifications
.VERSION   1.2	17-Feb-1989: Added tr_at
.VERSION   1.3	13-Jul-1989: Corrected bug in tr_atm when string is empty
				(returned status set to zero)
.VERSION   1.4	31-Jul-1990: Corrected bug in tr_atm when day not provided
				(only year + month)
.VERSION   1.5	04-Sep-1990: Corrected bug in tr_atm (be sure test p>pe OK)
.VERSION   1.6	10-Jun-1991: Assume CURRENT time only if "now" or a .
.VERSION   1.7	10-Nov-1993: Test on PC (Turbo-C). Merged all time functions
.VERSION   1.8	10-Dec-1998: Allow U(s) / S(s) seconds in Unix / 2000
.VERSION   1.9	29-Jun-2001: Accept AM/PM
.VERSION   1.91	01-Dec-2004: Bug with 'U' (Unix time) !!
.VERSION   2.0  15-Apr-2005: Accept dates outside Unix validity
.VERSION   2.1  25-May-2005: Accept 'D' = DMF time (origin = 1 Jan 1980)
.VERSION   2.2  10-Jan-2006: Decimal years ad Jyyy Byyy
.VERSION   2.21 06-Feb-2006: BUG fixed...
.VERSION   2.22 16-Jul-2006: -d alone means a TAB separator
.VERSION   2.3  29-Sep-2008: 64-bit
.VERSION   2.4  16-Jul-2009: accept delimiter
.VERSION   2.5  31-Jul-2011: Add format for edition of year. Accept negative yrs
.VERSION   2.51 24-Jun-2012: Bug in case of column(ized) input
.VERSION   2.6  15-Feb-2013: More possibilities in the interpretation of date
------------------------------*/

#define VERSION "2.6"

#include <time.h>
#include <ctype.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#ifndef int4
#define int4 int
#define int8 long long
#endif

#ifndef MAX_EDIT	/* Max. number of editions */
#define MAX_EDIT 20
#endif

/*===========================================================================
	Static variables for communications of internal routines
 *===========================================================================*/

typedef char	*S;
typedef S (*FTIME)();		/* Time edition function */
typedef struct { FTIME fct; char *arg; } fEDIT;
static fEDIT ftime[MAX_EDIT];	/* Max. # editions       */
static struct {			/* Error report	*/
  char *str;	/* Text where error occured	*/
  char *end;	/* Indicates end of str		*/
  struct tm *T;
 } Str;	/* Error Structure */
static	char	display_error = 1;	/* Could be changed */
static	char	err_level     = 1;	/* 0 for Warning    */
static	char	am            = 0;	/* 'A', 'P' or 0    */
static  char   *strp;			/* For fill_tm 	    */

#define ITEMS(a)        sizeof(a)/sizeof(a[0])
#define JD1970		2440587.5e0	/* JD   of 00:00:00 Jan. 1, 1970   */
#define s1980		(3652*86400)
#define s2000		(10957*86400)
#define LONG_MAX	2147483647		/* 2**31 - 1		*/
#define LONG_MIN	(-LONG_MAX-1)		/* -2**31		*/

static char month_list[] = {"\0Jan\01Feb\02Mar\03Apr\04May\05Jun\06Jul\07Aug\
\010Sep\011Oct\012Nov\013Dec\01Fev\03Avr\04Mai\05Jui\07Aou\011Okt\013Dez"};
static char day_list[]  = "Mon,Tue,Wed,Thu,Fri,Sat,Sun";
static unsigned char month[12]={31,28,31,30,31,30,31,31,30,31,30,31};

static int  verbop   = 0;  		/* Verbose option   */
static char *pgm;
static char usage[] =
	"Usage: %s [-dc] [-i colid] [-n|-a|-J|-U|-D|-S|-y[fmt]] [dates]\n";

#define GobbleSpaces(s,e)	while ((s < e) && (isspace(*s))) s++

/*===========================================================================
		Error report
 *===========================================================================*/

static int error(char *msg, char *err)
/*++++++
.PURPOSE Display Error of time parsing
.RETURNS 0 (no error) / 1 (error)
--------*/
  /* char	*msg;	IN: Error Message */
  /* char	*err;	IN: Position of bad character */
{
  char *p, c;

    if (! msg)	return(0);
    if (display_error) {
  	fprintf(stderr, "%s%s", err_level ? "#***" : "#+++", msg);
  	if (err) {
    	    c = *(Str.end), *(Str.end) = 0;
  	    fprintf(stderr, " in: %s\n         ",  Str.str);
  	    for (p = msg; *p; p++) fputc(' ', stderr);
  	    for (p = Str.str; p < err; p++) fputc(' ', stderr);
  	    fputc('^', stderr); fputc('\n', stderr);
    	    *(Str.end) = c;
     	}
     	else {		/* Use Time Structure */
     	    fprintf(stderr, ", tm=(%d-%d-%dT%d:%d:%d)\n",
     	        Str.T->tm_year, Str.T->tm_mon+1, Str.T->tm_mday,
     	        Str.T->tm_hour, Str.T->tm_min, Str.T->tm_sec);
    	}
    }
    return(1);
}

static int warning(char *msg, char *err)
/*++++++
.PURPOSE Display Warning message
.RETURNS 0 (no error) / 1 (error)
--------*/
  /* char	*msg;	IN: Error Message */
  /* char	*err;	IN: Position of bad character */
{
    err_level = 0;
    error(msg, err);
    err_level = 1;
    return(0);
}

static void help()
/*++++++++++++++++
.PURPOSE  Display help, then exit
.RETURNS  To OS
-----------------*/
{
  static char text[] = "\
  -dc: use c as column separator (defaut is tab)\n\
  -ix: input date starts at column (x+1)\n\
 -all: convert to all possible formats\n\
   -a: convert date to dd-mmm-yyyy \n\
   -g: convert date to YYYY-MM-DDThh:mm:ss (ANSI Gregorian)\n\
   -U: convert date to Unix time (seconds since 1970)\n\
   -D: convert date to DMF  time (seconds since 1980)\n\
   -S: convert date to Seconds   (seconds since 2000)\n\
   -y: convert date to (Besselian) decimal year, may be followed by format\n\
       e.g. -yB%9.4f or -y%g (default: %8.3f)\n\
  -By: convert date to Besselian year Byyyy.yyyyyyy (equiv. -yB%012.7f)\n\
  -JD: convert date to Julian date JDddddddd.fffff\n\
  -Jy: convert date to Julian year Jyyyy.yyyyyyy    (equiv. -yJ%012.7f)\n\
 date: Date in one of the above form or . for now\n\
       (default from stdin)\n\
" ;

    fprintf(stdout, usage, pgm);
    fprintf(stdout, "%s", text);
    exit(0);
}

static char *basename(char *name, int sub)
/*++++++++++++++++
.PURPOSE  Extract the 'name' from a complete path
.RETURNS  Converted
-----------------*/
  /* char *name; IN: Full file name */
  /* int  sub;   IN: Number of Subdirectories to take into account */
{
	char *p;

  for (p = name + strlen(name); --p >= name; ) {
	if (*p == '/') { if (--sub < 0) break; }
  }
  return(++p);
}

static int strloc(char *text, int c)
/*++++++++++++++++
.PURPOSE  Locate specified character
.RETURNS  Index of located char
-----------------*/
  /* char *text; IN: String to interpret */
  /* int  c;     IN: Char to look for */
{
  char *s;

    for ( s = text; *s; s++)	if (*s == c)	break;
    return(s-text);
}

/*===========================================================================
		Transformations between structure & Long integer
 *===========================================================================*/

static int cmp3(char *s1, char *s2)
/*+++
.PURPOSE  Comprare first 3 bytes of s1 and s2, case insensitive
.RETURNS  0 (match)
.REMARKS
---*/
  /* char *s1, *s2  :IN: Strings to compare */
{
  int r, i;

    for (r=0, i=3; (r == 0) && (--i >= 0); s1++, s2++)
       r = toupper(*s1) - toupper(*s2);
    return(r);
}

static int fill_tm(struct tm *T)
/*+++
.PURPOSE  Compute missing info in T structure
.RETURNS  0 (good date) / -1 (bad data: invalid month or day number ...)
.REMARKS
---*/
  /* struct tm *T :MOD:	Time as a structure */
{
  int4 y, j, t;
  char *err;

#ifdef DEBUG
    printf("....fill_tm called: %d.%d.%d.%d:%d:%d\n",
    	T->tm_year, T->tm_mon+1, T->tm_mday,
    	T->tm_hour, T->tm_min, T->tm_sec);
#endif

    month[1] = 29; err = (char *)0;

    if (T->tm_mday) {			/* Compute Day of 400 yr cycle 	*/
    	if ( (T->tm_mon > 11) || (T->tm_mon < 0) )
  	 	err = "bad month";
  	if (T->tm_mday <= 0)
  		err = "bad day_in_month";
	if (err) {
		error(err, strp);
  		return(-1);
  	}
    }
	/* Convert time	*/

    j = 0;
    t = ((T->tm_hour*60) + (T->tm_min))*60 + T->tm_sec;
    while (t < 0)		{ j--; t += 86400L; }
    while (t >= 86400L)	{ j++; t -= 86400L; }
#ifdef DEBUG
    printf("tr_tm: Time of day is %ld sec\n", t);
#endif
    T->tm_sec = t%60;
    t /= 60;
    T->tm_min = t%60;
    T->tm_hour= t/60;
    t = j;		/* Days	*/

    y = T->tm_year % 400;
    T->tm_year /= 400;
    if (y < 0)	{ y += 400; T->tm_year -= 1;}
    T->tm_year *= 400;
    if (T->tm_mday)			/* Compute Day of 400 yr cycle 	*/
    {	y += 400;			/* Positive interval	*/
  	j = y - (11 - T->tm_mon)/10;
	t += ((1461L*j)/4 + (306L * ((T->tm_mon+10)%12) + 5)/10
	    - (3* ((j + 100L)/100L))/4 + T->tm_mday + 59L) % 146097L;
	y -= 400;
    }
    else	t += 365*y + (y+3)/4 - (y-1)/100 + T->tm_yday;

    while (t < 0)	{ T->tm_year -= 400; t += 146097L;}

	/* Compute Week day	*/

    T->tm_wday = (t+6L)%7L;

	/* Compute Year	*/

    T->tm_year += 400L*(t/146097L);
    t %= 146097L;
    y = 4 *(t + (t-60)/36524L) / 1461L;
    T->tm_year += y;
    T->tm_yday = t - y*365L - (y+3)/4 + (y-1)/100;

	/* Compute mday + mon from yday	*/

    if (y&3)	month[1] = 28;
    if ( ((y%100) == 0) && (y))	month[1] = 28;
    if (T->tm_mday > month[T->tm_mon]) {
  	error("bad day_in_month", strp);
  	return(-1);
    }

    for (y=T->tm_yday, j=0; y>=month[j]; y -= month[j], j++)	;
    T->tm_mon  = j;
    T->tm_mday = y+1;
#ifdef DEBUG
    printf("Time %dh %dmn %ds\n", T->tm_hour, T->tm_min, T->tm_sec);
#endif

    return(0);
}

/*========================================================================*/
int tr_tm(struct tm *T)
/*+++
.PURPOSE  Compute missing info in T structure
.RETURNS  0 (good date) / -1 (bad data: invalid month or day number ...)
.REMARKS
---*/
  /*struct tm *T :MOD:	Time as a structure		*/
{
    strp = (char *)0;	/* If an error occurs, edit first tm	*/
    Str.T = T;

    return(fill_tm(T));
}

/*===========================================================================*/
int tr_ttm(int4 t, struct tm *T)
/*+++
.PURPOSE Transform integer time to structure
.RETURNS 0
.REMARKS
---*/
  /* int4 t :IN:	Time in seconds since Jan 1, 1970	*/
  /* struct tm *T :OUT: Time in the structure			*/
{
  int4 days, year;

#ifdef DEBUG
    printf("Input time is %ld sec", t);
#ifdef __MSDOS__
    printf(" -- timezone=%ld, daylight=%d, tzname=%c%c", timezone, daylight,
    	tzname[0], tzname[1]);
#endif
    printf("\n");
#endif
    days=  t/86400L;
    if (t < 0)	days--;
    t -= (days*86400L);
#ifdef DEBUG
    printf("Time of day is %ld sec\n", t);
#endif
    T->tm_hour = t/3600;
    T->tm_min  = 0;
    T->tm_sec  = (t%3600);
#ifdef DEBUG
    printf("Days since 1970: %ld\n", days);
    printf("Time  %dh %dmn %ds\n", T->tm_hour, T->tm_min, T->tm_sec);
#endif

    days   += (70L*365L + 18L);		/* Offset 1900	*/
    year    = (4L*days)/1461L;
    T->tm_year = year;
    T->tm_mon  = 0;
    T->tm_mday = 0;
    T->tm_yday = days - 365L*year - (year+3)/4;
    T->tm_year+= 1900;

    tr_tm(T);
#ifdef DEBUG
    printf("....tr_ttm gives: %d.%d.%d.%d:%d:%d\n",
    	T->tm_year, T->tm_mon+1, T->tm_mday,
    	T->tm_hour, T->tm_min, T->tm_sec);
    { struct tm *l;
      l = localtime( &t);
      printf("Local %dh %dmn %ds\n", l->tm_hour, l->tm_min, l->tm_sec);
    }
#endif
    return(0);
}

/*========================================================================*/
int tr_tmt(struct tm *T, int4 *t)
/*+++
.PURPOSE Transform structure time to time in seconds from Jan 1, 1970
.RETURNS 0 / -1 (time not in limits 1902-2037). Error message logged,
.REMARKS
---*/
  /* struct tm *T :IN:	Time as a structure		*/
  /* int4 *t      :OUT:	Time in seconds			*/
{
  register int4 n, y;

    *t = 0x80000000;	/* Default returned time...		*/
    strp = (char *)0;	/* If an error occurs, edit first tm	*/
    Str.T = T;

    if (fill_tm(T))		/* Compute missing information...	*/
	return(-1);

    if ((T->tm_year < 1902) || (T->tm_year >=2038)) {
    	error("Year outside limits [1902, 2038]", strp);
	return(-1);
    }

    n = T->tm_year - 1970;
    y = (n >= 0 ? n+1 : n-2);		/* Leap year counter */
    n = 365*n + y/4;

    *t = (n + T->tm_yday) * 86400L + T->tm_hour * 3600L
	+ T->tm_min * 60L + T->tm_sec;

    return(0);
}

/*===========================================================================
		Conversion JD
 *===========================================================================*/
double t2jd(int4 t)
/*+++
.PURPOSE Transform integer time to Julian Day
.RETURNS The equivalent Julian Date
---*/
  /* int4 t :IN: Time in seconds since Jan 1, 1970	*/
{
  double JD;	/* OUT: Julian date	*/

    JD  = t;
    JD /= 86400.e0;	/* In days since 1970	*/
    JD += JD1970;

    return(JD);
}

/*========================================================================*/
double tm2jd(struct tm *T)
/*+++
.PURPOSE Transform structure time to Julian Day
.RETURNS JD equivalence
.REMARKS
---*/
  /* struct tm *T :IN:	Time as a structure		*/
{
  int4 jm, j;
  double JD;

    strp = (char *)0;	/* If an error occurs, edit first tm	*/
    Str.T = T;
    fill_tm(T);		/* Fill what's missing in tm structure */

    if (T->tm_year <= -4712) {
  	j = 1 + (T->tm_year + 4712)/400;   	/* Number of 400-yr cycles */
  	jm = -j * 146097L;
	j  = T->tm_year + 400*j;
    }
    else jm = 0, j = T->tm_year;

    j  -= (11 - T->tm_mon)/10;
    jm += (1461L* ( j + 4712L))/4 + (306L * ((T->tm_mon+10)%12) + 5)/10
	- (3* ((j + 4900L)/100))/4 + T->tm_mday + 96;
    JD = jm;

    jm  = ((12+T->tm_hour))*3600L + T->tm_min*60 + T->tm_sec;
    JD += (jm/86400.e0);

    return(JD);
}

/*========================================================================*/
int4 jd2t(double JD)
/*+++
.PURPOSE Transform Julian Date to time in seconds from Jan 1, 1970
.RETURNS The time
.REMARKS
---*/
  /* double JD :IN: Julian date	*/
{
  int4 t; double x;

    x = ((JD - JD1970)*86400.e0);

    if ((x <= LONG_MIN)||(x >= LONG_MAX))
    	return(LONG_MIN);
    t = x;
    return(t);
}

int tr_jdt(double JD, int4 *t)
/*+++
.PURPOSE Transform Julian Date to time in seconds from Jan 1, 1970
.RETURNS 0 (success) / -1 (outside limits)
.REMARKS
---*/
  /* double JD :IN:  Julian date	*/
  /* int4 *t   :OUT: Date in seconds since Jan 1, 1970 */
{
  register double x;

    x = ((JD - JD1970)*86400.e0);

    if ((x <= LONG_MIN)||(x >= LONG_MAX))
	return(-1);
    *t = x;

    return(0);
}

/*========================================================================*/
int tr_jdtm(double JD, struct tm *T)
/*+++
.PURPOSE Transform Julian Date to time structure
.RETURNS 0
.REMARKS Assume Gregorian calendar
---*/
  /* double JD     :IN:	Julian date	*/
  /* struct tm *T :OUT: Date in structure	*/
{
  register int4 j, n4, nd10;
  double x;

    x = JD + 0.5e0 + 0.5e0/86400.e0;	/* Round day	*/
    j = x;
    if (x < 0)	 --j;
    x = x - j;				/* Fraction of day	*/

    T->tm_year = -4712;
    while (j <= 4480)	{ j += 146097L; T->tm_year -= 400;}

    n4 = 4*(j+((2*((4*j-17918L)/146097L)*3)/4+1)/2-37);
    nd10=10*( ((n4-237)%1461)/4)+5;

    T->tm_year += n4/1461;
    T->tm_mon  = (nd10/306+2)%12;
    T->tm_mday = (nd10%306)/10+1;

    x *= 86400.e0;			/* Seconds	*/
    j = x;
    T->tm_hour = j/3600;
    T->tm_min  = 0;
    T->tm_sec  = j%3600;

    tr_tm(T);				/* Add missing info...	*/

    return(0);
}

/*===========================================================================
		Interpretation from ascii strings
 *===========================================================================*/

int tr_atm(char *str, int len, struct tm *T)
/*++++++
.PURPOSE Converts a string to a date/time; input can be
	YYYY MM DD hh mm ss, or MMM DD YYYY hh mm ss,
	with month alphabetic or numeric.
.RETURNS -1 if error (error logged) /
	0 (empty string, curent time returned) /
	Number of Items present, from 1 (only
	year) to 6 (year + month + day + hours + min + sec)
.REMARKS An empty string gives the current date/time
--------*/
  /* char *str       :IN: String to scan	*/
  /* int  len        :IN: Length of string to scan	*/
  /* struct tm *T   :OUT: the date structure	*/
{
  char 	*p, *e, *a, err_flag;
  int	k, i=0, entered, tagged=0, neg=0;
  int components=0;	/* 1=s 2=min 4=hr 8=d 16=mon 32=yr 128=other */
  int4	value;
  double atof(), val;

    Str.str = p = str;
    Str.end = e = str + len;
    err_flag = 0;		/* No error */
    k = 0;			/* Returned status */

    while ((p < e) && (isspace(*p) || (*p=='+'))) p++;

    if ((p == e) || (((e-p)==1) && ispunct(*p))) {
    	/* Empty String or just . or - ... get Current Date / Time */
        value = time(0);
        tr_ttm(value, T);
  	return(0);
    }

    val = 0.;

    if (p[0] == 'M' && p[1] == 'J' && p[2] == 'D') {  /* MJD = JD - 2400000.5 */
	val = 2400000.5;
	p++;			/* Just skip the M */
    }
    if (p[0] == 'J' && p[1] == 'D') {	/* Julian Date (or MJD) */
    	p += 2;
    	GobbleSpaces(p, e);
    	strp = p;			/* Keep position in case of error */
	if (*p == '-')  p++; 		/* Accept sign */
    	if (isdigit(*p)) {
    	    while ((p < e) && (isdigit(*p))) p++;
    	    k=3; i=0;
    	    if ((p<e) && (*p == '.')) {
		/* Here i = number of decimals */
		++p;
    	    	while ((p<e) && (isdigit(*p))) p++, i++;
		if (tolower(*p) == 'e') {	/* Float */
		    p++; i -= atoi(p);	/* atoi(p) = Exponent */
		    if (*p == '+' || *p == '-') p++;
		    while ((p<e) && isdigit(*p)) p++;
		}
    	    	if (i >= 4) k = 6;	/* 1sec accuracy */
    	    	else if (i >= 2) k = 5;	/* 1min accuracy */
    	    	else if (i > 0) k = 4;  /* 1hr  accuracy */
    	    }
	    val += atof(strp);
    	    tr_jdtm(val, T);
    	    goto get_final;
    	}
    }

    /* Verify a decimal year */
    entered = 0;	/* Contains the time scale (B or J), -1 if error */
    if (isalpha(p[0]) && (!isalpha(p[1]))) { 
	entered = toupper(*p); 
	if ((entered != 'J') && (entered != 'B')) entered = -1;
	else p++; 
    }
    strp = p;			/* Keep position */
    if (*p == '-') p++;
    i = 0;			/* i = #decimals */
    if (isdigit(*p) && (entered>=0)) {
	while ((p<e) && isdigit(*p)) p++;
	if ((p<e) && (*p == '.')) {
	    for (++p; (p<e) && isdigit(*p); p++) i++;	/* i = #decimals */
	}
	while ((p<e) && isspace(*p)) p++;
    }
    if (p>=e) {			/* A single number ==> Year, converted to JD */
	if (entered == 'J') val = (atof(strp)-2000.0) * 365.25 + 2451545.0;
	else val = (atof(strp)-1900.0) * 365.242198781 + 2415020.31352;
	tr_jdtm(val, T);
	/* Estimate the accuracy , 1yr =~ 30Msec */
    	if (i >= 7) k = 6;	/* 1sec accuracy */
    	else if (i >= 5) k = 5;	/* 1min accuracy */
    	else if (i >= 4) k = 4;  /* 1hr  accuracy */
    	else if (i >= 2) k = 3;  /* 1d   accuracy */
    	else if (i >  0) k = 2;  /* 1mon accuracy */
	else k = 1;
	goto get_final;
    }
    p = strp; if(entered) p--;	/* Not a single number ==> beginning */
#if 0
    if (isalpha(p[0]) && isdigit(p[1])) {
	if (toupper(p[0]) == 'J') {
	    /* Conversion JYr to JD */
	    val = (atof(p+1) - 2000.0) * 365.25
		+ 2451545.0;
	}
	else if (toupper(p[0]) == 'B') {
	    /* Conversion of Besselian year to JD */
	    val = (atof(p+1) - 1900.0) * 365.242198781e0
		+ 2415020.31352;
	}
	if (val != 0.0) {
	    ++p;
	    tr_jdtm(val, T);
	    /* Estimate the accuracy , 1yr =~ 30Msec */
	    while ((p < e) && (isdigit(*p))) p++;
    	    if ((p<e) && (*p == '.')) {
		/* Here i = number of decimals */
		++p;
    	    	while ((p < e) && (isdigit(*p))) p++, i++;
		if (tolower(*p) == 'e') {	/* Float */
		    p++; i -= atoi(p);	/* atoi(p) = Exponent */
		    if (*p == '+' || *p == '-') p++;
		    while ((p<e) && isdigit(*p)) p++;
		}
    	    	if (i >= 7) k = 6;	/* 1sec accuracy */
    	    	else if (i >= 5) k = 5;	/* 1min accuracy */
    	    	else if (i >= 4) k = 4;  /* 1hr  accuracy */
    	    	else if (i >= 2) k = 3;  /* 1hr  accuracy */
    	    	else if (i >  0) k = 2;  /* 1hr  accuracy */
		else k = 1;
    	    }
	    goto get_final;
	}
    }
#endif

    /* Initialize the structure to zero	*/
    memset((char *)T, 0, sizeof(struct tm));
    T->tm_mon = -1;		    /* ... but with a bad month		*/
    entered = 0;

    while (p < e) {
    	if (k >= 6)		break;	/* We have all items!		*/
    	if (isspace(*p)) { p++; continue; }
    	strp = p;			/* Keep position in case of error */
	if (*p == '.') {		/* Could be a fraction ! */
	    if (entered == 0) break;	/* BAD, nothing known yet! */
	    for (a=p+1; (a<e) && isdigit(*a); a++) ;
	    if (a == e) {		/* It's a fraction */
		val = atof(p);
		p = a;
#if 0		/* Decimal year, already done */
		if (entered == 1) {	/* Fraction of yr  */
		    val *= 365.25;	/* Days */
		    if (neg<0) {	/* Negative fraction of year */
			T->tm_year += 1;
			val = 365.25 - val;
		    }
		    T->tm_yday = val;
		    val -= T->tm_yday;	/* Fraction of day */
		    entered = 3;	/* i.e. days */
		}
		if (entered >= 6) 	/* Ignore fractions of seconds */
		    continue;
		if (entered == 2) 	/* BAD */
		    break;
		if (entered <= 5) val *= 60;
		if (entered <= 4) val *= 60;
		if (entered <= 3) val *= 24;
		T->tm_sec = val;
#endif
		continue;
	    }
	}
	/* Other punctuation: may be a sign, implying a year */
	if (ispunct(*p)) {
	    if ((k==0) && isdigit(p[1])) {
		if (*p == '-') { neg = -1; p++; }
		else if (*p == '+') { neg = 1; p++; }
		else { p++; continue; }
	    }
	    else     { p++; continue; }
	}
	else neg = 0;
	if (isdigit(*p)) {   	/* A number comes now ...	*/
	    i = 0;
	    while ((p<e) && isdigit(*p)) {
	    	i = i*10 + (*p - '0');
	    	p++;
	    }
	    if (neg) { 		/* Can only be a year */
		entered=1; components |= 32;
		T->tm_year = i*neg; 
		k++; 
		continue; 
	    }
	    if ((tagged)&&(!islower(*p))) {
		/* Previous number was tagged (h m or s) */
		tagged++;
		if (tagged>6) tagged = 0;
	    }
	    else tagged = 0;	/* Set to 1(yr)...6(s) if number tagged */
	    if (islower(*p)) {	/* Can be followed by h m s */
		tagged = 4 + strloc("hms", tolower(*p));
		if (tagged<7) 
		    p++; 	/* Letter ok */
		else tagged = 0;
	    }
	    else if (*p == ':') {	/* Can only be a time */
		if (entered < 4) tagged = 4;
		else tagged = entered+1;
		p++;
	    }
	    if (tagged) {
		entered = tagged;
		if (entered == 4)      T->tm_hour = i;
		else if (entered == 5) T->tm_min  = i;
		else                   T->tm_sec  = i;
	    }
	    else if ((i>50)&&((components&32)==0)) {
		entered = 1; T->tm_year = i;
	    }
	    else switch(k) {
	      case 1:		/* 2nd number: Month or day number	*/
	    	if (T->tm_mon<0)	entered=2, T->tm_mon  = i-1;
	    	else if (i<32)		entered=3, T->tm_mday = i;
	    	else			entered=1, T->tm_year = i;
	    	break;
	      case 0: 	/* 1st number: Year or Day		*/
	      	if (i <= 31)	entered=3, T->tm_mday = i;
	      	else		entered=1, T->tm_year = i;
	    	break;
	      case 2:		/* Year or Day			*/
	      	if (T->tm_mday == 0)	entered=3, T->tm_mday = i;
		else if (ispunct(*p)&&(i<30))
		    			entered=4, T->tm_hour = i;
	      	else			entered=1, T->tm_year = i;
	      	break;
	      case 3:		/* Hours	*/
	      	entered=4, T->tm_hour = i;	break;
	      case 4:		/* Minutes	*/
	      	entered=5, T->tm_min = i;	break;
	      case 5:		/* Seconds	*/
	      	entered=6, T->tm_sec = i;	break;
	    }
	    components |= 1<<(6-entered);
    	}
	else if (isalpha(*p)) {		/* A text comes now ...		*/
	    while (isalpha(*p) && p < e) p++;
	    if (k >= 2)		continue;	/* Useless...	*/
	    for (i=1; i < sizeof(month_list); i += 4)
	    	if (cmp3(&month_list[i], strp) == 0)	break;
	    if (i >= sizeof(month_list)) {		/* Not Found...	*/
		if (!strchr(day_list, strp[0]))		/* Skip day of week */
	    	    warning("extra text ignored", strp);
	    	if (k > 0)	k++;
	    	continue;
	    }
	    entered = 2; components |= 16;	/* Month was entered */
	    T->tm_mon = month_list[i-1];
	}
	else {
	    error("spurious character", p);
	    err_flag = 1;
	    break;
	}
	if (components == (32|18|8)) 	/* Check YY MM DD */
	    err_flag = fill_tm(T);
	k++; continue;
    }
  get_final:
    if (am) {	/* Take the 'ap/pm' into account */
	if (am == 'A') {
	    if (T->tm_hour == 12)  T->tm_hour = 0 ;
	}
	else {
	    if (T->tm_hour < 12)   T->tm_hour += 12 ;
	}
    }
    GobbleSpaces(p, e);
    strp = p;
    if (! err_flag) err_flag = fill_tm(T);	/* Add missing info	*/
    if (p != e)
    	warning("extra text ignored", strp);

    /* Take current year if year not specified */
    if ((components&(8|32))==8) {
	T->tm_year = 1970 + (time(0)/31557600);
    }

    return(err_flag ? -1 : k);
}

struct tm *a2tm(char *str)
/*++++++
.PURPOSE Converts a string to a date/time;
.RETURNS The time tm structure (declared in time.h) /
	NULL if error
.REMARKS An empty string gives the current date/time
--------*/
  /* char *str :IN: String to scan	*/
{
  static struct tm T; int stat;

    stat = tr_atm(str, strlen(str), &T);
    return(stat < 0 ? (struct tm *)0 : &T);
}

/*===========================================================================*/
int tr_at(char *str, int len, int4 *t)
/*++++++
.PURPOSE Converts a string to a date/time; input can be
	YYYY MM DD hh mm ss, or MMM DD YYYY hh mm ss,
	with month alphabetic or numeric.
.RETURNS Length of the scanned text / -1 if error (error logged)
.REMARKS An empty string (value 0 returned) gives the current date/time
--------*/
  /* char *str :IN: String to scan	*/
  /* int  len  :IN: Length of string to scan	*/
  /* int4 *t  :OUT: the date (seconds since 1970)	*/
{
  struct tm T;
  int	status;

    *t = 0x80000000;
    if (len > 0)	status = tr_atm(str, len, &T);
    else		status = 0;

    if(status <= 0)	*t = time(0);
    else {
    	if (T.tm_year < 100)	T.tm_year += 1900;
  	if (tr_tmt(&T, t) < 0)	status = -1;
    }
    return(status);
}

int4 a2t(char *str)
/*++++++
.PURPOSE Converts a string to a date/time;
.RETURNS The time / 0x80000000 if error
.REMARKS An empty string gives the current date/time
--------*/
  /* char *str :IN: String to scan	*/
{
  int4 t;
    tr_at(str, strlen(str), &t);
    return(t);
}

/*===========================================================================
		Edition of date as a standard string
 *===========================================================================*/
static char wday[] = "SunMonTueWedThuFriSat";

static char *edT(struct tm *T)
/*++++++
.PURPOSE Edit date+time
.RETURNS The edited string as a standard 23-byte string
--------*/
  /* struct tm *T :IN: Date structure	*/
{
  static char buf[30];
  char c, *p;

    p = wday + 3*(T->tm_wday);
    c = p[3], p[3] = 0;
    sprintf(buf, "%s %4d-%02d-%02dT%02d:%02d:%02d", p,
	T->tm_year, T->tm_mon+1, T->tm_mday,
	T->tm_hour, T->tm_min,   T->tm_sec);
    p[3] = c;
    return(buf);
}

char *tm2a(struct tm *T)
/*++++++
.PURPOSE Edit date+time
.RETURNS The edited string as a standard 23-byte string
--------*/
  /* struct tm *T :IN: Date structure	*/
{
    strp = (char *)0;	/* If an error occurs, edit first tm	*/
    Str.T = T;
    fill_tm(T);		/* Fill what's missing in tm structure */
    return(edT(T));
}

char *t2a(int4 t)
/*++++++
.PURPOSE Edit date+time
.RETURNS The edited string as a standard 23-byte string
--------*/
  /* int4 t :IN: Date in seconds  */
{
  struct tm T;

    tr_ttm(t, &T);
    return(edT(&T));
}

/*==================================================================*/
static int tratm(char *str, struct tm *T)
/*++++++++++++++++
.PURPOSE  Convert Alphabetic to Numeric
.RETURNS  Status (0 = OK)
-----------------*/
  /* char *str     :IN: Time written in ASCII */
  /* struct tm *T :OUT: Time as LONG         */
{
  double atof();
  int status;
  int4 itime;
  char *p ;

    p = str+1;
    if (*p == '-') p++ ;
    if (isalpha(*str) && isdigit(*p)) switch(toupper(str[0])) {
      case 'J':
	if (toupper(str[1]) == 'D') {
	    status = tr_jdtm(atof(str+2), T);
	    return(status) ;
	}
	break ;
      case 'U':	/* Unix Time */
	while (isdigit(*p) || (*p == '.')) p++;
	if (*p) break ;
	itime = atol(str+1) ;
	status = tr_ttm(itime, T);
	return(status) ;
      case 'D':	/* DMF time  (from 1980) */
	while (isdigit(*p) || (*p == '.')) p++;
	if (*p) break ;
	itime = atol(str+1) + s1980;
	status = tr_ttm(itime, T);
	return(status) ;
      case 'S':	/* 2000 Time */
	while (isdigit(*p) || (*p == '.')) p++ ;
	if (*p) break ;
	itime = atol(str+1) + s2000 ;
	status = tr_ttm(itime, T);
	return(status) ;
    }

    return(tr_atm(str, strlen(str), T));
}
#if 0
static int trat(str, itime)
/*++++++++++++++++
.PURPOSE  Convert Alphabetic to Numeric
.RETURNS  Status (0 = OK)
-----------------*/
  char *str;	/* IN: Time written in ASCII */
  int4 *itime;	/* OUT: Time as LONG         */
{
  double atof();
  char *p ;

    p = str+1;
    if (*p == '-') p++ ;
    switch(toupper(str[0])) {
      case 'J':
	if (toupper(str[1]) == 'D') {
	    *itime = jd2t(atof(str+2));
	    return(0) ;
	}
	break ;
      case 'U':	/* Unix Time */
	while (isdigit(*p) || (*p == '.')) p++;
	if (*p) break ;
	*itime = atol(str+1) ;
	return(0) ;
      case 'D':	/* 1980 Time */
	while (isdigit(*p) || (*p == '.')) p++ ;
	if (*p) break ;
	*itime = atol(str+1) + s1980 ;
	return(0) ;
      case 'S':	/* 2000 Time */
	while (isdigit(*p) || (*p == '.')) p++ ;
	if (*p) break ;
	*itime = atol(str+1) + s2000 ;
	return(0) ;
    }

    return(tr_at(str, strlen(str), itime));
}

static char *edA(itime)
/*++++++++++++++++
.PURPOSE  Edition in Alphabetic format
.RETURNS  Converted
-----------------*/
  int4 itime;	/* IN: Time to edit */
{
  static char buf[25];
  char *p, *pm;
    p = t2a(itime);	/* Edited time */
    buf[0] = p[12]; buf[1] = p[13]; buf[2] = '-';
    pm = month_list + 4*atoi(p+9) - 3;
    buf[3] = pm[0]; buf[4] = pm[1]; buf[5] = pm[2]; buf[6] = '-';
    buf[7] = p[4]; buf[8] = p[5]; buf[9] = p[6]; buf[10] = p[7];
    buf[11] = 0;
    return(buf);
}
#endif

static char *edtmA(struct tm *T, char *arg /* unused argument*/)
/*++++++++++++++++
.PURPOSE  Edition in Alphabetic format
.RETURNS  Converted
-----------------*/
{
  static char buf[25];
  char *p, *pm=0;
    if (arg) fprintf(stderr, "#+++unused argument '%s'\n", arg);
    if ((T->tm_year>9999)||(T->tm_year<1582)) pm = (char *)1; /* error */
    else if (T->tm_year == 1582) {	/* Limit Gregorian = 15 Oct. 1582 */
	 if (T->tm_mon<=10) pm = (char *)1;
    }
    if (pm) {
	error("outside Gregorian calendar", 0);
	return("");
    }
    p = edT(T);
    buf[0] = p[12]; buf[1] = p[13]; buf[2] = '-';
    pm = month_list + 4*atoi(p+9) - 3;
    buf[3] = pm[0]; buf[4] = pm[1]; buf[5] = pm[2]; buf[6] = '-';
    buf[7] = p[4]; buf[8] = p[5]; buf[9] = p[6]; buf[10] = p[7];
    buf[11] = 0;
    return(buf);
}

#if 0
static char *edN(itime)
/*++++++++++++++++
.PURPOSE  Edition in Numeric format
.RETURNS  Converted
-----------------*/
  int4 itime;	/* IN: Time to edit */
{
    return(t2a(itime)+4);
}
#endif

static char *edtmN(struct tm *T, char *arg /* unused argument*/)
/*++++++++++++++++
.PURPOSE  Edition in Numeric format
.RETURNS  Converted
-----------------*/
{
    if (arg) fprintf(stderr, "#+++unused argument '%s'\n", arg);
    return(edT(T)+4);
}

#if 0
static char *edJD(itime)
/*++++++++++++++++
.PURPOSE  Edition in Numeric format
.RETURNS  Converted
-----------------*/
  int4 itime;	/* IN: Time to edit */
{
  static char buf[25];
    sprintf(buf, "JD%013.5f", t2jd(itime));
    return (buf);
}
#endif

static char *edtmJD(struct tm *T, char *arg /* unused argument*/)
/*++++++++++++++++
.PURPOSE  Edition in JD
.RETURNS  Converted
-----------------*/
{
  static char buf[25];
    if (arg) fprintf(stderr, "#+++unused argument '%s'\n", arg);
    sprintf(buf, "JD%013.5f", tm2jd(T));
    return (buf);
}

static char *edtmJy(struct tm *T, char *arg /* unused argument*/)
/*++++++++++++++++
.PURPOSE  Edition in Julian years
.RETURNS  Converted
-----------------*/
{
  static char buf[25];
  double value;
    if (arg) fprintf(stderr, "#+++unused argument '%s'\n", arg);
    value = ((tm2jd(T) - 2451545.0)/365.25) + 2000.;
    sprintf(buf, "J%012.7f", value);
    return (buf);
}

static char *edtmBy(struct tm *T, char *arg /* unused argument*/)
/*++++++++++++++++
.PURPOSE  Edition in Besselian years
.RETURNS  Converted
-----------------*/
{
  static char buf[25];
  double value;
    if (arg) fprintf(stderr, "#+++unused argument '%s'\n", arg);
    value = ((tm2jd(T) - 2415020.31352)/365.242198781) + 1900.;
    sprintf(buf, "B%012.7f", value);
    return (buf);
}

static char *edtmy(struct tm *T, char *arg)
/*++++++++++++++++
.PURPOSE  Edition in Besselian years -- use the format specified in arg
.RETURNS  Converted
-----------------*/
{
  static char buf[40];
  double value;
    if (!arg) arg = "%8.3f";
    value = toupper(*arg) == 'J' ?
	    ((tm2jd(T) - 2451545.0)/365.25) + 2000.:
	    ((tm2jd(T) - 2415020.31352)/365.242198781) + 1900. ;
    /* if (snprintf(buf, sizeof(buf), arg, value) >= sizeof(buf))
          fprintf(stderr, "#+++Edition of time truncated to %d bytes (%s)\n",
	                  sizeof(buf), arg);   */
    if (strlen(arg)>20) {
	fprintf(stderr, "#+++Edition of time too long, default assumed: %s\n",
		arg);
	arg = "%8.3f";
    }
    sprintf(buf, arg, value);
    return (buf);
}

#if 0
static char *edU(itime)
/*++++++++++++++++
.PURPOSE  Edition in Unix format
.RETURNS  Converted
-----------------*/
  int itime;	/* IN: Time to edit */
{
  static char buf[25];
    sprintf(buf, "U%010d", itime) ;
    return (buf);
}
#endif

static char *edtmU(struct tm *T, char *arg /* unused argument*/)
/*++++++++++++++++
.PURPOSE  Edition in Unix format
.RETURNS  Converted
-----------------*/
{
  static char buf[25];
  int itime;
    if (arg) fprintf(stderr, "#+++unused argument '%s'\n", arg);
    if (tr_tmt(T, &itime)) *buf = 0;
    else sprintf(buf, "U%010d", itime) ;
    return (buf);
}

#if 0
static char *edS(itime)
/*++++++++++++++++
.PURPOSE  Edition in Unix format
.RETURNS  Converted
-----------------*/
  int itime;	/* IN: Time to edit */
{
  static char buf[25];
    sprintf(buf, "S%010d", itime-s2000) ;
    return (buf);
}
#endif

static char *edtmS(struct tm *T, char *arg /* unused argument*/)
/*++++++++++++++++
.PURPOSE  Edition in Unix format
.RETURNS  Converted
-----------------*/
{
  static char buf[25];
  int itime;
    if (arg) fprintf(stderr, "#+++unused argument '%s'\n", arg);
    if (tr_tmt(T, &itime)) *buf = 0;
    else sprintf(buf, "S%010d", itime-s2000) ;
    return (buf);
}

static char *edtmD(struct tm *T, char *arg /* unused argument*/)
/*++++++++++++++++
.PURPOSE  Edition in Unix format
.RETURNS  Converted
-----------------*/
{
  static char buf[25];
  int itime;
    if (arg) fprintf(stderr, "#+++unused argument '%s'\n", arg);
    if (tr_tmt(T, &itime)) *buf = 0;
    else sprintf(buf, "D%010d", itime-s1980) ;
    return (buf);
}

#if 0
static char *edY(int itime)
/*++++++++++++++++
.PURPOSE  Edition in Unix format
.RETURNS  Converted
-----------------*/
{
  static char buf[25];
  double y ;
    y = (itime-s2000)/(365.25*86400.) ;
    y += 2000. ;
    sprintf(buf, "%.3f", y) ;
    return (buf);
}

static char *edtmY(struct tm *T, char *arg /* unused argument*/)
/*++++++++++++++++
.PURPOSE  Edition in years
.RETURNS  Converted
-----------------*/
{
  static char buf[25];
  double y ;
    y = 1970. + (tm2jd(T)-JD1970)/365.25;
    sprintf(buf, "%.3f", y) ;
    return (buf);
}
#endif

static fEDIT edtime[] = {	/* All functions */
    {edtmJD}, {edtmN}, {edtmJy}, {edtmBy}, {edtmU}, {edtmS}, {edtmA}
};

/*==================================================================*
		Program
 *==================================================================*/
#define add_ftime(f)	if (nedit < MAX_EDIT-1) ftime[nedit++].fct = f; \
	       else fprintf(stderr, "#+++Not more than %d methods!\n", nedit);
int main (int argc, char **argv)
{
  struct tm T;
  int nc, i, o, len, status;
  FTIME fct;
  /* static S (*af)() = edtmJD;
     S (*fct)();
  */
  char line[1025], *p, *aline, x;
  int nedit = 0;
  char dc = 0;

    nc = o = 0;

	/* Check the Arguments */

    pgm = basename(*argv, 0);

    while (-- argc > 0) {
	p = *++argv;
	if ((*p == '-') && (!isdigit(p[1])))  switch (*++p) {
	  case 'i':		/* Skip column start  */
		if (isdigit(p[1])) o = atoi(p+1) ;
		else o = atoi(*++argv), argc-- ;
		continue ;
	  case 'B': case 'b':
		add_ftime(edtmBy);	/* Bessielian    Year */
		continue ;
	  case 'g' :
	  case 'n' :
		add_ftime(edtmN);	/* Convert to Numeric */
		continue;
	  case '*':			/* All functions */
	  	memcpy(ftime, edtime, sizeof(edtime));
		nedit = ITEMS(edtime);
		continue;
	  case 'a' :
		if (p[1] == 'l') {	/* All conversions... */
		    memcpy(ftime, edtime, sizeof(edtime));
		    nedit = ITEMS(edtime);
		}
		else {
		    add_ftime(edtmA);  	/* Convert to Roman   */
		}
		continue;
	  case 'J' : case 'j':		/* Julian day or year */
		if(tolower(p[1]) == 'y') { add_ftime(edtmJy); }
		else                     { add_ftime(edtmJD); }
		continue;
	  case 'D':			/* DMF format (1980)  */
		add_ftime(edtmD);
		continue ;
	  case 's': case 'S':		/* Seconds since 2000 */
		add_ftime(edtmS);
		continue ;
	  case 'u': case 'U':		/* Seconds since 1970 */
		add_ftime(edtmU);
		continue ;
	  case 'y': case 'Y':
		add_ftime(edtmy);	/* Simple year (Besselian) */
		--nedit; ftime[nedit].arg = "%8.3f";
		if (strchr(p, '%')) {	/* There is a format */
		    ftime[nedit].arg = malloc(strlen(p)+2); 
		    strcpy(ftime[nedit].arg, p+1);
		    for (p=strchr(p, '%'); *p && (!strchr("efg", *p)); p++);
		    if (!*p) { 	/* Missing format ==> added */
			strcat(ftime[nedit].arg, "f");
			fprintf(stderr, "#+++added 'f' in format: %s\n",
				ftime[nedit].arg);
		    }
		}
		nedit++;
		continue ;
	  case 'd':
		dc = p[1];
		if (!dc) dc = '\t';	/* V2.22 */
		continue;
	  case 'v':                     /* Verbose      */
		if (strncmp(p, "ver", 3) == 0) {
		    printf("jdate (CDS) Version %s\n", VERSION);
		    exit(0);
		}
		verbop = 1;
		/* fprintf(stderr, "... %s: Examine argument ", pgm+4); */
		continue;
	  case 'h':                     /* Help         */
		help();
		continue;
	  case 0:			/* Data in stdin */
		continue;
	  default:
	  /* bad_option: */
		fprintf(stderr, "#***Bad option: %s", p);
		exit(1);
	} else {
		if (nedit == 0) ftime[nedit++].fct = edtmJD;
		status = tratm(p, &T);
		for (i=0; i<nedit; i++) {
		    if (i) printf(" = ");
		    fct = ftime[i].fct;
		    printf("%s", (*fct)(&T, ftime[i].arg));
		    nc++;
		}
		printf("\n");
		continue;
	}
    }
    if (nc)	exit(0);
    if (nedit == 0) ftime[nedit++].fct = edtmJD;

    while(fgets(line, sizeof(line), stdin)) {
	/* Remove the end-of-line */
	p = line + strlen(line) -1;
	if ((p>=line) && (*p == '\n')) *(p--) = 0;
	if ((p>=line) && (*p == '\r')) *(p--) = 0;
	if (p<line) p = line;
	/* Find out where the date is located... */
	if (dc) for (aline=line, i=o; i>0; i--) {
	    aline += strloc(aline, dc);
	    if (*aline) aline++;
	}
	else {
	    aline = line + o ;
	    if (aline>p) aline = p;
	}
	if (aline>line) fwrite(line, aline-line, 1, stdout) ;
	len = strloc(aline, dc);
	am = 0 ;	/* 'A', 'P' for am pm */
#if 0
	if (dc == ':') {
	    p = aline+len+1;	/* p -> mm:ss */
	    if(isdigit(p[0]) && isdigit(p[1]) && (p[2] == ':')) len += 3;
	    p = aline+len+1;	/* p -> ss.cc */
	    if(isdigit(p[0]) && isdigit(p[1])) len += 3;
	    p = aline+len;
	    if ((p[0] == '.') || (p[0] == ':')) for (++p; isdigit(*p); p++) ;
	    am = toupper(p[0]) ;
	    if (((am == 'A') || (am == 'P')) && (toupper(p[1] == 'M')))
		 p += 2 ;
	    else am = 0 ;
	}
#endif
	x = aline[len]; aline[len] = 0;
	if (len>0) {		/* Empty date, do nothing */
	    status = tratm(aline, &T);
	    for (i=0; i<nedit; i++) {
	        if (i) printf(" = ");
	        fct = ftime[i].fct;
	        printf("%s", (*fct)(&T, 0));
	    }
	}
	if (x) aline[len] = x, printf("%s", aline+len);
	printf("\n");
    }
    exit(0);
}
