/*++++++++++++++++++++++++
.TYPE		Module
.IDENTIFICATION	qbox.c
.AUTHOR		Francois Ochsenbein [ESO]
.VERSION  1.0	28-Dec-1998: Creation from coocell.c
.VERSION  1.1	06-Jan-1999: Main Program
.VERSION  2.0	22-Oct-1999: More discrimant choices...
.VERSION  2.1	22-Jan-2005: Added qbox_lower qbox_upper (of current level)
			Addition of qbox_ob, qbox_obrot
.VERSION  2.2	01-Feb-2007: qbox_adj (adjacent qbox), qbox_nearby
.VERSION  2.3	07-May-2007: Function qbox_a
.VERSION  2.4	13-Aug-2007: accept named qboxes [NOPQRS][a-p]*[0-3]?
.VERSION  2.5	25-Oct-2007: function qbox_size
.VERSION  2.6	02-Apr-2009: check_radius uses same algorithm as Java version
.VERSION  2.61	24-Sep-2009: accept abox as an integer
.Language	C
.ENVIRONMENT	Any
.KEYWORDS	Coordinates indexing
.COMMENTS
\begin{TeX}
	The routines provided in this module all deal with "qboxes".
	The qboxes are a way to split the celestial sphere in zones
	of approximately same surfaces and extension.

	The dichotomic procedure uses the projections (X,Y) of the
	directions on the 6 faces of a cube
       	numbered 1~($z=1$),  2~($y=1$),  3~($x=1$),
		4~($x=-1$), 5~($y=-1$) and 6~($z=-1$).

	Each cube face is then divided into 4 areas; each area is further
	divided into four pieces, etc.
	The number of valid qboxes is

	\begin{tabular}{|r r r|} \hline
	Level & Lower qboxes number  & Number of qboxes \\
	0 & 9 &     6 \\
	1 & 36 &    24 \\
	2 & 144 &    96 \\
	3 & 576 &   384 \\
	4 & 2304 &  1536 \\
	5 & 9216 &  6144 \\
	6 & 36864 & 24576 \\
	7 & 147456 & 98304 \\
	8 & 589824 & 393216 \\
	9 & 2359296 & 1572864 \\
	10& 9437184 & 6291456 \\
	11& 37748736 & 25165824 \\
	12& 150994944 & 100663296 \\
        \hline\end{tabular}

	where level is a number which can be changed at any time using
	{\tt qbox\_set}({\em level}) procedure.

	The present implementation assumes that {\em level $\leq 12$},
	corresponding to a maximum of 28 bits in the qbox number;
	this condition can be removed with a redefinition of the
	{\tt QBOX\_BITS} definition in this module;
	constants can be obtained by compiling the program with the INIT
	option set to 1.

	For a level 6, the qbox number is a short integer with bits
	(from left to right)
	{\fbox{{\tt 1ppp xyxy xyxy xyxy}}}.
	The leftmost `1' bit allows the recognition of the level;
	bits {\tt p} represent the face number (1 to 6),
	{\tt x} and {\tt y} the position, expressed with {\em level} bits,
	along the axises defined above for the face; going one level down
	in the hierarchy is therefore just a shift of 2 bits and
	updating the 2 rightmost bits.

	Qboxes essentially represent a "pixel" on the sky; 
	they may be edited either as numbers (e.g. 1.300000 which represents,
	the cell close to the North Pole with RA between 0 and 6hr in level=6),
	or with letters NPOQRS for faces (N=North, O=0h, P=6h, Q=12h, R=18h,
	S=South), followed by letters [a-p] for 2-level aggregations,
	and terminated by a number [0-3] for odd level (Nmaa = 1.300000)

	The selection of valid qboxes makes usage of a `callback routine',
	a user--supplied function which is called after the tests
	and has synopsis {\tt routine}({\em qbox\_no}, {\em status})
	where {\em status} may take the values
	\begin{itemize}
	\item 	{\tt QBOX\_NONE} (0): no object in {\em qbox\_no} can meet
		the position requirements.
	\item 	{\tt QBOX\_ANY} (-1): any object in {\em qbox\_no} meets
		the position requirements; there is therfore no need for a test.
	\item 	{\tt QBOX\_SOME} (1): objects in {\em qbox\_no} may meet
		the position requirements; individual tests must be performed.
	\end{itemize}

A representation of the faces of the cube is:
\begin{verbatim}

           +---------+
           |(N) ^    |
           |    |    |
           |   1+--->|
           |      x  |
           |         |
 +---------+---------+---------+---------+
 |(R) ^    |(O) ^    |(P)      |(Q)      |
 |    |    |    |    |         |         |
 |   5+--->|  3 +--->|   2+--->|   4+--->|
 |      x  |      x  |   x|    |   x|    |
 |         |         |    v    |    v    |
 +---------+---------+---------+---------+
           |(S) ^    |
           |    |x   |
           |<---+    |
           |    6    |
           |         |
           +---------+
\end{verbatim}
\end{TeX}

--------------------*/

/* Problem with inconsistent definition of pointer to double as 3x3 matrix */
#include <qbox.h>		/* Definitions for qboxes	*/
#include <trigo.h>		/* Spherical Function		*/
#include <math.h>		/* All math functions		*/
#include <ctype.h>		/* isalpha isdigit etc		*/
#include <stdio.h>		/* sprintf			*/
#include <string.h>		/* memcpy			*/
#include <stdlib.h>		/* memcpy			*/

#ifndef INIT
#define INIT 	0 		/* Printed results if INIT set to 1  */
#endif

#ifndef QBOX_BITS
#define QBOX_BITS	28	/* Bits used for a complete qbox definition */
#endif

#define QBOX_LEVEL  	((QBOX_BITS-4)>>1)	\
				/* Level in the dichotomic procedure	*/

struct QBOX_def {		/* Definitions of Qbox */
    int	maxlevel;		/* Maximal level	*/
    int	level;			/* Current level	*/
    double 	radius[1+QBOX_LEVEL];	/* Max distance	*/
    double 	sinrad[1+QBOX_LEVEL];	/* sinus 	*/
    double	minrad[1+QBOX_LEVEL];	/* Min distance */
#ifdef QLIMITS
    double	qbox_limit[1+(1<<(QBOX_LEVEL-1))] ;
#endif
 };

/* List of adjacent faces, when moving toward one of +x +y -y -x
 * The face is flagged by 8 when a rotation in the axes occurs
 * (always a rotation of +pi/2)
 * */
static  char adj_face[4][8] = {		/* Adjacent faces:	*/
    {0,  2,8|6,8|2,  6,  3,8|3,0}, 	/* ... when x >= 1 (0)	*/
    {0,8|4,  4,  1,8|5,8|1,  5,0},	/* ... when y >= 1 (1)	*/
    {0,  3,8|3,8|6,  2,  6,8|2,0},	/* ... when y < -1 (2)	*/
    {0,8|5,  1,  5,8|1,8|4,  4,0},	/* ... when x < -1 (3)	*/
};
static char adj_move[4][4] = {		/* Adjacent cells	*/
    { 2, 3, 4, 5 },			/* (0) toward +x	*/
    { 1, 4, 3, 6 },			/* (1) toward +y	*/
    { 5, 0, 7, 2 },			/* (2) toward -y	*/
    { 6, 7, 0, 1 },			/* (3) toward -x	*/
};
static char face_letter[] = ".NPOQRS" ;	/* Alphabetic face name	*/

#if INIT
#define QBOX_AXIS	(1<<(1+QBOX_LEVEL))	/* dimension on 1 axis   */
#define Initialize	if_not(qbox_init_done) qbox_init();
static int qbox_init();
static double min_s2d() ;
static double qbox_limit [1+QBOX_AXIS];
static double qbox_surf_min[1+QBOX_LEVEL];
static double qbox_surf_max[1+QBOX_LEVEL];
static char qbox_init_done  = 0 ;	/* flag init done 	*/
static int    ijc;

static struct QBOX_def qboxes = {QBOX_LEVEL, QBOX_LEVEL};

#else

#define Initialize
static struct QBOX_def qboxes = { 12, 12,
    {    54.735610317245  /* Level  0:   radius  54.73561/ 45.00000 */
    ,    30.361193404822  /* Level  1:   radius  30.36119/ 20.94102 */
    ,    16.403086517223  /* Level  2:   radius  16.40309/  9.39111 */
    ,     8.639594386107  /* Level  3:   radius   8.63959/  4.35381 */
    ,     4.449359566016  /* Level  4:   radius   4.44936/  2.08468 */
    ,     2.259674851696  /* Level  5:   radius   2.25967/  1.01857 */
    ,     1.138910822988  /* Level  6:   radius   1.13891/  0.50326 */
    ,     0.571763981623  /* Level  7:   radius   0.57176/  0.25011 */
    ,     0.286464127736  /* Level  8:   radius   0.28646/  0.12468 */
    ,     0.143378219959  /* Level  9:   radius   0.14338/  0.06224 */
    ,     0.071725726570  /* Level 10:   radius   0.07173/  0.03110 */
    ,     0.035872027118  /* Level 11:   radius   0.03587/  0.01554 */
    ,     0.017938305727  /* Level 12:   radius   0.01794/  0.00777 */
    }, /* Sine of above radii */
    { 0.8164965809277260  /* Level  0: Surface  6.875e+03 / 6.875e+03 */
    , 0.5054494651244236  /* Level  1: Surface  1.719e+03 / 1.719e+03 */
    , 0.2823931345695149  /* Level  2: Surface  4.170e+02 / 4.825e+02 */
    , 0.1502185901567493  /* Level  3: Surface  9.780e+01 / 1.250e+02 */
    , 0.0775779474805375  /* Level  4: Surface  2.345e+01 / 3.154e+01 */
    , 0.0394285430318791  /* Level  5: Surface  5.729e+00 / 7.904e+00 */
    , 0.0198764347423619  /* Level  6: Surface  1.415e+00 / 1.977e+00 */
    , 0.0099789983974146  /* Level  7: Surface  3.517e-01 / 4.944e-01 */
    , 0.0049997213877594  /* Level  8: Surface  8.766e-02 / 1.236e-01 */
    , 0.0025024194021835  /* Level  9: Surface  2.188e-02 / 3.090e-02 */
    , 0.0012518497600656  /* Level 10: Surface  5.466e-03 / 7.725e-03 */
    , 0.0006260849416742  /* Level 11: Surface  1.366e-03 / 1.931e-03 */
    , 0.0003130824920569  /* Level 12: Surface  3.415e-04 / 4.828e-04 */
    },
    {    45.000000000000  /* Level  0:   radius  54.73561/ 45.00000 */
    ,    20.941020472244  /* Level  1:   radius  30.36119/ 20.94102 */
    ,     9.391113935712  /* Level  2:   radius  16.40309/  9.39111 */
    ,     4.353808198538  /* Level  3:   radius   8.63959/  4.35381 */
    ,     2.084680032733  /* Level  4:   radius   4.44936/  2.08468 */
    ,     1.018569701478  /* Level  5:   radius   2.25967/  1.01857 */
    ,     0.503260558091  /* Level  6:   radius   1.13891/  0.50326 */
    ,     0.250114431503  /* Level  7:   radius   0.57176/  0.25011 */
    ,     0.124677058407  /* Level  8:   radius   0.28646/  0.12468 */
    ,     0.062243342044  /* Level  9:   radius   0.14338/  0.06224 */
    ,     0.031097855853  /* Level 10:   radius   0.07173/  0.03110 */
    ,     0.015542971843  /* Level 11:   radius   0.03587/  0.01554 */
    ,     0.007769996615  /* Level 12:   radius   0.01794/  0.00777 */
#ifdef QLIMITS
    }, /* Limits of Qboxes... */
    {     0.000000000000  /* n=    0 */
    ,     0.000383495216  /* n=    1 */
    ,     [...]           /* n=[...] */
    ,     0.999233303593  /* n= 2047 */
    ,     1.000000000000  /* n= 2048 */
#endif
    } };
#endif

	/* Shared Variables: */

static int qlevel = 0;			/* Shared by qbox_cu / qbox_cp 	*/

static double sqrtf2 = 0.7071067811865476;	/* sqrt(1/2) */
static double sqrtf3 = 0.5773502691896257;	/* sqrt(1/3) */

	/* Shared Variables for Selection */

typedef int (*FCT)();

#if INIT
static int qbox_init();
#endif
static int qbox_dflt();			/* Default Action Routine	*/
static int (*fct)() = qbox_dflt;	/* The Current Action Routine	*/

static double thecenter[3];		/* For Circle Selection		*/
static double theradius, s2radius;	/* s2radius = sin^2(r/2)	*/
static double lambda ;			/* Used by min_s2d		*/

static	double R1[3][3] = {		/* Unit Matrix	*/
	{1.,0.,0.},
	{0.,1.,0.},
	{0.,0.,1.}	};
static double *Rlon, *Rlat ;
static double llim[2], blim[2], zlim[2] ;
static double dl, tandl;
static int complement ;			/* Used for complementing Results */

#define if_not(cond)	if(!(cond))

/*===========================================================================*
 *		Internal Functions
 *===========================================================================*/

#define dotprod(a,b)   (a[0]*b[0]+a[1]*b[1]+a[2]*b[2])   /* Scalar product */
static void vecprod(double *v1, double *v2, double *r)
/*+++++++
.PURPOSE Vectorial product
.RETURNS Value
-------*/
{ 
    r[0] = v1[1]*v2[2] - v1[2]*v2[1];
    r[1] = v1[2]*v2[0] - v1[0]*v2[2];
    r[2] = v1[0]*v2[1] - v1[1]*v2[0];
}

static int qbox_dflt(qboxno, status)
/*+++++++
.PURPOSE Default function used when a qbox is found
.Returns status
.REMARKS Does nothing
-------*/
  int qboxno;	/* IN: The qbox number */
  int status;	/* IN: The status as QBOX_NONE / QBOX_ANY / QBOX_SOME */
{
    return(status);
}

static int corners(qboxno, qbox0, qbox1)
/*+++++++
.PURPOSE Compute the Corners of a Qbox from qbox number
.RETURNS The cube face number in range [1...6] / 0
.REMARKS qlevel is set.
-------*/
  int qboxno;		/* IN: The qbox number		*/
  double qbox0[2] ;	/* OUT: Box lower left corner 	*/
  double qbox1[2] ;	/* OUT: Box upper right corner 	*/
{
  int 	xy[2] ;
  double f ;

    Initialize;

    qlevel = -1 ;
    if (qboxno <= 0) return(0) ;

    xy[0] = xy[1] = 0 ;

    for (qlevel=0; qboxno&(~15); qlevel++) {
	xy[1] |= (qboxno & 1) << qlevel ;
	qboxno >>= 1 ;
	xy[0] |= (qboxno & 1) << qlevel ;
	qboxno >>= 1 ;
    }
    f = 1 << qlevel ;
    /* Lower  left corner */
    qbox0[0] = tan(((xy[0]/f) - 0.5e0) * (M_PI/2.));
    qbox0[1] = tan(((xy[1]/f) - 0.5e0) * (M_PI/2.));
    /* Upper right corner */
    xy[0] += 1 ; xy[1] += 1 ;
    qbox1[0] = tan(((xy[0]/f) - 0.5e0) * (M_PI/2.));
    qbox1[1] = tan(((xy[1]/f) - 0.5e0) * (M_PI/2.));

    return (qboxno&7);
}

static int ucorners(qboxno, uq)
/*+++++++
.PURPOSE Compute the Direction Cosines of the 4 Corners
.RETURNS Face number
-------*/
  int qboxno;		/* IN: The qbox number	*/
  double uq[4][3] ;	/* OUT: The cod.dir. of the 4 corners */
{
  double y, qbox0[2], qbox1[2] ;
  int plane ;
    plane = corners(qboxno, qbox0, qbox1) ;
    tr_pu6(qbox0, uq[0], plane);
    tr_pu6(qbox1, uq[2], plane);
    y = qbox0[1] ; qbox0[1] = qbox1[1] ; qbox1[1] = y ;
    tr_pu6(qbox0, uq[3], plane);
    tr_pu6(qbox1, uq[1], plane);
    { int i; printf("#...ucorners(%d=%s):\n", qboxno, qbox2a(qboxno));
	for(i=0; i<4; i++) printf("#   %+19.15f %+19.15f %+19.15f\n", 
		uq[i][0], uq[i][1], uq[i][2]);
    }
    return(plane);
}

#if INIT
/*===========================================================================
#ifndef tr_Pu
#define tr_Pu(a,b)	tr_pu6(a,b,1)
#endif
 *===========================================================================*/

static int qbox_init()
/*+++++++
.PURPOSE Initialize
.Returns 1
-------*/
{
  int 	i, j, n, nij;
  double p[2], p2[2];
  char	first_byte;

    if_not(qbox_init_done) {	/* Initialize */
	qbox_limit[0] = -1.e0;
	qbox_limit[1] = 1.e0; 	/* qbox level 0 	*/

	for (nij=1,n=0; n <= QBOX_LEVEL; n++) {	/* n is the dichotomy level   */
  	    for (j=nij; j>=1; j--)   	/* loop to move old limits    */
		qbox_limit[j<<1] = qbox_limit[j];
	    nij <<= 1; 			/* double number of limits    */

					/* compute centers            */
	    p[0] = sqrt(2.e0);
	    for (j=1; j<=nij; j+= 2) {
		p[1] = sqrt (1.e0 + qbox_limit[j+1] * qbox_limit[j+1]);
		qbox_limit[j] = (p[1]*qbox_limit[j-1] +
				 p[0]*qbox_limit[j+1])/(p[0]+p[1]);
		p[0] = p[1];
	    }
					/*     Estimate qboxes.radius */
	    p[0]  = qbox_limit[nij-1];	/* from center (n-1,n-1)      */
	    p[1]  = qbox_limit[nij-1];	/* to corner (n-2,n)          */
	    p2[0] = 1.e0;
	    p2[1] = qbox_limit[nij-2];
	    qboxes.radius[n] = dist_p(p, p2);

	    ijc = (nij>>1) ;
	    p[0]  = qbox_limit[nij-1];	/* from center (n-1,0)        */
	    p[1]  = qbox_limit[ijc+1];	/* to corner (n-2,n)          */
	    p2[0] = qbox_limit[nij-1];
	    p2[1] = qbox_limit[ijc];
	    qboxes.minrad[n] = dist_p(p, p2);
#if 0	/* Verification of the Minimal Radius 	*/
	    { int i, j; double r, rm; /* Compute all distances */
	      double u0[3], s2d, s2dmin, uq[4][3] ;
		s2dmin = 1. ;
		for (i=1, rm=90. ; i < nij; i+= 2) {
		    p[0] = qbox_limit[i] ;
		    for (j = 1; j <= i; j += 2) {
			p[1] = qbox_limit[j] ;
			tr_Pu(p, u0) ;
			printf("....i=%d,j=%d:", i, j) ;
			p2[0] = qbox_limit[i-1]; p2[1] = qbox_limit[j-1];
			    tr_Pu(p2, uq[0]) ;
			p2[0] = qbox_limit[i+1]; p2[1] = qbox_limit[j-1];
			    tr_Pu(p2, uq[1]) ;
			p2[0] = qbox_limit[i+1]; p2[1] = qbox_limit[j+1];
			    tr_Pu(p2, uq[2]) ;
			p2[0] = qbox_limit[i-1]; p2[1] = qbox_limit[j+1];
			    tr_Pu(p2, uq[3]) ;
			s2d = min_s2d(u0, uq[0], uq[1]);
			    if(s2d < s2dmin) s2dmin = s2d ;
			s2d = min_s2d(u0, uq[1], uq[2]);
			    if(s2d < s2dmin) s2dmin = s2d ;
			s2d = min_s2d(u0, uq[2], uq[3]);
			    if(s2d < s2dmin) s2dmin = s2d ;
			s2d = min_s2d(u0, uq[3], uq[0]);
			    if(s2d < s2dmin) s2dmin = s2d ;
			printf("[%.6f]", 2.*asind(sqrt(s2dmin))) ;
			p2[0] = p[0]; p2[1] = qbox_limit[j+1] ;
			printf(" %.4f", r = dist_p(p, p2)) ;
			if (r < rm) rm = r ;
			p2[0] = p[0]; p2[1] = qbox_limit[j-1] ;
			printf(" %.4f", r = dist_p(p, p2)) ;
			if (r < rm) rm = r ;
			p2[0] = qbox_limit[i-1]; p2[1] = p[1] ;
			printf(" %.4f", r = dist_p(p, p2)) ;
			if (r < rm) rm = r ;
			p2[0] = qbox_limit[i+1]; p2[1] = p[1] ;
			printf(" %.4f", r = dist_p(p, p2)) ;
			if (r < rm) rm = r ;
			printf("\n") ;
		    }
		}
		printf("----Chosen:  %.6f (ijc=%d) min=%.6f\n",
		    qboxes.minrad[n], ijc, rm) ;
	    }
#endif

	    if (ijc&1)	ijc--;
	    p[0]  = qbox_limit[ijc];	/* Center	*/
	    p[1]  = qbox_limit[ijc];
	    p2[0] = qbox_limit[ijc + 2];
	    p2[1] = qbox_limit[ijc + 2];
	    qbox_surf_max[n] = surf_p(p, p2);

	    p[0]  = 1;
	    p[1]  = qbox_limit[ijc + 2];	/* right	*/
	    p2[0] = qbox_limit[nij-2];
	    p2[1] = qbox_limit[ijc];
	    qbox_surf_min[n] = surf_p(p, p2);
	}
	qbox_init_done = 1;

	printf("    %c* Number   Lower Limit in X/Y \n", '/');
	for (i=0; i<= QBOX_AXIS; i++)    {
	    if (i < 10) goto show_xy;
	    if (i < (QBOX_AXIS/2)-5) {
		if (i == 10) printf("        [....]\n") ;
		continue;
	    }
	    if (i < (QBOX_AXIS/2)+5) goto show_xy;
	    if (i < QBOX_AXIS-9) continue;
	    if (i == QBOX_AXIS-9) printf("        [....]\n") ;
	    show_xy:
	    printf("       %6d %20.15f\n", i, qbox_limit[i]);
	}
	printf("    */\n");

	printf("static struct QBOX_def qboxes = { %d, %d\n",
	    QBOX_LEVEL, QBOX_LEVEL);

	for (i=0, first_byte='{'; i<= QBOX_LEVEL; i++, first_byte = ',')
	    printf("    %c %18.12f  %c* Level %2d:   radius %9.5f/%9.5f */\n",
			first_byte, qboxes.radius[i], '/', i,
			qboxes.radius[i], qboxes.minrad[i]);
	printf("    }, %c* Sine of above radii */\n", '/');

	for (i=0, first_byte='{'; i<= QBOX_LEVEL; i++, first_byte = ',')
	    qboxes.sinrad[i] = sind(	qboxes.radius[i]),
	    printf("    %c %18.16f  %c* Level %2d: Surface %10.3e /%10.3e */\n",
			first_byte, qboxes.sinrad[i], '/', i,
			qbox_surf_min[i], qbox_surf_max[i]);
	printf("    },\n") ;

	for (i=0, first_byte='{'; i<= QBOX_LEVEL; i++, first_byte = ',')
	    printf("    %c %18.12f  %c* Level %2d:   radius %9.5f/%9.5f */\n",
			first_byte, qboxes.minrad[i], '/', i,
			qboxes.radius[i], qboxes.minrad[i]);

#ifdef QLIMITS
	printf("    }, %c* Limits of Qboxes... */\n", '/');
	for (i=QBOX_AXIS/2, first_byte='{', j=0; i<= QBOX_AXIS ;
	    i+=2, j++, first_byte = ',')
	    qboxes.qbox_limit[j] = qbox_limit[i],
	    printf("    %c %18.12f  %c* n=%5d */\n",
			first_byte, qbox_limit[i], '/', j) ;
#endif
	printf("    } };\n");

  }

  return(1);
}
#endif

/*===========================================================================*
 *		Public Functions
 *===========================================================================*/

int qbox_get()
/*+++++++
.PURPOSE Get the current Qbox level
.Returns Current Qbox level
-------*/
{
    return(qboxes.level);
}

int qbox_set(lev)
/*+++++++
.PURPOSE Define the Qbox level
.Returns Previously defined qbox level
.REMARKS Limited in range [0,QBOX_LEVEL]
-------*/
  int	lev;	/* IN: The new asked qbox level	*/
{
  int 	old_lev;

    old_lev = qboxes.level;
    qboxes.level = lev;
    if ((lev < 0) || (lev > QBOX_LEVEL))
	qboxes.level = QBOX_LEVEL;
    return(old_lev);
}

char *qbox2a(qboxno)
/*+++++
.PURPOSE Edit qbox number
.RETURN Pointer to edited char (static buffer)
.REMARK Edit according to successive levels
-----*/
  int	qboxno;	/* IN: Qbox number to edit */
{
  static char buf[24];
  char *p ; int	level ;

    buf[0] = 0 ;
    level = qbox_level(qboxno);
    if (level < 0) return(buf);
    p = buf + level + 2 ;		/* Edit levels in base 4 */
    *p = 0 ;				/* Terminate the String  */
    while (level > 0) {
	*--p = '0' | (qboxno&3) ;
	level-- ;
	qboxno >>= 2 ;
    }
    *--p = '.' ;
    *--p = '0' | (qboxno&7) ;

    return(buf);
}

char *qbox8a(qboxno, buf)
/*+++++
.PURPOSE "Alphabetic" edition of a qbox
.RETURN Pointer to edited char (static buffer)
.REMARK buf is NULL ==> internal buffer
-----*/
  int	qboxno;	/* IN: Qbox number to edit */
  char *buf;	/* OUT: Buffer with edited value (min. 8 bytes) */
{
  static char buf9[9];
  unsigned int qb;
  int i, level;
    if (!buf) buf = buf9;	/* Edit in internal buffer */
    level = qbox_level(qboxno);
    if (level < 0) { buf[0] = 0; return(buf); }
    qb = (unsigned int)qboxno;
    i = (level>>1) + 1;		/* Length of edited qbox   */
    if (level&1) {		/* Last figure is [0-3]    */
	++i;
	if (i<8) buf[i] = 0;
	buf[--i] = '0' | (qb&3);
	qb >>= 2;
    }
    else {
	if (i<8) buf[i] = 0;
    }
    while (qb&(~0xf)) { buf[--i] = 'a' + (qb&15); qb >>= 4; }
    buf[0] = face_letter[qb&7];

    return(buf);
}

int qbox_a(char *str)
/*+++++
.PURPOSE Interpret a qbox number (either numeric or alphabetic)
.RETURN Value (0 for no interpretation)
.REMARK str is the result of qbox2a or qbox8a, e.g. 1.3120
-----*/
{
  int qb;
  char *p;

    p = str;
    while (*p == ' ') p++;
    if (! *p) return(0);

    if (isupper(*p)) {		/* V2.4: interpret alphabetic */
	for (qb=1; (face_letter[qb] != *p); qb++) {
	    if (face_letter[qb] == 0)	/* Invalid */
		return(0);
	}
	qb |= 8;
	for (++p; isgraph(*p); p++) {
	    if (islower(*p))      { qb <<= 4; qb |= *p - 'a'; }
	    else if (isdigit(*p)) { qb <<= 2; qb |= *p - '0'; }
	    else break;
	}
	return(qb);
    }

    /* Numeric */
    if (isdigit(p[1])) {
	qb = atoi(p);
	while(isdigit(*p)) p++;
    }
    else {
        qb = *(p++) - '0';
        if ((qb<1) || (qb>6))
	    return(-1);
        if (*(p++) != '.')
	    return(qb);
        qb |= 8;
        while ((*p >= '0') && (*p < '4')) {
	    qb = (qb<<2) | (*p&3);
	    p++;
	}
    }

    return(qb);
}

/*===========================================================================*/

int qbox_level (qboxno)
/*+++++++
.PURPOSE Extract the Level part from a Qbox number
.RETURNS The level number in range [0...QBOX_LEVEL] 
	/ negative number if bad qbox number
.METHOD  Use shifts.
-------*/
  int qboxno;	/* IN: The qbox number	*/
{
  int 	i;
  unsigned int n;
    /* for (i = -1, n = ((unsigned)qboxno) >> 2; n; n >>= 2) i++; */
    for (i = 0, n = (unsigned)qboxno; n & (~15) ; n >>= 2) i++;
    if ((n >= 9) && (n <= 14) && (i <= QBOX_LEVEL)) return(i) ;
    return(-1) ;
}

int qbox_face (qboxno)
/*+++++++
.PURPOSE Extract the Face number from a Qbox number
.Returns The Face number in range [1...6] / 0 if bad qbox number
.METHOD  Use shifts.
-------*/
   int qboxno;	/* IN: The qbox number	*/
{
  unsigned int n;
    for (n = (unsigned)qboxno; n & (~15); n >>= 2) ;
    return(n&7);
}

int qbox_lower (qboxno)
/*+++++++
.PURPOSE Find what is the smallest box number
.Returns The lower (level<<(level<<1)). Use current level if qboxno = 0.
-------*/
   int qboxno;	/* IN: The qbox number	*/
{
  int lev;
    lev = qbox_level(qboxno);
    if (lev<0) lev = qboxes.level;
    return(9<<(lev<<1));
}

int qbox_upper (qboxno)
/*+++++++
.PURPOSE Find what is the smallest box number
.Returns The lower (level<<(level<<1)). Use current level if qboxno = 0.
-------*/
   int qboxno;	/* IN: The qbox number	*/
{
  int lev;
    lev = qbox_level(qboxno);
    if (lev<0) lev = qboxes.level;
    return(((unsigned)(15<<(lev<<1)))-1);
}

/*===========================================================================*/

int qbox_o(o)
/*+++++++
.PURPOSE Compute the Qbox number from Position
.Returns The Qbox Number
.METHOD	 Use Unit vector
-------*/
  double o[2];	/* IN: Position (degrees)	*/
{
  double	u[3];
    tr_ou(o, u);
    return(qbox_u(u));
}

int qbox_u(u)
/*+++++++
.PURPOSE Compute the Qbox number from Unit Vector
.Returns The Qbox Number / 0 if Error
.REMARK  The current level (defined via qbox_set) is used.
-------*/
  double u[3];	/* IN: Unit vector used for Qbox */
{
  int i; 
  unsigned int n;
  double p[2];

    Initialize;

    i = tr_up6(u, p, 0);	/* Basic 1...6 number */
    n = (unsigned)i | 8;	/* Insert bit to indicate level */

#ifndef QLIMITS 	/* Compute x,y with ATAN */

    if (qboxes.level) {
	n = qbox__u(qboxes.level, u);
    }
#else
  int i, x, y, count;
  unsigned int 	n;
  double xy, p[2];
    /* Estimate sign */
    if (qboxes.level) {
	n <<= 2 ;
	if (p[0] >= 0) n |= 1, sgn[0] = 0 ; else p[0] = -p[0], sgn[0] = 1;
	if (p[1] >= 0) n |= 2, sgn[1] = 0 ; else p[1] = -p[1], sgn[1] = 1;
	i = 1<<(qboxes.level-1) ;
	low[0] = low[1] = qboxes.qbox_limit ;
	while (i & (~1)) {
	    n <<= 2 ;  i >>= 1;
	    if (sgn[0]) {	/* Negative x */
		if (p[0] <= low[0][i]) n |= 2 ;
		else low[0] += i ;
	    }
	    else {		/* Positive x  */
		if (p[0] <= low[0][i]) ;
		else n |= 2, low[0] += i ;
	    }
	    if (sgn[1]) {	/* Negative y */
		if (p[1] <= low[1][i]) n |= 1 ;
		else low[1] += i ;
	    }
	    else {		/* Positive y  */
		if (p[1] <= low[1][i]) ;
		else n |= 1, low[1] += i ;
	    }
	}
    }
#endif

   return(n);
}

int qbox__o(lev, o)
/*+++++++
.PURPOSE Compute the Qbox number from Level + Position
.Returns The Qbox Number
.METHOD	 Use Unit vector
-------*/
  int lev;	/* IN: Level in range 0..12	*/
  double o[2];	/* IN: Position (degrees)	*/
{
  double	u[3];
    tr_ou(o, u);
    return(qbox__u(lev, u));
}

int qbox__u(lev, u)
/*+++++++
.PURPOSE Compute the Qbox number from Unit Vector
.Returns The Qbox Number / 0 if Error
.REMARK  The current level (defined via qbox_set) is used.
-------*/
  int lev;	/* IN: Level in range 0..12	*/
  double u[3];	/* IN: Unit vector used for Qbox */
{
  int i, x, y, count;
  unsigned int 	n;
  double xy, p[2];

    Initialize;

#ifndef QLIMITS 	/* Compute x,y with ATAN */
    i = tr_up6(u, p, 0);	/* Basic 1...6 number */
    n = (unsigned)i | 8;	/* Insert bit to indicate level */

    if ((lev < 0) || (lev > QBOX_LEVEL))	/* Bad level */
	return(0);		/* ERROR */

    if (qboxes.level) {
        count = 1 << lev;
        xy = (.5e0 + (2./M_PI)*atan(p[0])) * count;
	x = xy ; if (x >= count) x = count-1 ;
        xy = (.5e0 + (2./M_PI)*atan(p[1])) * count;
	y = xy ; if (y >= count) y = count-1 ;
	n <<= (lev << 1) ;
	for (count >>= 1, i = lev-1 ; count ; count >>= 1, i-- ) {
	    n |= (((x&count)<<i)<<1) | ((y&count)<<i) ;
	}
    }
#else
    fprintf(stderr, "****qbox__u: QLIMITS shouldn't be defined...\n");
    return(0);
#endif

   return(n);
}

/*===========================================================================*/

int qbox_co (qboxno, o)
/*+++++++
.PURPOSE Compute the Center of a Qbox from qbox number
.Returns The level in range [0...QBOX_LEVEL] / -1 if qboxno is not a
		valid qbox number
.METHOD  Use unit vectors.
-------*/
  int qboxno;	/* IN: The qbox number	*/
  double o[2];	/* OUT: The center of the Qbox (degrees)	*/
{
  double u[3];
    qbox_cu(qboxno, u) ;
    tr_uo(u, o);
    return(qlevel);
}

int qbox_cu (qboxno, u)
/*+++++++
.PURPOSE Compute the Center of a Qbox from qbox number
.Returns The level in range [0...QBOX_LEVEL] / -1 if qboxno is not a
		valid qbox number
.REMARK
-------*/
  int qboxno;	/* IN: The qbox number	*/
  double u[3];	/* OUT: The center of the Qbox	*/
{
  int 	plane;
  double p[2];
    if ((plane = qbox_cp(qboxno, p)))
	 tr_pu6(p, u, plane);
    else u[0] = 0, u[1] = 0, u[2] = 1;	/* Choose North Pole */
    return(qlevel);
}

int qbox_cp (qboxno, p)
/*+++++++
.PURPOSE Compute the Center of a Qbox from qbox number
.Returns The cube face number in range [1...6] / 0
		if qboxno is not a valid qbox number
-------*/
  int qboxno;	/* IN: The qbox number	*/
  double p[2];	/* OUT: The center	*/
{
  int 	xy[2] ;
  double f ;

    Initialize;

    qlevel = -1 ;
    if (qboxno <= 0) return(0) ;

    xy[0] = xy[1] = 0 ;

    for (qlevel=0; qboxno&(~15); qlevel++) {
	xy[1] |= (qboxno & 1) << qlevel ;
	qboxno >>= 1 ;
	xy[0] |= (qboxno & 1) << qlevel ;
	qboxno >>= 1 ;
    }
    f = 1 << qlevel ;

    p[0] = tan((((xy[0]+0.5)/f) - 0.5e0) * (M_PI/2.));
    p[1] = tan((((xy[1]+0.5)/f) - 0.5e0) * (M_PI/2.));

    return (qboxno&7);
}

int qbox_limp (qboxno, pmin, pmax)
/*+++++++
.PURPOSE Compute the limits on the projections
.Returns The cube face number in range [1...6] / 0
		if qboxno is not a valid qbox number
.REMARKS  NULL pointers => not computed
-------*/
  int qboxno;	/* IN: The qbox number	*/
  double *pmin;	/* OUT: Minimal values  */
  double *pmax;	/* OUT: Minimal values  */
{
  double qbox0[2], qbox1[2] ;
  int face ;
    Initialize;
    face = corners(qboxno, qbox0, qbox1) ;
    if (pmin) memcpy(pmin, qbox0, 2*sizeof(double));
    if (pmax) memcpy(pmax, qbox1, 2*sizeof(double));
    return(face);
}

int qbox_limu (qboxno, umin, umax)
/*+++++++
.PURPOSE Compute the limits on direction cosines
.Returns The cube face number in range [1...6] / 0
		if qboxno is not a valid qbox number
.REMARKS  NULL pointers => not computed
	Note that umin and umax are NOT direction cosines,
	but (xmin, ymin, zmin)
-------*/
  int qboxno;	/* IN: The qbox number	*/
  double *umin;	/* OUT: Minimal values  */
  double *umax;	/* OUT: Minimal values  */
{
  double uq[4][3], x;
  int face, i, j;
    Initialize;
    if (qboxno <= 0) return(0);
    if ((qboxno&(~15)) == 0) {	/* Special, level = 0 */
	face = qboxno&7;
	uq[0][0] = uq[0][1] = uq[0][2] = -sqrtf2;
	uq[1][0] = uq[1][1] = uq[1][2] =  sqrtf2;
	if (face&4) {
	    uq[0][4^face] = -1;
	    uq[1][4^face] = -sqrtf3;
	}
	else {
	    uq[0][3^face] = sqrtf3;
	    uq[1][3^face] = 1.;
	}
    }
    else {			/* Take min/max of corners */
	face = ucorners(qboxno, uq);
	for (i=1; i<4; i++) for (j=0; j<3; j++) {
	    if (uq[i][j] < uq[0][j]) {
		x = uq[0][j]; uq[0][j] = uq[i][j]; uq[i][j] = x;
	    }
	    if (uq[i][j] > uq[1][j]) {
		x = uq[1][j]; uq[1][j] = uq[i][j]; uq[i][j] = x;
	    }
	}

    }
    if (umin) memcpy(umin, &uq[0][0], 3*sizeof(double));
    if (umax) memcpy(umax, &uq[1][0], 3*sizeof(double));
    return(face);
}

double qbox_surf (qboxno)
/*+++++++
.PURPOSE Compute the Surface of a Qbox
.RETURNS The surface in square degrees
-------*/
  int qboxno;	/* IN: The qbox number	*/
{
  double qbox0[2], qbox1[2] ;
    Initialize;
    if (qboxno <= 0) return((360.*360./M_PI)) ;	/* Full Sphere */
    corners(qboxno, qbox0, qbox1) ;
    return (surf_p(qbox0, qbox1)) ;
}

double qbox_radius(qboxno)
/*+++++++
.PURPOSE Give the Radius of a Qbox
.rETURNS The radius (in degrees)  / 0 if qboxno is not a valid qbox number
-------*/
  int qboxno;	/* IN: The qbox number	*/
{
    Initialize;
    if (qboxno <= 0) return(180.) ;		/* Full Sphere 	   */
#if 0
    corners(qboxno, qbox0, qbox1) ;		/* Compute corners */
    qbox1[0] = (qbox0[0] + qbox1[0]) / 2.;
    qbox1[1] = (qbox0[1] + qbox1[1]) / 2.;
    return(dist_p(qbox0, qbox1)) ;
#else
    return(qboxes.radius[qbox_level(qboxno)]) ;
#endif
}

double qbox_size(qboxno)
/*+++++++
.PURPOSE Give the Size (square side) of a Qbox
.rETURNS The size (in degrees)  / 0 if qboxno is not a valid qbox number
-------*/
  int qboxno;	/* IN: The qbox number	*/
{
  int level;
  double f;
    level = qbox_level(qboxno);
    if (level<0)  return(0);
    if (level==0) return(90.);
    f = (1<<level);
    return(atand(2./f));
}

/*===========================================================================*
 *		Adjacent boxes
 *===========================================================================*/

int qbox_adj(int qboxno, int dir)
/*+++++++
.PURPOSE Move to the nearest qbox
.RETURNS The adjacent in direcretion 0=+x, 1=+y, 2=-y, 3=-x
.REMARKS Problems at limits...
---------*/
{
  static char move_x[4] = { 1, 0, 0, 1 }; /* 1 for movements on x */
  int k, qb, r, shifted, comb;

    r = shifted = 0;
    qb = qboxno;
    dir &= 3;				/* Only 4 directions */
    while (qb&(~15)) {
	k = adj_move[dir][qb&3];
	if (k&4) {			/* Not possible at this level */
	    r |= (k^4)<<shifted;
	    qb >>= 2;
	    shifted += 2;
	}
	else {
	    qb = (qb&(~3))|k;
	    break;
	}
    }

    if ((qb&(~15)) == 0) {		/* Here at level=0 */
	k = adj_face[dir][qb&7];
	/* In comes cases, a rotation occurs, indicated by
	 * the lask 8 in the adj_face table
	 */
	if (k&8) {
	    /* printf("....must change r...."); */
	    for (qb=shifted, comb=0; qb>0; qb -= 2) comb = (comb<<2) | 2;
#if 1	/* Compact code */
	    if (move_x[dir]) comb >>= 1;
	    r = (qboxno&comb)^comb;	/* r contains -y bits */
	    if (move_x[dir]) r <<= 1;
	    else             r >>= 1;
	    if (dir&2) r |= comb;	/* set x|y to maximum */
#else	/* Details for each direction */
	    if (dir == 0) {		/* Move +x */
		comb >>= 1;
		r = (qboxno&comb)^comb;	/* r contains -y bits */
		r <<= 1;		/* change y -> x bits */
	    }
	    else if (dir == 1) {	/* Move +y */
		r = (qboxno&comb)^comb;	/* r contains -x bits */
	        r >>= 1;		/* change x -> y bits */
	    }
	    else if (dir == 2) {	/* Move -y */
		r = (qboxno&comb)^comb; /* r contains -x bits */
		r >>= 1;		/* change x -> y bits */
		r |= comb;		/* set x to maximum   */
	    }
	    else if (dir == 3) {	/* Move -x */
		comb >>= 1;
		r = (qboxno&comb)^comb;	/* r contains -y bits */
		r <<= 1;		/* change y -> x bits */
		r |= comb;		/* Set y to maximum   */
	    }
#endif
	}
	qb = 8|k;
    }

    if (shifted) {
	qb <<= shifted;
	qb |= r;
    }

    return(qb);
}

int qbox_nearby(int qboxno, int *nearby)
/*+++++++
.PURPOSE Find the 8 nearby boxes
.RETURNS Number of boxes (8, or 7 in the corners)
.REMARKS Order of the computed nearby boxes:
    +---+---+---+
    | 7 | 1 | 4 |
    +---+---+---+
    | 3 | . | 0 |
    +---+---+---+
    | 6 | 2 | 5 |
    +---+---+---+
---------*/
{
  int i, qbmax, nb=4;
  char same_face[4];	/* Set to 1 for nearby boxes on same facet *
          * A simple test to be sure 2 boxes are on the same facet
          * is   (b1^b2)/max(b1,b2) >= 1 */

    /* Directly adjacent: straightforward */
    printf("....qbox_nearby=%s \n", qbox2a(qboxno));
    qbmax = 1 << (qbox_level(qboxno)<<1);
    for (i=0; i<4; i++) {
	nearby[i] = qbox_adj(qboxno, i);
	same_face[i] = (nearby[i] ^ qboxno) < qbmax;
			/* TRUE (1) only if same facet */
	printf("    nearby[%d]=%s, same_face[%d]=%d\n", 
		i, qbox2a(nearby[i]), i, same_face[i]);
    }
    /* Explore how to deal with corners :
     * set same_face[i] = 1 or 2 to explore from there in y or x direction.
     */
    nearby[4] = nearby[5] = nearby[6] = nearby[7] = 0;
    if (same_face[0]) {
	nearby[4] = qbox_adj(nearby[0], 1);
	nearby[5] = qbox_adj(nearby[0], 2);
	nb += 2;
    }
    if (same_face[3]) {
	nearby[6] = qbox_adj(nearby[3], 1);
	nearby[7] = qbox_adj(nearby[3], 2);
	nb += 2;
    }
    if (nb>=7) return(nb);

    if (same_face[1]) {
	if (!nearby[4]) nearby[4] = qbox_adj(nearby[1], 0), nb++;    /* +x */
	if (!nearby[7]) nearby[7] = qbox_adj(nearby[1], 3), nb++;    /* -x */
    }
    if (nb>=7) return(nb);

    if (same_face[2]) {
	if (!nearby[5]) nearby[5] = qbox_adj(nearby[2], 0), nb++;    /* +x */
	if (!nearby[6]) nearby[6] = qbox_adj(nearby[2], 3), nb++;    /* -x */
    }
    return(nb);
}

int qbox_around(double *uc, int *contig, double *dist2)
/*+++++++
.PURPOSE Find the qboxes contiguous to a position, ordered by increasing
		distance to the central position.
.RETURNS Number of boxes (9, 8 or 5)
.REMARKS dist2 contains in return the values of (2 sin(r/2))^2, ordered
	by increasing distance. Pointer may be null, and then distances
	are not giben.
---------*/
{
  static char nearby_xy[] = {	/* Neighbour coordinate indices */
      025, 043, 041, 005, 023, 021, 001, 003 };
  double sin2d2[9], coin[4], cXY[2], XY[2], u[3], du, w; 
  int qbox, face, nb, i, ix, iy;
    
    qbox = qbox_u(uc);
    face = corners(qbox, coin, coin+2);
    nb = qbox_nearby(qbox, contig);
    tr_up6 ( uc , cXY , face );		/* Get projection of center */

    /* Compute the distances (4sin^2(r/2)) of 8 neighbors */
    for (i=0; i<8; i++) {
	if (contig[i] == 0) {		/* in a corner of the cube */
	    sin2d2[i] = 100.;		/* Very large to put at end */
	    continue;
	}
	ix = nearby_xy[i]>>3;		/* 1st octal index */
	iy = nearby_xy[i] &7; 		/* 2nd octal index */
	XY[0] = (ix&4)==0 ? coin[ix] : cXY[ix^4];
	XY[1] = (iy&4)==0 ? coin[iy] : cXY[iy^4];
	tr_pu6(XY, u, face);
	du = (u[0] - uc[0]); w  = du*du;
	du = (u[1] - uc[1]); w += du*du;
	du = (u[2] - uc[2]); w += du*du;
	sin2d2[i] = w;
    }

    /* Simple swap sort */
    for (ix=0; ix<7; ix++) for (iy=ix+1; iy<8; iy++) {
	if (sin2d2[ix] <= sin2d2[iy]) continue;
	/* swap distances & qbox# */
	w = sin2d2[ix]; sin2d2[ix] = sin2d2[iy]; sin2d2[iy] = w;
	i = contig[ix]; contig[ix] = contig[iy]; contig[iy] = i;
    }

    /* Move qbox as the first */
    for (i=8; i>0; i--) {
	contig[i] = contig[i-1];
	sin2d2[i] = sin2d2[i-1];
    }
    contig[0] = qbox;
    sin2d2[i] = 0;

    /* Copy the distances */
    if (dist2) memcpy(dist2, sin2d2, 9*sizeof(double));

    return(nb+1);		/* The addition qbox including the target */
}

/*===========================================================================*/

static int call_down(qboxno, status, cf)
/*+++++++
.PURPOSE Move down the tree of qboxes, using the results of a check function
.RETURNS 0
.REMARKS The Action (defined by qbox_act) is called:
	 - either at the leaf level
	 - or at a node when no match can exist (QBOX_NONE returned
	   by the check function cf)
	The check function is of type 	int (*cf)(int qboxno)
	and must return QBOX_ANY / QBOX_NONE / QBOX_SOME
-------*/
  int qboxno;	/* IN: The qbox number 	*/
  int status;	/* IN: NONE/ANY/SOME	*/
  FCT cf;	/* IN: Check function 	*/
{
  int n, level ;
  FCT f ;

#if INIT
    printf("....call_down(%-8s, %d)\n", qbox2a(qboxno), status) ;
#endif

    /* At leaf level ? Call then action !	*/
    if ((unsigned)qboxno >= (unsigned)((9 << qboxes.level) << qboxes.level))
	return((*fct)(qboxno, status)) ;

    /* Zero status ?  Stop at whatever node...	*/
    if (!status)
	return((*fct)(qboxno, status)) ;

    level = qbox_level(qboxno) ;
    if (level < 0) {		/* Whole sky...	*/
	for (n=9; n<=14; n++) call_down(n, (*cf)(n, status), cf) ;
	return(0) ;
    }

    /* Recursively call the check function...	*/
    f = status == QBOX_ANY ? qbox_dflt : cf ;
    n = qboxno << 2 ;
    call_down(n,   (*f)(n,   status), f) ;  	/* Quarter left  / down */
    call_down(n|1, (*f)(n|1, status), f) ;  	/* Quarter right / down */
    call_down(n|2, (*f)(n|2, status), f) ;  	/* Quarter left  / up   */
    call_down(n|3, (*f)(n|3, status), f) ;  	/* Quarter right / up   */
    return (0);
}

int qbox_next(qboxno)
/*+++++++
.PURPOSE Convert a qbox number to the current level (set by qbox_set).
.RETURNS The converted number, 0 when no more qbox number is available.
-------*/
  int qboxno;	/* IN: The qbox number (init) or 0 (next number) */
{
  static int the_qbox;
  static int max_qbox;
  int ix, iy ;

    if (qboxno) {		/* Initialize the values */
	max_qbox = the_qbox = 0;	/* Default Value */
	ix = qbox_level(qboxno);	/* Current Level */
	iy = qboxes.level - ix;		/* Suppl. bits   */
	if ((ix >= 0) && (iy >= 0)) {
	    the_qbox = qboxno << (iy << 1) ;
	    max_qbox = the_qbox + (1<<(iy << 1)) ;
	}
	else return(0) ;
   }
   				/* Compute Next Qbox	 */
    if (the_qbox >= max_qbox) return(0) ;
    ix = the_qbox ;
    ++the_qbox ;
    return(ix) ;
}

/*===========================================================================*
 *			Selection of Qbox
 *===========================================================================*/

FCT qbox_act(f)
/*+++++++
.PURPOSE Define the `action routine', i.e. function to be called during
	qbox selection.
.Returns The old function
-------*/
  FCT f;	/* IN: The function to install. Arguments (qboxno qboxestatus) */
{
  FCT	old_f;
    old_f = fct;
    fct = (f ? f : qbox_dflt);
    return(old_f);
}

/*===========================================================================*
 *		Check a Circle
 *===========================================================================*/

static double min_prod(u0, u1, u2, a)
/*+++++++
.PURPOSE Compute the vector which maximizes (or minimizes)
	the scalar product between u0 and a vector
	located between u1 and u2.
.RETURNS The value of lambda such that  a = u1 + lambda u2
-------*/
  double u0[3];	/* IN: Center */
  double u1[3];	/* IN: 1st Vertice */
  double u2[3];	/* IN: 1nd Vertice */
  double  a[3];	/* OUT: Result vet */
{
  double lambda, den ;
    a[0] = u1[0]*u2[0] + u1[1]*u2[1] + u1[2]*u2[2] ;
    a[1] = u0[0]*u1[0] + u0[1]*u1[1] + u0[2]*u1[2] ;
    a[2] = u0[0]*u2[0] + u0[1]*u2[1] + u0[2]*u2[2] ;
    lambda = (a[1]*a[0] - a[2])/(a[2]*a[0] - a[1]) ;

    /* Compute the Vector */
    den = sqrt(1. + lambda*lambda + 2.*lambda*a[0]) ;
    a[0] = (u1[0] + lambda*u2[0])/den ;
    a[1] = (u1[1] + lambda*u2[1])/den ;
    a[2] = (u1[2] + lambda*u2[2])/den ;

    return(lambda) ;
}

#if 0
static double min_s2d(u0, u1, u2)
/*+++++++
.PURPOSE Compute the minimal distance between u0 and the arc (u1,u2)
.RETURNS The minimal distance occurs within the arc,
           and -1 id the distance is outside (v1,v2)
.REMARKS The value of lambda is set.
.METHOD  We use a vector
\begin{TeX}
$\vec{u} = \vec{u_1} + lambda vec{u_2}$, and we look for
$\lambda$ gives extremum of the function
$$ \frac{\vec{u_1}.\vec{u_0}+\lambda\vec{u_2}.\vec{u_0}}
        {sqrt{1+\lambda^2 + 2\lambda\vec{u_1}.\vec{u_2}}}
\end{TeX}
-------*/
  double u0[3];	/* IN: Center */
  double u1[3];	/* IN: 1st Vertice */
  double u2[3];	/* IN: 1nd Vertice */
{
  double a[3] ;
    lambda = min_prod(u0, u1, u2, a) ;

    /* Compute the Minimal Distance */
    return(s2d_u(u0, a)) ;
}
#endif

static int check_radius(qboxno, status)
/*+++++++
.PURPOSE Checks the qboxes which must be searched for a target
.RETURNS QBOX_NONE / QBOX_ANY / QBOX_SOME
.REMARKS Recursive
.METHOD  The function fct is called for matching qboxes.
-------*/
  int qboxno;	/* IN: The qbox number */
  int status ;	/* IN: e.g. QBOX_SOME  */
{
  double uc[3], d, uq[4][3], r, v[3], qbox0[2], qbox1[2], s2r;
  int level, plane, i, inside;

    if (theradius >= 180.e0)	return(QBOX_ANY);

    level = qbox_cu(qboxno, uc);
    d = dist_u(uc, thecenter);

    if (d >= (qboxes.radius[level] + theradius))/* No point belongs to circle */
  	return(QBOX_NONE);

    if (theradius >= (qboxes.radius[level] + d))/* All points belong to circle*/
  	return(QBOX_ANY);

    /* if (level < 2) return(QBOX_SOME) ; */
    /* Check now more accurately: can the box be completely included ?
       Easy: if all corners are inside the circle ...
    */
    inside = 0;
    plane = corners(qboxno, qbox0, qbox1) ;
    tr_pu6(qbox0, uc, plane);
    if (s2d_u(uc, thecenter) <= s2radius) inside++;
    tr_pu6(qbox1, uc, plane);
    if (s2d_u(uc, thecenter) <= s2radius) inside++;
    /* Compute the other 2 corners */
    r = qbox0[1] ; qbox0[1] = qbox1[1] ; qbox1[1] = r ;
    tr_pu6(qbox0, uc, plane);
    if (s2d_u(uc, thecenter) <= s2radius) inside++;
    tr_pu6(qbox1, uc, plane);
    if (s2d_u(uc, thecenter) <= s2radius) inside++;
    printf("#...check_radius(%d=%s[%s]) => inside=%d\n", 
	    qboxno, qbox2a(qboxno), qbox8a(qboxno,0), inside);
    if (inside == 4) return(QBOX_ANY) ;
    if (inside >  0) return(QBOX_SOME) ;
    
    /* No corner of Qbox inside circle.
     * Look for possible intersection of the circle with 
     * any side of the Qbox:
     * -> at least one intersection => overlapping
     * -> no intersection: overlapping only if circle center inside the box.
     The distance between the center of the circle and a Qbox side 
     is computed from vector product if (a,b) = vertices and C = center
          sin(dist) = (a^b).C/||a^b||
     */

    s2r = 4.*s2radius - s2radius*s2radius;   /* sin^2r */
    ucorners(qboxno, uq) ;
    uq[4][0] = uq[0][0];
    uq[4][1] = uq[0][1];
    uq[4][2] = uq[0][2];
    for (i=0; i<4; i++) {
	vecprod(uq[i], uq[i+1], v);
	r = dotprod(thecenter, v);   /* sin(v1,v2) * sin(distance to circle) */
	if (r >= 0) inside++;
	printf("#   i=%d r=%-16g ||v||=%g\n", i, r, dotprod(v,v));
	if (r*r <= s2r*dotprod(v,v)) return(QBOX_SOME);
    }
    return(inside == 4 ? QBOX_SOME : QBOX_NONE);
}

int qbox_or(o, radius)
/*+++++++
.PURPOSE Select the qboxes which match a small circle centered at o
.Returns 0
.METHOD  Use unit vectors.
.REMARKS The `action routine' (see qbox_act) is called to gather the results.
-------*/
  double o[2];	/* OUT: Center of the Circle (degrees)	*/
  double radius;	/* IN: Radius of the circle (degrees)	*/
{
  double u[3];
    tr_ou(o, u);
    return(qbox_ur(u, radius));
}

int qbox_ur(u, radius)
/*+++++++
.PURPOSE Select the qboxes which match a small circle.
.Returns 0
.REMARKS The `action routine' (see qbox_act) is called to gather the results.
.METHOD  Compare distance of qbox center to radii.
-------*/
  double u[3];	/* IN: Position of the center of the circle */
  double radius;	/* IN: Radius of the circle (degrees)	*/
{
  int 	i;

    for (i=0; i<3; i++)	thecenter[i] = u[i];	/* Local copy 	*/
    theradius = radius;
    s2radius = sind(radius/2.); s2radius *= s2radius ;

    call_down(0, QBOX_SOME, check_radius) ;

    return(0);
}

/*===========================================================================*
 *		Check Limits in RA / Dec / Lon / Lat
 *===========================================================================*/

static int check_lon(qboxno, status)
/*+++++++
.PURPOSE Checks the qboxes which must be searched for a target
.RETURNS QBOX_NONE / QBOX_ANY / QBOX_SOME
.REMARKS Interval aussed to be less than 90 degrees.
-------*/
  int qboxno;	/* IN: The qbox number */
  int status ;	/* IN: e.g. QBOX_SOME  */
{
  double uc[3], uq[4][3], x, r, L, psi;
  int level, domain, inside, i, j ;

    if (dl <= 0.)	return(complement/*^QBOX_NONE*/);

    level = qbox_cu(qboxno, uc);
    tr_uu(uc, uc, Rlon); 	/* Express qbox center in Local */
    r = sqrt(uc[0]*uc[0] + uc[1]*uc[1]) ;	/* x^2 + y^2	*/

    if (qboxes.sinrad[level] <= r) {	/* Pole not in Qbox	*/
	psi = asind(qboxes.sinrad[level] / r);
	L = fabs(atan2d(uc[1], uc[0]));

	if ((L-psi) >= dl)	/* No point belongs to range 	*/
  	    return(complement/*^QBOX_NONE*/);

	if ((L+psi) <= dl)	/* All points belong to range   */
  	    return(complement^QBOX_ANY);
    }

    if (level < 2)	return(QBOX_SOME) ;

    /* Verify more closely whether the region is concerned or not */
    ucorners(qboxno, uq) ;
    for (i=0; i<4; i++) tr_uu(uq[i], uq[i], Rlon) ;
    domain = inside = 0 ;
    /* printf("....%12s: Verify Corners, r=%.8f\n", qbox2a(qboxno), r) ; */

    /* When pole outside our domain, just use the projection
       on the plane x=1: the valid domain is between
       the vertical lines  y = +/-tan(dl)
    */
    if (qboxes.sinrad[level] <= r) {	/* Pole not in Qbox     */
        for (i=0; i<4; i++) {
	    if (uq[i][0] <= 0) domain += (uq[i][1] > 0 ? +1 : -1) ;
	    else if ((x = uq[i][1]/uq[i][0]) > tandl) domain++ ;
	    else if (x < -tandl) domain-- ;
	    else inside++ ;
	}
        if (inside == 4) return(complement^QBOX_ANY);
        if ((domain == 4) || (domain == -4))
	    return(complement/*^QBOX_NONE*/);
    }
    /* The test of pole inside the domain uses the projection
       onto the plane z=1. The valid domain is then the angle
       between +/-dl
    */
    else {				/* Pole really inside ? */
        for (i=0; i<4; i++) 		/* Standard Projection 	*/
	   uq[i][0] /= (x = fabs(uq[i][2])), uq[i][1] /= x ;
	/* Verify the pole is in, with the classical test	*/
        for (i=0; i<4; i++) {
	    j = (i+1)&3 ;
	    x = uq[i][0] * uq[j][1] - uq[i][1] * uq[j][0] ;
	    if (x == 0) 		/* Pole on the polygon	*/
	       return(QBOX_SOME) ;
	    if (x>0) domain++;
	    else domain-- ;
	    if (fabs(uq[i][1]) <= uq[i][0] * tandl) inside++ ;
	}
	/* qbox_co(qboxno, uc) ;
	   printf("....%12s: %010.6f%+010.6f domain=%d, inside=%d\n",
	    qbox2a(qboxno), uc[0], uc[1], domain, inside) ;
	*/
	if ((domain == 4) || (domain == -4)) 	/* Pole inside	*/
	    return(QBOX_SOME) ;
	if (inside == 4) return(complement^QBOX_ANY);
	if (inside == 0) return(complement/*^QBOX_NONE*/);
    }

    return(QBOX_SOME) ;
}

int qbox_ra(lim)
/*+++++++
.PURPOSE Select the qboxes which match a selection on ra
.Returns 0
.REMARKS The `action routine' (see qbox_act) is called to gather the results.
.METHOD  Use qbox_lon with unit matrix.
-----------*/
  double lim[2];	/* IN: Limit in Longitudes (the order is important!) */
{
    return(qbox_lon(lim,R1));
}

int qbox_lon(lim, R)
/*+++++++
.PURPOSE Select the qboxes which match a selection on ra / longitude
.Returns 0
.REMARKS The `action routine' (see qbox_act) is called to gather the results.
.METHOD
\begin{TeX}
The qbox is taken as a circle of radius $\rho$, centered at $(L,\phi)$
in the local frame with $x$--axis in the middle of the longitude range
($\frac{1}{2}(l_0+l_1)$). The limiting circle of the qbox projects onto the
$(x,y)$ plane as an ellipse centered at a distance $\cos\phi\cos\rho$
from center, with semi-axises $(\sin\rho , \sin\rho \sin\phi)$.
This ellipse is included bewtween its tangents with polar
angles $L\pm\psi$, where $\psi = \frac{\sin\rho}{\cos\phi}$.
\end{TeX}
-------*/
  double lim[2];	/* IN: Limit in Longitudes (the order is important!) */
  double R[3][3];	/* IN: Rotation matrix expressing longitude / latitude*/
{
  double o[2], Rloc[3][3];

    llim[0] = lim[0], llim[1] = lim[1];		/* Local Copy 	*/
    if (llim[0] > llim[1]) llim[1] += 360.e0;	/* Verify Range */

	/* Build the Rloc matrix */
    dl = (llim[1] - llim[0])/2.e0;
    o[0] = (llim[0] + llim[1])/2.e0;	o[1] = 0.e0;
    /* For large intervals, use complement */
    complement = 0 ;
    if (dl > 90.) complement = -1, dl = 180. - dl, o[0] += 180. ;
    tr_oR(o, Rloc);
    tr_RR(R,Rloc,Rloc);	/* Matrix Rloc = Rloc * R */

    Rlon = (double *)Rloc ;
    tandl = tand(dl) ;

    call_down(0, QBOX_SOME, check_lon) ;

    return(0);
}

/*===========================================================================*
 *		Check Limits in Dec / Lat
 *===========================================================================*/

static int check_lat(qboxno, status)
/*+++++++
.PURPOSE Checks the qboxes which must be searched for a longitude range
.RETURNS 0
.REMARKS Recursive
.METHOD  The function fct is called for matching qboxes.
-------*/
  int qboxno;	/* IN: The qbox number */
  int status ;	/* IN: e.g. QBOX_SOME  */
{
  double uc[3], b, uq[4][3], zbound[2] ;
  int level, i;

    level = qbox_cu(qboxno, uc);

    tr_uu(uc, uc, Rlat);	/* Express qbox center in Local Frame	*/
    b = asind(uc[2]);

    if ((b <= (blim[0] - qboxes.radius[level])) ||
  	(b >= (blim[1] + qboxes.radius[level])) )
  		return(QBOX_NONE);

    if ((b >= (blim[0] + qboxes.radius[level])) &&
  	(b <= (blim[1] - qboxes.radius[level])) )
  		return(QBOX_ANY);

    if (level < 2) return(QBOX_SOME) ;

    /* Verify more closely: compute the Min/Max of Latitude within
       the Qbox. Use for this the min_prod function applied on
       z-axis (==R1[2])
    */

    ucorners(qboxno, uq) ;
    for (i=0; i<4; i++) tr_uu(uq[i], uq[i], Rlat) ;

    zbound[0] = 1 ; zbound[1] = -1 ;

    for (i=0; i<4; i++) {
        if (uq[i][2] < zbound[0]) zbound[0] = uq[i][2] ;
        if (uq[i][2] > zbound[1]) zbound[1] = uq[i][2] ;
	lambda = min_prod(R1[2], uq[i], uq[(i+1)&3], uc) ;
	if (lambda > 0) { 	/* Extremum */
	    if (uc[2] < zbound[0]) zbound[0] = uc[2] ;
	    if (uc[2] > zbound[1]) zbound[1] = uc[2] ;
	}
    }
    /*
    qbox_co(qboxno, uc) ; printf(
    "....%12s: %010.6f%+010.6f zlim=(%.6f,%.6f), zbound=(%.6f,%.6f)\n",
    qbox2a(qboxno), uc[0], uc[1], zlim[0], zlim[1], zbound[0], zbound[1]) ;
    */

    if ((zbound[0] > zlim[1]) || (zbound[1] < zlim[0]))
        return(QBOX_NONE);

    if ((zbound[0] > zlim[0]) && (zbound[1] < zlim[1]))
        return(QBOX_ANY);


    return(QBOX_SOME) ;
}

int qbox_dec(lim)
/*+++++++
.PURPOSE Select the qboxes which match a selection on Declination
.Returns 0
.REMARKS The `action routine' (see qbox_act) is called to gather the results.
.METHOD  Use qbox_lat with unit matrix.
-------*/
  double lim[2];	/* IN: Limit in Longitudes (the order is important!) */
{
    return(qbox_lat(lim,R1));
}

int qbox_lat(lim, R)
/*+++++++
.PURPOSE Select the qboxes which match a selection on Latitude.
.Returns 0
.REMARKS The `action routine' (see qbox_act) is called to gather the results.
.METHOD  Straightforward
-------*/
  double lim[2];	/* IN: Limit in Latitudes (reordered) */
  double R[3][3];	/* IN: Rotation matrix expressing longitude / latitude*/
{
    if (lim[0] <= lim[1])	blim[0] = lim[0], blim[1] = lim[1];
    else			blim[0] = lim[1], blim[1] = lim[0];
    zlim[0] = sind(blim[0]) ;
    zlim[1] = sind(blim[1]) ;

    Rlat = (double *)R ;

    call_down(0, QBOX_SOME, check_lat) ;

    return(0);
}

/*===========================================================================*
 *		Check Limits in both RA / Dec / Lon / Lat
 *===========================================================================*/

static int check_lonlat(qboxno, status)
/*+++++++
.PURPOSE Checks the qboxes which must be searched for a longitude/lat range
.RETURNS SOME / NONE / ANY
.METHOD  The function fct is called for matching qboxes.
-------*/
  int qboxno;	/* IN: The qbox number */
  int status ;	/* IN: e.g. QBOX_SOME  */
{
  int st1, st2 ;
    st1 = check_lat(qboxno, status) ;
    if (st1 == QBOX_NONE) return(QBOX_NONE) ;
    st2 = check_lon(qboxno, st1) ;
    return(st1&st2) ;	/* is QBOX_ANY only if both are	*/
}

int qbox_radec(alim, dlim)
/*+++++++
.PURPOSE Select the qboxes which match a selection on both RA + Declination
.REMARKS The `action routine' (see qbox_act) is called to gather the results.
.METHOD  Use qbox_lonlat with unit matrix.
-------*/
  double alim[2];	/* IN: Limit in Longitudes (the order is important!) */
  double dlim[2];	/* IN: Limit in Longitudes (the order is important!) */
{
    return(qbox_lonlat(alim, dlim, R1));
}

int qbox_lonlat(alim, dlim, R)
/*+++++++
.PURPOSE Select the qboxes which match a selection on Latitude.
.Returns 0
.REMARKS The `action routine' (see qbox_act) is called to gather the results.
-------*/
  double alim[2];	/* IN: Limit in Longitudes (order!!)  */
  double dlim[2];	/* IN: Limit in Latitudes (reordered) */
  double R[3][3];	/* IN: Rotation matrix expressing longitude / latitude*/
{
  double o[2], Rloc[3][3] ;

    if (dlim[0] <= dlim[1])	blim[0] = dlim[0], blim[1] = dlim[1];
    else			blim[0] = dlim[1], blim[1] = dlim[0];
    zlim[0] = sind(blim[0]) ;
    zlim[1] = sind(blim[1]) ;

    llim[0] = alim[0], llim[1] = alim[1];	/* Local Copy 	*/
    if (llim[0] > llim[1]) llim[1] += 360.e0;	/* Verify Range */

    /* Build the matrices */
    o[0] = (llim[0] + llim[1])/2.e0;	o[1] = 0. ;
    tr_oR(o, Rloc);
    tr_RR(R,Rloc,Rloc);			/* Matrix Rloc = Rloc * R */
    Rlon = (double *)Rloc ;
    dl = (llim[1] - llim[0])/2.e0;
    tandl = tand(dl) ;
    Rlat = (double *)R ;

    call_down(0, QBOX_SOME, check_lonlat) ;

    return(0);
}

/*===========================================================================*
 *		Check Limits in a Box
 *===========================================================================*/

int qbox_ob(center, adim)
/*+++++++
.PURPOSE Select the qboxes which match a selection in a Box
.RETURNS 0
.REMARKS The `action routine' (see qbox_act) is called to gather the results.
.METHOD  Use qbox_obrot with unit matrix.
-------*/
  double center[2];	/* IN: Center of box	*/
  double adim[2];	/* IN: Dimension of Box (in x/y, degrees) */
{
    return(qbox_obrot(center, adim, R1));
}

int qbox_obrot(center, adim, R)
/*+++++++
.PURPOSE Select the qboxes which match a selection in a Box
.RETURNS 0
.REMARKS The `action routine' (see qbox_act) is called to gather the results.
.METHOD  Use qbox_lonlat
-------*/
  double center[2];	/* IN: Center of box	*/
  double adim[2];	/* IN: Dimension of Box (in x/y, degrees) */
  double R[3][3] ;	/* IN: Rotation matrix, e.g. gal_2000	*/
{
  double Rloc[3][3], lon[2], lat[2] ;

    /* Build the matrix Rloc such that  u(loc) = Rloc * u(sys0 	*/
    tr_oR(center, Rloc) ;
    tr_RR(R, Rloc, Rloc);	/* Matrix Rloc = Rloc * R */

    /* In this local frame, the condition is just +/-Limits 	*/
    lat[1] = adim[1]/2. ;	lat[0] = -lat[1] ;
    lon[1] = adim[0]/2./cosd(lat[1]);
    if (lon[1] > 180.) lon[1] = 180. ;
    lon[0] = -lon[1] ;

    return(qbox_lonlat(lon, lat, Rloc));
}

/******************************************************************************
 *		Test Program
 ******************************************************************************/

#if INIT
#include <astropos.h>
#include <stdlib.h>
#define getS(buf)	getrim(buf, sizeof(buf), stdin)
char *getrim(char *buf, int bytes, FILE *f) {
    char *p; int i;
    if ((p = fgets(buf, bytes, f))) 
	for (i=strlen(buf); (--i>=0) && isspace(buf[i]); ) buf[i] = 0;
    return(p);
}
extern  double gal_1950[3][3];
static double oo[2];		/* To communicate with list_aqbox */
static int cur_level;

int list_qbox(qboxno, status)
/*+++++
.PURPOSE Test routine for qbox list
	(computes distance from target / qbox center)
.RETURN  status
-----*/
  int	qboxno;	/* IN: Qbox number */
  int	status;	/* IN: status as QBOX_ANY / QBOX_NONE / QBOX_SOME */
{
  double oc[2];
  char *p;

    qbox_co(qboxno, oc);
    switch(status) {
      case QBOX_NONE: p = "  none"; return(status) ;
      case QBOX_ANY : p = "++ any"; break;
      default       : p = "**some"; break;
    }
    printf(" %s object(s) in qbox %s (%8.4f,%8.4f). Distance %8.4f\n",
   	p, qbox8a(qboxno,0), oc[0], oc[1], dist_o(oo,oc));
    return(status);
}

int list_aqbox(qboxno, status)
/*+++++
.PURPOSE Test routine for qbox
.RETURN  status
-----*/
  int	qboxno;	/* IN: Qbox number */
  int	status;	/* IN: status as QBOX_ANY / QBOX_NONE / QBOX_SOME */
{
  double oc[2];
  char *p;

    qbox_co(qboxno, oc);
    switch(status) {
      case QBOX_NONE: p = "  none"; return(status);
      case QBOX_ANY : p = "++ any"; break;
      default       : p = "**some"; break;
    }
    printf(" %s object(s) in qbox %s (%8.4f,%8.4f)\n",
   	p, qbox8a(qboxno,0), oc[0], oc[1]);
    return(status);
}

#if 0	/* Test with Galactic Coordinates */
int list_gqbox(qboxno, status)
/*+++++
.PURPOSE Test routine for qbox in Galactic Coordinates
.RETURN  status
-----*/
  int	qboxno;	/* IN: Qbox number */
  int	status;	/* IN: sttus as QBOX_ANY / QBOX_NONE / QBOX_SOME */
{
  double oc[2], u[3], og[2];
  char *p;
  int  n, k;

    qbox_cu(qboxno, u);		/* Center in Equatorial */

    switch(status) {
      case QBOX_NONE: p = "  none"; return(status);
      case QBOX_ANY : p = "++ any"; break;
      default       : p = "**some"; break;
    }
    tr_uo(u, oc);			/* Equatorial		*/
    tr_f4g(u, u);			/* To galactic...	*/
    tr_uo(u, og);
    printf(" %s object(s) in qbox %s (%8.4f,%8.4f) G%08.4f,%+08.4f\n",
   	p, qbox2a(qboxno), oc[0], oc[1], og[0], og[1]);
    return(status);
}
#endif

int main(int argc, char **argv)
/*+++++
.PURPOSE Test implementation
.RETURN To OS
-----*/
{
  char 	buf[80];
  double 	oc[2], o[2], uc[3], u[3], p[2];
  double	rc, qbox_radius();
  double	lim[4];
  int 	the_level, i, lev, n, pno;
#if 0
  static	double R0[3][3] = {
		{1.,0.,0.},
		{0.,1.,0.},
		{0.,0.,1.}	};
#endif

    qbox_init();
    qbox_act(list_qbox);

    printf("Choose your Level in range [0..%d]: ", QBOX_LEVEL); getS(buf);
    qbox_set(atoi(buf));
    the_level = qbox_get();
    printf("----Level: %d (qbox_number between %d and %d\n", the_level,
	(9<<the_level)<<the_level, ((15<<the_level)<<the_level)-1) ;

    /* Provide the list of interesting qboxes */
    while(1) {
	printf("----Give a qbox number at lower level (0 to terminate): ");
	if (!getS(buf)) break;
	if (!buf[0])    break ;
	n = atoi(buf) ;
	if (n <= 0) break ;
	printf("====Box %8d = %-12s[%s]  Level=%d, R=%10.6f, S=%12.6f\n",
	    n, qbox2a(n), qbox8a(n,0), qbox_level(n), 
	    qbox_radius(n), qbox_surf(n)) ;
	printf("----List of qboxes (level %d) included in %d (level %d):\n",
	    the_level, n, qbox_level(n)) ;
	for (i = qbox_next(n); i > 0; i = qbox_next(0))
	    printf("%10d", i) ;
	putchar('\n') ;
    }

    while(1) {
	printf("Give position: ");
  	if(!getS(buf))	break;
	if(!buf[0]) break ;
  	if (tr_ao(buf, strlen(buf), oo) < 0) continue ;
  	printf("Input coordinates: %12.6f, %12.6f\n", oo[0], oo[1]);
  	tr_ou(oo, u);
  	printf("Input unit vector: %12.9f, %12.9f, %12.9f\n",
  		u[0], u[1], u[2]);
  	n = tr_up6(u, p, 0);
  	printf("Cube projectin #%d: %12.6f, %12.6f\n", n, p[0], p[1]);
  	tr_pu6(p, u, n);
  	printf("      unit vector: %12.9f, %12.9f, %12.9f\n",
  		u[0], u[1], u[2]);
  	tr_uo(u, o);
  	printf("      coordinates: %12.6f, %12.6f\n", o[0], o[1]);

	for (lev=0; lev<=the_level; lev++){
	    qbox_set(lev);
	    cur_level = lev;
	    printf("===============Level %d ==================\n", lev);

  	    n = qbox_u(u);		/* Compute Qbox Number */
  	    printf("Cube qbox number  : %s[%s] (%d)\n", 
		    qbox2a(n), qbox8a(n,0), n);

  	    pno = qbox_cu(n, uc);
  	    tr_uo(uc, oc);
  	    printf("Qbox center       : %12.6f, %12.6f [Dist=%6.3f/%6.3f]\n",
  		oc[0], oc[1], dist_o(o,oc),qbox_radius(n));
	    i = tr_up6(uc, p, 0) ;
  	    printf(".....projectin #%d: %12.6f, %12.6f\n", i, p[0], p[1]);
	}
	printf("Give Radius: ");  	if(!getS(buf))	break;
  	tr_ao0(buf, strlen(buf), &rc);
  	printf("....radius (degrees):%9.4f\n", rc);
  	qbox_ur(u, rc);
    }

    qbox_act(list_aqbox);
    while(1) {
	printf("Give lower longitude limit: "); if (!getS(buf)) break;
	if(!buf[0]) break ;
	tr_ao0(buf, strlen(buf), &lim[0]);
	printf("Give upper longitude limit: "); if (!getS(buf)) break;
	tr_ao0(buf, strlen(buf), &lim[1]);
  	qbox_ra(lim);
	printf("+Add lower latitude limit: "); if (!getS(buf)) break;
	if(!buf[0]) break ;
	tr_ao0(buf, strlen(buf), &lim[2]);
	printf("+Add upper latitude limit: "); if (!getS(buf)) break;
	tr_ao0(buf, strlen(buf), &lim[3]);
	qbox_radec(lim, lim+2) ;
    }

#if 0	/* Test with Galactic coordinates */
    qbox_act(list_gqbox);
    while(1) {
	printf("Give lower Galactic longitude limit: "); if (!gets(buf)) break;
	if(!buf[0]) break ;
	tr_ao0(buf, strlen(buf), &lim[0]);
	printf("Give upper Galactic longitude limit: "); if (!gets(buf)) break;
	tr_ao0(buf, strlen(buf), &lim[1]);
  	qbox_lon(lim, gal_1950);
	printf("+Add lower Galactic latitude limit: "); if (!gets(buf)) break;
	if(!buf[0]) break ;
	tr_ao0(buf, strlen(buf), &lim[2]);
	printf("+Add upper Galactic latitude limit: "); if (!gets(buf)) break;
	tr_ao0(buf, strlen(buf), &lim[3]);
  	qbox_lonlat(lim, lim+2, gal_1950);
    }
#endif

    qbox_act(list_aqbox);
    while(1) {
	printf("Give lower latitude limit: "); if (!getS(buf)) break;
	if(!buf[0]) break ;
	tr_ao0(buf, strlen(buf), &lim[0]);
	printf("Give upper latitude limit: "); if (!getS(buf)) break;
	tr_ao0(buf, strlen(buf), &lim[1]);
  	qbox_dec(lim);
    }

#if 0	/* Test with Galactic coordinates */
    qbox_act(list_gqbox);
    while(1) {
	printf("Give lower Galactic latitude limit: "); if (!gets(buf)) break;
	if(!buf[0]) break ;
	tr_ao0(buf, strlen(buf), &lim[0]);
	printf("Give upper Galactic latitude limit: "); if (!gets(buf)) break;
	tr_ao0(buf, strlen(buf), &lim[1]);
  	qbox_lat(lim, gal_1950);
    }
#endif
    exit(0);
}
#endif
